/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iomanip>

#include "infiniband/ibdm/SubnMgt.h"
#include "infiniband/ibdm/CredLoops.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include "infiniband/ibutils/output_stream_primitives.h"

extern IBDiagClbck ibDiagClbck;


u_int16_t AdditionalRoutingData::sw_supports_rn_count = 0;
u_int16_t AdditionalRoutingData::max_sub_group_direction_block = 0;
u_int8_t AdditionalRoutingData::max_direction_block = 0;
u_int8_t AdditionalRoutingData::max_plft = 0;
u_int16_t AdditionalRoutingData::max_string_block = 0;
u_int8_t AdditionalRoutingData::max_ports_block = 0;


/****************************************************/
/******************private methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpUCFDBSInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // check if not a switch
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_SW_NODE)
            continue;

        stringstream ss;
        ss << "osm_ucast_mgr_dump_ucast_routes: Switch " << PTR(p_curr_node->guid_get()) << endl;

        u_int8_t maxPLFT =  p_curr_node->getMaxPLFT();
        for (u_int8_t pLFT = 0; pLFT <= maxPLFT ; ++pLFT) {
            ss << "PLFT_NUM: " << DEC(pLFT) << endl
               << "LID    : Port : Hops : Optimal" << endl;

            u_int16_t lfdbTop = (u_int16_t)p_curr_node->LFT[pLFT].size();

            if (p_curr_node->isPLFTEnabled())
                lfdbTop = p_curr_node->getLFDBTop(pLFT);

            for (unsigned int i = 1; i <= lfdbTop; ++i) {
                u_int8_t port_for_curr_lid = p_curr_node->getLFTPortForLid((lid_t)i, pLFT);

                // memset(curr_uc_fdb_line, 0, sizeof(curr_uc_fdb_line));
                if (port_for_curr_lid == IB_LFT_UNASSIGNED)
                    ss << PTR(i, 4) << " : UNREACHABLE";
                else
                    ss << PTR(i, 4) << " : " << DEC(port_for_curr_lid, 3, '0') << "  : 00   : yes";

                ss << endl;
            }
            ss << endl;
        }
        sout << ss.rdbuf() << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::GetAREnabledNum(u_int64_t& ar_enabled_num,
                            u_int64_t& hbf_enabled_num,
                            u_int64_t& hbf_sup_num)
{
    IBDIAG_ENTER;

    ar_enabled_num = 0;
    hbf_enabled_num = 0;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null pointer in Switches.");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //HBF is a subset of AR configuration. AR is enabled on a switch if bit AR_INF0.E is on
        //isAREnable and isHBFEnable places stronger restrictions by checking that
        //at least a single SL is configured to work with AR or HBF.
        //So, it is theoretically possible that SL exists only for HBF or AR.
        if (p_curr_node->isAREnable() || p_curr_node->isHBFEnable())
            ++ar_enabled_num;

        if (p_curr_node->isHBFSupported())
            ++hbf_sup_num;

        if (p_curr_node->isHBFEnable())
            ++hbf_enabled_num;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpMCFDBSInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char buffer[2096];
    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {
        IBNode *p_curr_node = *nI;

        if (!p_curr_node) {
            this->SetLastError("DB error - found null pointer in Switches.");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        snprintf(buffer, sizeof(buffer),
                 "Switch " U64H_FMT "\n"   \
                 "LID    : Out Port(s)",
                 p_curr_node->guid_get());
        sout << buffer << endl;
        for (unsigned int i = 0; i <= p_curr_node->MFT.size(); ++i) {
            list_phys_ports ports_for_curr_lid = p_curr_node->getMFTPortsForMLid((lid_t)(i + 0xc000));
            if (ports_for_curr_lid.empty() == true)
                continue;

            snprintf(buffer, sizeof(buffer), U16H_FMT " : ", i + 0xc000);
            sout << buffer;
            for (list_phys_ports::iterator it = ports_for_curr_lid.begin();
                    it != ports_for_curr_lid.end(); ++it) {
                snprintf(buffer, sizeof(buffer), "0x%03x ", *it);
                sout << buffer;
            }
            sout << endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::ReportNonUpDownCa2CaPaths(IBFabric *p_fabric, list_pnode rootNodes, string& output)
{
    IBDIAG_ENTER;
    if (SubnRankFabricNodesByRootNodes(p_fabric, rootNodes)) {
        output += "-E- fail to rank the fabric by the given root nodes.\n";
        IBDIAG_RETURN(1);
    }
    IBDIAG_RETURN(SubnReportNonUpDownCa2CaPaths(p_fabric));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
int IBDiag::RetrieveUCFDBSEntry(IBNode *p_curr_node, direct_route_t *p_curr_route,
                                list_p_fabric_general_err& retrieve_errors,
                                ProgressBarNodes &progress_bar, clbck_data_t &clbck_data, int &ret,
                                const set_lid &lids)
{
    IBDIAG_ENTER;

    if (!p_curr_node || !p_curr_node->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    p_curr_node->appData1.val = 0;

        //check if hca
    if (p_curr_node->type == IB_CA_NODE || p_curr_node->type == IB_RTR_NODE)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (p_curr_node->isAREnable() || p_curr_node->isHBFEnable())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

        //get switch info for this node
    const SMP_SwitchInfo *p_curr_switch_info = this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
    if (!p_curr_switch_info)   //no switch info - didn't build db or no response for this node
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

        //get direct route for this node
    if (!p_curr_route) {
        p_curr_route = this->GetDR(p_curr_node);
        if (!p_curr_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }

    //check LinearFDBTop value
    if (p_curr_switch_info->LinearFDBTop > (0xc000 - 1)) {
        retrieve_errors.push_back(
                    new FabricErrNodeWrongConfig(p_curr_node, "LinearFDBTop exceeds 0xc000"));
        ret = IBDIAG_ERR_CODE_FABRIC_ERROR;
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    //resize LFT table
    p_curr_node->resizeLFT((u_int16_t)(p_curr_switch_info->LinearFDBTop+1));

    //get ucfdbs info
    u_int16_t num_of_lid_to_port_blocks = (u_int16_t)((p_curr_switch_info->LinearFDBTop +
                                IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS) / IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS);
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has LinearFDBTop=%u ==> Blocks=%u\n",
                                p_curr_node->getName().c_str(),
                                p_curr_switch_info->LinearFDBTop, num_of_lid_to_port_blocks);

    vector<bool> blocks = vector<bool>(num_of_lid_to_port_blocks, false);
    this->MarkBlocksToSend(blocks, lids, IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS);
    for (u_int16_t curr_block = 0; curr_block < num_of_lid_to_port_blocks; ++curr_block) {
        if (!blocks[curr_block])
            continue;

        SMP_LinearForwardingTable curr_linear_forwarding_table = {0};

        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_data2 = (void *)(uintptr_t)curr_block;

        progress_bar.push(p_curr_node);

        this->ibis_obj.SMPLinearForwardingTableGetByDirect(p_curr_route, curr_block,
                    &curr_linear_forwarding_table, &clbck_data);

        int rc = ibDiagClbck.GetState();
        if (rc)
            IBDIAG_RETURN(rc);

        if (p_curr_node->appData1.val != 0)
            break; //proceed to the next_node;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::MarkBlocksToSend(vector<bool> &blocks, const set_lid &lids, size_t block_size) const
{
    IBDIAG_ENTER;

    if (lids.empty()) {
        fill(blocks.begin(), blocks.end(), true);
        IBDIAG_RETURN_VOID;
    }

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Marking blocks to be send; total blocks=%lu,"
                                " single block size=%lu\n",
                                blocks.size(), block_size);

    for (lid_t lid : lids) {
        if (!lid)
           continue;

        //assume situation when block_size is 16
        //and we are checking lid=16 and we have 24 lids -- 2 blocks then block=1 should be allowed
        //min and max lids provided by a block
        //min_lid = (block * num_blocks);
        //max_lid = (block * num_blocks + (num_blocks - 1));

        size_t index = (lid)/block_size;
        if (index < blocks.size()) {
            blocks[index] = true;
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "block=%lu is allowed to be sent for lid=%u\n",
                                        index, lid);
            continue;
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Calculated wrang block=%lu for lid=%u\n", index, lid);
    }

    IBDIAG_RETURN_VOID;
}

int IBDiag::RetrieveUCFDBSInfo(list_p_fabric_general_err& retrieve_errors,
                            direct_route_list &from_routes, const set_lid &lids)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    int ret = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPLinearForwardingTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes
    for (const auto& entry : from_routes) {
        rc = this->RetrieveUCFDBSEntry(entry.first, entry.second,
                                    retrieve_errors, progress_bar, clbck_data, ret, lids);
        if (rc)
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            ret = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(ret);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveUCFDBSInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    int ret = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPLinearForwardingTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        rc = this->RetrieveUCFDBSEntry(this->fabric_extended_info.getNodePtr(i), NULL,
                                    retrieve_errors, progress_bar, clbck_data, ret);
        if (rc)
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            ret = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(ret);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveMCFDBSInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPMulticastForwardingTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;

    // go over all nodes:
    struct SMP_SwitchInfo *p_curr_switch_info;
    struct SMP_MulticastForwardingTable curr_multicast_forwarding_table;
    CLEAR_STRUCT(curr_multicast_forwarding_table);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_curr_node->appData1.val = 0;

        //check if hca or router
        if (p_curr_node->type == IB_CA_NODE || p_curr_node->type == IB_RTR_NODE)
            continue;

        //nodes with num of ports > IBDIAG_MAX_SUPPORTED_NODE_PORTS not supported in our code
        if (p_curr_node->numPorts > IBDIAG_MAX_SUPPORTED_NODE_PORTS) {
            retrieve_errors.push_back(
                    new FabricErrNodeWrongConfig(
                            p_curr_node,
                            "number of ports exceeds maximum supported, can not support fetch of mcfdbs"));
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;   //try next node
        }

        //get switch info for this node
        p_curr_switch_info = this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
        if (!p_curr_switch_info)   //no switch info - didn't build db or no response for this node
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        //check MCastFDBCap value
        if (p_curr_switch_info->MCastFDBCap > (0xffff - 0xc000 + 1)) {
            retrieve_errors.push_back(
                    new FabricErrNodeWrongConfig(p_curr_node, "MCastFDBCap exceeds range"));

            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;   //try next switch
        }

        //get mcfdbs info
        u_int16_t max_mcast_fdb = 0;

        if (p_curr_switch_info->MCastFDBTop == 0 ) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has MCastFDBTop=%u - use CastFDBCap:%u\n",
                       p_curr_node->getName().c_str(),
                       p_curr_switch_info->MCastFDBTop,
                       p_curr_switch_info->MCastFDBCap);
            max_mcast_fdb =  p_curr_switch_info->MCastFDBCap;
        }
        else if (p_curr_switch_info->MCastFDBTop < 0xc000 ) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has MCastFDBTop=%u - no multicast configuration\n",
                    p_curr_node->getName().c_str(), p_curr_switch_info->MCastFDBTop);
            continue;
        } else {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has MCastFDBTop=%u\n",
                       p_curr_node->getName().c_str(),
                       p_curr_switch_info->MCastFDBTop);
            max_mcast_fdb = (u_int16_t)(p_curr_switch_info->MCastFDBTop - 0xc000 + 1);
        }

        lid_t num_of_lid_to_port_blocks = (lid_t)((max_mcast_fdb +
                                                   IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS - 1) / IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has max_mcast_fdb=%u ==> Blocks=%u\n",
                p_curr_node->getName().c_str(), max_mcast_fdb, num_of_lid_to_port_blocks);

        for (u_int16_t curr_block = 0; curr_block < num_of_lid_to_port_blocks; ++curr_block) {
            for (u_int8_t curr_port_group = 0;
                    curr_port_group < ((p_curr_node->numPorts + IBIS_IB_MAD_SMP_MFT_PORT_MASK_SIZE - 1) / IBIS_IB_MAD_SMP_MFT_PORT_MASK_SIZE);
                    ++curr_port_group) {

                clbck_data.m_data1 = p_curr_node;
                clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
                clbck_data.m_data3 = (void *)(uintptr_t)curr_port_group;

                progress_bar.push(p_curr_node);
                this->ibis_obj.SMPMulticastForwardingTableGetByDirect(p_curr_direct_route,
                        curr_port_group, curr_block, &curr_multicast_forwarding_table, &clbck_data);

            }   //port groups loop

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //not try to get all the rest of the tables
        }   //blocks loop

    }   //nodes loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::RetrieveVLArbitrationTable(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarPorts progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPVLArbitrationGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    SMP_PortInfo *p_curr_port_info;
    SMP_VLArbitrationTable curr_vl_arbitration_table;
    CLEAR_STRUCT(curr_vl_arbitration_table);

    // go over all nodes:
    for (u_int32_t nodes_i = 0; nodes_i < this->fabric_extended_info.getNodesVectorSize(); ++nodes_i) {

        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nodes_i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_curr_node->appData1.val = 0;

        // go over all ports of this node:
        for (phys_port_t i = 1; i <= p_curr_node->numPorts; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);

            // Skip ports that don't exist or are down or have undefined state
            // or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_curr_port_info)
                continue;

            //get direct route for this port
            direct_route_t *p_curr_direct_route = this->GetDR(p_curr_port);

            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to port=%s",
                        p_curr_port->getName().c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            clbck_data.m_data1 = p_curr_port;
            progress_bar.push(p_curr_port);

            for (u_int32_t block = LOWER_LOW_PRIORITY_VL_ARBITRATION_TABLE;
                    block <= UPPER_HIGH_PRIORITY_VL_ARBITRATION_TABLE; ++block) {

                switch (block) {
                    case LOWER_LOW_PRIORITY_VL_ARBITRATION_TABLE:
                        if (p_curr_port_info->VLArbLowCap == 0)
                            continue;
                        break;

                    case UPPER_LOW_PRIORITY_VL_ARBITRATION_TABLE:
                        if (!(p_curr_port_info->VLArbLowCap > 32))
                            continue;
                        break;

                    case LOWER_HIGH_PRIORITY_VL_ARBITRATION_TABLE:
                        if (p_curr_port_info->VLArbHighCap == 0)
                            continue;
                        break;

                    case UPPER_HIGH_PRIORITY_VL_ARBITRATION_TABLE:
                        if (!(p_curr_port_info->VLArbHighCap > 32))
                            continue;
                        break;
                }

                u_int32_t attribute_modifier = 0;
                if (p_curr_node->type == IB_SW_NODE)
                    attribute_modifier = i; // switch port in bits 0-7

                // send to all blocks. block num in bits 16-32.
                clbck_data.m_data2 = (void *)(uintptr_t)block;
                attribute_modifier |= (block << 16);

                this->ibis_obj.SMP_VLArbitrationMadGetByDirect(p_curr_direct_route,
                                                               attribute_modifier,
                                                               &curr_vl_arbitration_table,
                                                               &clbck_data);
            } // blocks

            if (ibDiagClbck.GetState())
                goto exit;

        } // ports
    } // nodes

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::DumpVLArbitrationCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_VL_ARBITRATION_TBL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNum,Priority,VL,Weight" << endl;

    csv_out.WriteBuf(sstream.str());

    // go over all nodes:
    for (u_int32_t nodes_i = 0; nodes_i < this->fabric_extended_info.getNodesVectorSize(); ++nodes_i) {

        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nodes_i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        // go over all ports of this node:
        for (phys_port_t i = 1; i <= p_curr_node->numPorts; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);

            // Skip ports that don't exist or are down or have undefined state
            // or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            SMP_PortInfo *p_curr_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_curr_port_info)
                continue;

            for (u_int32_t block = LOWER_LOW_PRIORITY_VL_ARBITRATION_TABLE;
                    block <= UPPER_HIGH_PRIORITY_VL_ARBITRATION_TABLE; ++block) {

                SMP_VLArbitrationTable* p_curr_vl_arbitration_table =
                        this->fabric_extended_info.getSMPVLArbitrationTable(
                                p_curr_port->createIndex, block);

                if (!p_curr_vl_arbitration_table)
                    continue;

                u_int32_t cap = 0;
                string priority;

                switch (block) {
                    case LOWER_LOW_PRIORITY_VL_ARBITRATION_TABLE:
                        cap = p_curr_port_info->VLArbLowCap;
                        priority = "low";
                        break;

                    case UPPER_LOW_PRIORITY_VL_ARBITRATION_TABLE:
                        cap = p_curr_port_info->VLArbLowCap - 32;
                        priority = "low";
                        break;

                    case LOWER_HIGH_PRIORITY_VL_ARBITRATION_TABLE:
                        cap = p_curr_port_info->VLArbHighCap;
                        priority = "high";
                        break;

                    case UPPER_HIGH_PRIORITY_VL_ARBITRATION_TABLE:
                        cap = p_curr_port_info->VLArbHighCap - 32;
                        priority = "high";
                        break;
                }

                for (u_int32_t entry = 0; entry < cap && entry < 32; ++entry) {

                    sstream.str("");

                    // NodeGUID,PortGUID,PortNum,Priority,VL,Weight
                    sstream << PTR(p_curr_node->guid_get())                       << ","
                            << PTR(p_curr_port->guid_get())                       << ","
                            << +i  /* port num */                                 << ","
                            << priority                                           << ","
                            << +p_curr_vl_arbitration_table->VLArb[entry].VL      << ","
                            << +p_curr_vl_arbitration_table->VLArb[entry].Weight  << endl;

                    csv_out.WriteBuf(sstream.str());

                } // entries
            } // blocks
        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_VL_ARBITRATION_TBL);

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteUCFDBSFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Unicast FDBS",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpUCFDBSInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteMCFDBSFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Multicast FDBS",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpMCFDBSInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

/////////
/// PLFT
/////////

int IBDiag::RetrievePLFTData(list_p_fabric_general_err& retrieve_errors, direct_route_list &from_routes, bool is_ibdiagpath)
{
    IBDIAG_ENTER;

    direct_route_list plft_routes;
    int rc = this->GetSwitchesDirectRouteList(from_routes, plft_routes, NULL);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = this->RetrievePLFTInfo(retrieve_errors, plft_routes, is_ibdiagpath);
    if (rc)
        IBDIAG_RETURN(rc);

    if (plft_routes.empty())
        IBDIAG_RETURN(rc);

    GetDiscoverFabricPtr()->setPLFTEnabled();

    rc = this->RetrievePLFTMapping(retrieve_errors, plft_routes, is_ibdiagpath);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = this->RetrievePLFTTop(retrieve_errors, plft_routes, is_ibdiagpath);

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTData(list_p_fabric_general_err& retrieve_errors,
                             unsigned int &supportedDev)
{
    IBDIAG_ENTER;
    supportedDev = 0;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    direct_route_list directRouteList;

    rc = GetSwitchesDirectRouteList(directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrievePLFTInfo(retrieve_errors, directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    if (directRouteList.empty())
        IBDIAG_RETURN(rc);

    supportedDev = (unsigned int)directRouteList.size();
    GetDiscoverFabricPtr()->setPLFTEnabled();

    rc = RetrievePLFTMapping(retrieve_errors, directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrievePLFTTop(retrieve_errors, directRouteList);

    IBDIAG_RETURN(rc);

}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTInfo(list_p_fabric_general_err& retrieve_errors,
                            direct_route_list &directRouteList, bool is_ibdiagpath)
{
    IBDIAG_ENTER;

    if (!is_ibdiagpath) {
        if (!this->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    }

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPPLFTInfoGetClbck>;

    // go over all nodes:
    struct ib_private_lft_info plft_info;
    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();

    CLEAR_STRUCT(plft_info);

    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;


        if (!this->capability_module.IsSupportedSMPCapability(
            p_curr_node, EnSMPCapIsPrivateLinearForwardingSupported) ||
            !this->capability_module.IsSupportedSMPCapability(
            p_curr_node, EnSMPCapIsPortSLToPrivateLFTMapSupported))
            continue;

        clbck_data.m_data1 = p_curr_node;
        this->ibis_obj.SMPPLFTInfoMadGetSetByDirect(p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &plft_info, &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    if (rc == IBDIAG_SUCCESS_CODE) {
        struct SMP_SwitchInfo *p_curr_switch_info = NULL;
        //remove nodes that do not support PLFT
        iter = directRouteList.begin();
        direct_route_list_iter prevIter = directRouteList.begin();
        for (;iter != directRouteList.end(); ) {
            prevIter = iter;
            ++iter;
            p_curr_node = prevIter->first;
            if (!p_curr_node->isPLFTEnabled()) {
                //get switch info for this node
                p_curr_switch_info =
                    this->fabric_extended_info.getSMPSwitchInfo(
                        p_curr_node->createIndex);

                if (!p_curr_switch_info)   //no switch info - didn't build db or no response for this node
                    p_curr_node->setLFDBTop(0, 0);
                else
                   p_curr_node->setLFDBTop(0, p_curr_switch_info->LinearFDBTop);

                directRouteList.erase(prevIter);
            }
        }
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTMapping(list_p_fabric_general_err& retrieve_errors,
                                direct_route_list &directRouteList, bool is_ibdiagpath)
{
    IBDIAG_ENTER;

    if (!is_ibdiagpath) {
        if (!this->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    }

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPPortSLToPrivateLFTMapGetClbck>;

    // go over all nodes:
    struct ib_port_sl_to_private_lft_map plft_map;
    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();

    CLEAR_STRUCT(plft_map);

    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;

        p_curr_node->appData1.val = 0;

        //get ucfdbs info
        u_int8_t num_of_port_blocks = (u_int8_t)((p_curr_node->numPorts +
                                                  IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE) /
                                                 IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has %u ports ==> Blocks=%u\n",
                p_curr_node->getName().c_str(), p_curr_node->numPorts, num_of_port_blocks);

        for (u_int8_t curr_block = 0; curr_block < num_of_port_blocks; ++curr_block) {
            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
            this->ibis_obj.SMPPortSLToPrivateLFTMapGetSetByDirect(
                p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                curr_block,
                &plft_map, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //goto next_node;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTTop(list_p_fabric_general_err& retrieve_errors,
                            direct_route_list &directRouteList, bool forced)
{
    IBDIAG_ENTER;

    if (!forced) {
        if (!this->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    }

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPPrivateLFTTopGetClbck>;

    // go over all nodes:
    struct ib_private_lft_map plft_map;
    CLEAR_STRUCT(plft_map);
    plft_map.LFT_TopEn = 1;

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();
    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;

        p_curr_node->appData1.val = 0;

        for (u_int8_t pLFT = 0; pLFT <= p_curr_node->getMaxPLFT(); ++pLFT) {
            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = (void *)(uintptr_t)pLFT;

            this->ibis_obj.SMPPLFTMapMadGetSetByDirect(
                p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                pLFT,
                &plft_map, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //goto next_node;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//////
// AR
//////

int IBDiag::GetSwitchesDirectRouteListEntry(IBNode *p_node, direct_route *p_route,
                                    direct_route_list &ar_routes,
                                    AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;

    if (!p_node || !p_node->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //check if hca
    if (p_node->type == IB_CA_NODE)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

        // skip if AR and HBF are not enabled
    if (!p_node->isAREnable() && !p_node->isHBFEnable())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //get direct route for this node
    if (!p_route) {
        p_route = this->GetDR(p_node);
        if (!p_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }

    ar_routes.push_back(make_pair(p_node, p_route));
    p_node->appData1.val = 0;
    p_node->appData2.val = 0;

    if (p_routing_data_map) {
        adaptive_routing_info *p_ar_info = this->fabric_extended_info.getARInfo(p_node->createIndex);
        if (p_ar_info)
            AddRNDataMapEntry(p_routing_data_map,
                            p_node,
                            p_route,
                            p_ar_info);
    }
    else
        p_node->appData3.ptr = NULL;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::GetSwitchesDirectRouteList(direct_route_list &ar_routes,
                                       AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        rc = this->GetSwitchesDirectRouteListEntry(this->fabric_extended_info.getNodePtr(i), NULL,
                                                ar_routes, p_routing_data_map);
        if (rc)
            break;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::GetSwitchesDirectRouteList(direct_route_list &from_routes,
                            direct_route_list &ar_routes,
                            AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (auto& entry : from_routes) {
        rc = this->GetSwitchesDirectRouteListEntry(entry.first, entry.second,
                                                ar_routes, p_routing_data_map);
        if (rc)
            break;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::RetrieveHBFData(list_p_fabric_general_err& retrieve_errors,
                            unsigned int &supportedDev)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = this->RetrieveHBFConfig(retrieve_errors, supportedDev);
    if (rc)
        IBDIAG_RETURN(rc);

    //retrieve WHBFConfig
    rc = this->RetrieveWeightsHBFConfig(retrieve_errors);

    //Set FABRIC_ERROR only after all available data is collected
    if (!rc && !retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

int IBDiag::Retrieve_pFRNData(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = this->Retrieve_pFRNConfig(retrieve_errors);

    IBDIAG_RETURN(rc);
}

int IBDiag::RetrieveARData(list_p_fabric_general_err& retrieve_errors,
                        direct_route_list &from_routes,
                        AdditionalRoutingDataMap *p_routing_data_map,
                        bool is_ibdiagpath, const set_lid  &lids)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    direct_route_list ar_routes;
    rc = this->GetSwitchesDirectRouteList(from_routes, ar_routes, p_routing_data_map);

    IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "AR is supported on =%d\n", (int)ar_routes.size());

    if (ar_routes.empty())
        IBDIAG_RETURN(rc);

    rc = this->RetrieveARGroupTable(retrieve_errors, ar_routes, is_ibdiagpath);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrieveARLinearForwardingTable(retrieve_errors, ar_routes, is_ibdiagpath, lids);

    //Set FABRIC_ERROR only after all available data is collected
    if (!rc && !retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARData(list_p_fabric_general_err& retrieve_errors,
                           unsigned int &supportedDev,
                           AdditionalRoutingDataMap *p_routing_data_map,
                           bool skip_lfts)
{
    IBDIAG_ENTER;
    supportedDev = 0;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    direct_route_list directRouteList;

    rc = GetSwitchesDirectRouteList(directRouteList, p_routing_data_map);
    if (rc)
        IBDIAG_RETURN(rc);

    if (directRouteList.empty())
        IBDIAG_RETURN(rc);

    supportedDev = (unsigned int)directRouteList.size();

    rc = RetrieveARGroupTable(retrieve_errors, directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    if (!skip_lfts)
        rc = RetrieveARLinearForwardingTable(retrieve_errors, directRouteList);

    //Set FABRIC_ERROR only after all available data is collected
    if (!rc && !retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildARInfoDBEntry(ProgressBarNodes &progress_bar, clbck_data_t &clbck_data,
                            IBNode *p_node, direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;

    adaptive_routing_info ar_info = {0};

    if (!p_node || !p_node->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    // skip HCAs
    if (p_node->type == IB_CA_NODE)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    // check if Adaptive Routing Supported.
    if (!this->capability_module.IsSupportedSMPCapability(
                                p_node, EnSMPCapIsAdaptiveRoutingRev1Supported))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //get direct route for this node
    if (!p_direct_route)
        p_direct_route = this->GetDR(p_node);

    if (!p_direct_route) {
        this->SetLastError("DB error - can't find direct route to node=%s", p_node->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    clbck_data.m_data1 = p_node;
    progress_bar.push(p_node);

    this->ibis_obj.SMPARInfoGetSetByDirect(p_direct_route,
                                        IBIS_IB_MAD_METHOD_GET,
                                        false, //get_cap
                                        &ar_info,
                                        &clbck_data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//this function is executed aby the ibdiagpath while discovery is running
int IBDiag::PathDisc_BuildARInfo(list_p_fabric_general_err& ar_errors, direct_route_list& routes)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &ar_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPARInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (auto& entry: routes) {
        rc = this->BuildARInfoDBEntry(progress_bar, clbck_data,
                            entry.first, entry.second);
        if (rc)
            break;

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildARInfoDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPARInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        rc = this->BuildARInfoDBEntry(progress_bar, clbck_data,
                            this->fabric_extended_info.getNodePtr(i), NULL);
        if (rc)
            break;

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARGroupTable(list_p_fabric_general_err& retrieve_errors,
                                 direct_route_list & directRouteList, bool is_ibdiagpath)
{
    IBDIAG_ENTER;

    if (!is_ibdiagpath) {
        if (!this->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    }

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPARGroupTableGetClbck>;

    direct_route_list routeList = directRouteList;

    // go over all nodes:
    struct ib_ar_group_table ar_group_table;
    CLEAR_STRUCT(ar_group_table);

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    bool isSend = false;

    while (!routeList.empty()) {
        direct_route_list_iter iter = routeList.begin();
        direct_route_list_iter prevIter = routeList.begin();
        isSend = false;
        for ( ; iter != routeList.end() ; ) {

            p_curr_node = iter->first;
            p_curr_direct_route = iter->second;
            prevIter = iter;
            ++iter;

            if (p_curr_node->appData2.val == RETRIEVE_STAGE_REC_WAIT)
                continue;
            else if (p_curr_node->appData2.val == RETRIEVE_STAGE_REC_DONE) {
                routeList.erase(prevIter);
                continue;
            }

            clbck_data.m_data1 = p_curr_node;
            p_curr_node->appData2.val = RETRIEVE_STAGE_REC_WAIT;

            this->ibis_obj.SMPARGroupTableGetSetByDirect(
                p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                (u_int16_t)p_curr_node->appData1.val, //group_block
                0, //group_table
                0, //pLFT - global group
                &ar_group_table, &clbck_data);

            isSend = true;
        }

        if(!isSend && !routeList.empty())
            this->ibis_obj.MadRecAll();

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        //else if (!retrieve_errors.empty())
        //    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARLinearForwardingTable(list_p_fabric_general_err& retrieve_errors,
                                            direct_route_list & directRouteList,
                                            bool is_ibdiagpath, const set_lid  &lids)
{
    IBDIAG_ENTER;

    if (!is_ibdiagpath) {
        if (!this->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    }

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPARLinearForwardingTableGetClbck>;

    //The ibdiagpath sets the pointer in order to take all LIDs.
    //It does not have fully discovered fabric yet and it cannot check
    //if a lid exists in the fabric.
    if (is_ibdiagpath)
        clbck_data.m_data4 = &is_ibdiagpath;
    else
        clbck_data.m_data4 = NULL;

    // go over all nodes:
    struct ib_ar_linear_forwarding_table_sx ar_linear_forwarding_table;
    CLEAR_STRUCT(ar_linear_forwarding_table);

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    for (u_int8_t pLFT = 0 ; pLFT < MAX_PLFT_NUM; ++pLFT ) {
        direct_route_list_iter iter = directRouteList.begin();
        for ( ; iter != directRouteList.end() ; ++iter) {

            p_curr_node = iter->first;
            p_curr_direct_route = iter->second;

            if (p_curr_node->getMaxPLFT() < pLFT)
                continue;

            if (pLFT == 0)
                p_curr_node->appData1.val = 0;

            u_int16_t lfDBTop = p_curr_node->getLFDBTop(pLFT);

            p_curr_node->resizeLFT((u_int16_t)(lfDBTop+1), pLFT);
            p_curr_node->resizeARLFT((u_int16_t)(lfDBTop+1), pLFT);

            u_int16_t num_of_lid_to_port_blocks =
                (u_int16_t)((lfDBTop + IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX) /
                                  IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX);
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "Switch=%s pLFT:%d has LinearFDBTop=%u ==> Blocks=%u\n",
                       p_curr_node->getName().c_str(), (int)pLFT, lfDBTop,
                       num_of_lid_to_port_blocks);

            vector<bool> blocks = vector<bool>(num_of_lid_to_port_blocks, false);
            this->MarkBlocksToSend(blocks, lids, IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX);
            for (u_int16_t curr_block = 0; curr_block < num_of_lid_to_port_blocks; ++curr_block) {
                if (!blocks[curr_block])
                    continue;

                clbck_data.m_data1 = p_curr_node;
                clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
                clbck_data.m_data3 = (void *)(uintptr_t)pLFT;

                this->ibis_obj.SMPARLinearForwardingTableGetSetByDirect(
                    p_curr_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    curr_block,
                    pLFT,
                    &ar_linear_forwarding_table, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
                if (p_curr_node->appData1.val != 0)
                    break; //goto next_node;
            }

            if (ibDiagClbck.GetState())
                break;
        }

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    //else if (!retrieve_errors.empty())
    //    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//////
// RN
//////

int IBDiag::AddRNDataMapEntry(AdditionalRoutingDataMap *p_routing_data_map,
                              IBNode *p_node,
                              direct_route_t *p_direct_route,
                              adaptive_routing_info *p_ar_info)
{
    pair <IBNode *, AdditionalRoutingData> entry;
    entry.first = p_node;
    entry.second.p_node = p_node;
    entry.second.p_direct_route = p_direct_route;

    memcpy(&(entry.second.ar_info), p_ar_info, sizeof(*p_ar_info));

    AdditionalRoutingDataMapInsertRes res = p_routing_data_map->insert(entry);
    if (!res.second)
        //If an equivalent key already existed. Should not happen
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    AdditionalRoutingData & routing_data = res.first->second;

    //appData3.ptr is used by ARLinearForwardingTableGetClbck,
    //to update AdditionalRoutingData.ar_lft_table_vec
    //and by IBDiagSMPARGroupTableGetClbck
    //to update AdditionalRoutingData.group_table_vec
    p_node->appData3.ptr = &routing_data;

    if (p_ar_info->is_arn_sup || p_ar_info->is_frn_sup) {

        routing_data.sw_supports_rn_count++;

        //RNSubGroupDirectionTable
        routing_data.top_sub_group_direction_block =
                (u_int16_t) (((p_ar_info->group_top + 1) *
                        (p_ar_info->sub_grps_active + 1) - 1) /
                        IBIS_IB_MAD_SMP_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE);

        routing_data.sub_group_direction_table_vec.resize(
                routing_data.top_sub_group_direction_block + 1);

        AdditionalRoutingData::max_sub_group_direction_block = max(
                AdditionalRoutingData::max_sub_group_direction_block,
                routing_data.top_sub_group_direction_block);

        //RNGenStringTable
        routing_data.top_direction_block =
                (u_int8_t)(p_ar_info->direction_num_sup /
                        IBIS_IB_MAD_SMP_RN_GEN_STRING_TABLE_BLOCK_SIZE);

        u_int8_t maxPLFT = p_node->getMaxPLFT();

        routing_data.gen_string_table_vec.resize(maxPLFT + 1);

        for (u_int8_t pLFT = 0 ; pLFT <= maxPLFT ; ++pLFT) {
            routing_data.gen_string_table_vec[pLFT].resize(
                    routing_data.top_direction_block + 1);
        }

        AdditionalRoutingData::max_direction_block =
                max(AdditionalRoutingData::max_direction_block, routing_data.top_direction_block);

        AdditionalRoutingData::max_plft = max(AdditionalRoutingData::max_plft, maxPLFT);

        //RNRcvString
        routing_data.top_string_block =
                (u_int16_t)((0x1 << (p_ar_info->string_width_cap)) - 1)/
                IBIS_IB_MAD_SMP_RN_RCV_STRING_BLOCK_SIZE;

        routing_data.rcv_string_vec.resize(
                routing_data.top_string_block + 1);

        AdditionalRoutingData::max_string_block =
                max(AdditionalRoutingData::max_string_block, routing_data.top_string_block);

        //RNXmitPortMask

        if (p_ar_info->rn_xmit_enabled) {
            routing_data.top_ports_block =
                (u_int8_t)(p_node->numPorts / IBIS_IB_MAD_SMP_RN_XMIT_PORT_MASK_BLOCK_SIZE);
            routing_data.xmit_port_mask_vec.resize(routing_data.top_ports_block + 1);

            AdditionalRoutingData::max_ports_block =
                max(AdditionalRoutingData::max_ports_block, routing_data.top_ports_block);
        }
    }

    routing_data.top_group_table_block = 0;
    routing_data.top_ar_lft_table_block = 0;

    if (p_ar_info->group_top){
        u_int16_t top_block =
                (u_int16_t)(((p_ar_info->group_top + 1) *
                        (p_ar_info->sub_grps_active + 1)) /
                        IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE);

        routing_data.group_table_vec.resize(top_block + 1);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveRNData(list_p_fabric_general_err& retrieve_errors,
                           AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    AdditionalRoutingDataMapIter routing_data_iter = p_routing_data_map->begin();
    AdditionalRoutingData *p_routing_data = NULL;

    //RNSubGroupDirectionTable
    struct rn_sub_group_direction_tbl sub_group_direction_table;
    CLEAR_STRUCT(sub_group_direction_table);

    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRNSubGroupDirectionTableGetClbck>;

    for (u_int16_t block_num = 0 ;
         block_num <= AdditionalRoutingData::max_sub_group_direction_block;
         ++block_num) {

        clbck_data.m_data2 = (void *)(uintptr_t)block_num;

        for (routing_data_iter = p_routing_data_map->begin();
             routing_data_iter != p_routing_data_map->end();
             ++routing_data_iter){

            p_routing_data = &(routing_data_iter->second);

            if (!p_routing_data->isRNSupported())
                continue; //continue to next sw

            if (p_routing_data->top_sub_group_direction_block < block_num)
                continue; //continue to next sw

            clbck_data.m_data1 = p_routing_data;

            this->ibis_obj.SMPRNSubGroupDirectionTableGetSetByDirect(
                p_routing_data->p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                &sub_group_direction_table, &clbck_data);

            if (ibDiagClbck.GetState())
                break;
        }

        if (ibDiagClbck.GetState())
            break;
    }

    //RNGenStringTable
    struct rn_gen_string_tbl gen_string_table;
    CLEAR_STRUCT(gen_string_table);

    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRNGenStringTableGetClbck>;

    for (u_int8_t pLFT = 0 ; pLFT <= AdditionalRoutingData::max_plft;
         ++pLFT ) {

        clbck_data.m_data3 = (void *)(uintptr_t)pLFT;

        for (u_int8_t block_num = 0 ;
             block_num <= AdditionalRoutingData::max_direction_block;
             ++block_num) {

            clbck_data.m_data2 = (void *)(uintptr_t)block_num;

            for (routing_data_iter = p_routing_data_map->begin();
                 routing_data_iter != p_routing_data_map->end();
                 ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->top_direction_block < block_num ||
                    p_routing_data->p_node->getMaxPLFT() < pLFT)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.SMPRNGenStringTableGetSetByDirect(
                    p_routing_data->p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    block_num,
                    pLFT,
                    &gen_string_table, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
            }//routing_data_iter
        } //block_num
    }//pLFT

    //RNGenBySubGroupPriority
    struct rn_gen_by_sub_group_prio gen_by_sg_priority;
    CLEAR_STRUCT(gen_by_sg_priority);

    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRNGenBySubGroupPriorityGetClbck>;

    for (routing_data_iter = p_routing_data_map->begin();
         routing_data_iter != p_routing_data_map->end();
         ++routing_data_iter){

        p_routing_data = &(routing_data_iter->second);

        if (!p_routing_data->isRNSupported())
            continue; //continue to next sw

        clbck_data.m_data1 = p_routing_data;

        this->ibis_obj.SMPRNGenBySubGroupPriorityMadGetSetByDirect(
            p_routing_data->p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            &gen_by_sg_priority, &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }


    //RNRcvString
    struct rn_rcv_string rcv_string;
    CLEAR_STRUCT(rcv_string);

    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRNRcvStringGetClbck>;

    for (u_int16_t block_num = 0 ;
         block_num <= AdditionalRoutingData::max_string_block;
         ++block_num) {

        clbck_data.m_data2 = (void *)(uintptr_t)block_num;

        for (routing_data_iter = p_routing_data_map->begin() ;
             routing_data_iter != p_routing_data_map->end() ;
             ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->top_string_block < block_num)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.SMPRNRcvStringGetSetByDirect(
                    p_routing_data->p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    block_num,//string_block,
                    &rcv_string, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
        }
    }

    //RNXmitPortMask
    struct rn_xmit_port_mask xmit_port_mask;
    CLEAR_STRUCT(xmit_port_mask);

    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRNXmitPortMaskGetClbck>;

    for (u_int8_t block_num = 0 ;
         block_num <= AdditionalRoutingData::max_ports_block;
         ++block_num) {

        clbck_data.m_data2 = (void *)(uintptr_t)block_num;

        for (routing_data_iter = p_routing_data_map->begin();
             routing_data_iter != p_routing_data_map->end();
             ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported() || !p_routing_data->ar_info.rn_xmit_enabled)
                    continue; //continue to next sw

                if (p_routing_data->top_ports_block < block_num)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.SMPRNXmitPortMaskGetSetByDirect(
                    p_routing_data->p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    block_num, //ports_block,
                    &xmit_port_mask, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
        }
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

int IBDiag::CountSkipRoutingChecksNodes(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "CountSkipRoutingChecksNodes Start \n");
    SubnMgtCountSkipRoutingChecksNodes(&(this->discovered_fabric));
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "CountSkipRoutingChecksNodes End \n");

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    IBDIAG_RETURN(rc);
}

int IBDiag::CheckSL2VLTables(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    //CA <--> CA connectivity
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "CheckSL2VLTables Start \n");
    SubnMgtCheckSL2VLTables(&(this->discovered_fabric));
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "CheckSL2VLTables End \n");

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_IBDM_ERR / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::ReportFabricQualities(string& output, const char *outDir, bool ar_enabled, bool static_ca2ca)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    if (!ar_enabled || static_ca2ca) {
        //CA <--> CA histogram
        SubnMgtVerifyAllCaToCaRoutes(&(this->discovered_fabric), outDir);
    } else {
        cout << "-I- AR enabled, skipping verify all CA to CA for static routing" << endl;
    }

    SubnMgtVerifyFNMPorts(&(this->discovered_fabric), ar_enabled);

    //LID --> Port histogram
    SubnMgtCheckFabricMCGrps(&(this->discovered_fabric));
    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_IBDM_ERR / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::ReportCreditLoops(string& output, bool is_fat_tree, bool checkAR)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    list_pnode root_nodes;
    char *buffer;
    if (is_fat_tree) {
        // Calculate the minhop table for the switches
        if (SubnMgtCalcMinHopTables(&(this->discovered_fabric))) {
        	cout << "-E- Report Credit Loop failure. Fail to update Min Hops Tables." << endl;
        	this->SetLastError("Report Credit Loop failure. Fail to update Min Hops Tables.");
        	IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
        }

        // Analyze the fabric to find its root nodes using statistical methods
        // on the profiles of min hops to CAs
        root_nodes = SubnMgtFindRootNodesByMinHop(&(this->discovered_fabric));

        buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            this->SetLastError("Failed to allocate buffer for ibdm output");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
        }
        output += buffer;
        free(buffer);
    }

    if (root_nodes.empty() == false) {      //fat tree
        char buffer2[128];
        snprintf(buffer2, sizeof(buffer2), "\n-I- Found %u Roots:\n", (unsigned int)root_nodes.size());
        output += buffer2;
        for (list_p_node::iterator it = root_nodes.begin(); it != root_nodes.end(); ++it) {
            output += "    ";
            output += (*it)->getName();
            output += "\n";
        }
        this->ReportNonUpDownCa2CaPaths(&(this->discovered_fabric), root_nodes, output);
    } else {                                //generic - not fat tree
        CrdLoopAnalyze(&(this->discovered_fabric), checkAR);
    }

    buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}

/****************** AR Routing Tests **********************************/

//Returns: SUCCESS_CODE / ERR_CODE_IBDM_ERR / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::ReportFabricARConnectivity(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    //CA <--> CA connectivity
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtVerifyAllARCaToCaRoutes Start \n");
    SubnMgtVerifyAllARCaToCaRoutes(&(this->discovered_fabric));
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtVerifyAllARCaToCaRoutes End \n");

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}

int IBDiag::ReportFabricAREmptyGroups(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    //AR empty groups check
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtVerifyAREmptyGroups Start \n");
    SubnMgtVerifyAREmptyGroups(&(this->discovered_fabric));
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtVerifyAREmptyGroups End \n");

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}

int IBDiag::ReportFabricARValidation(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    if (!this->discovered_fabric.IsSMDBApplied()) {
        cout << "-I- SMDB file wasn't applied, skipping AR validation." << endl;
    } else {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtValidateARRouting Start \n");
        SubnMgtValidateARRouting(&(this->discovered_fabric));
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtValidateARRouting End \n");
    }

    cout << "---------------------------------------------------------------------------" << endl;
    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    output += buffer;
    free(buffer);

    IBDIAG_RETURN(rc);
}

/****************************************************/

int IBDiag::ParseSLVLFile(const string & file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parseSLVLFile(file_name);

    char *buffer= ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteSLVLFile(const string &file_name,
                          list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("SLVL Table",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    // add format explanation
    sout << "# File format explanation:" << endl
         << "#   Field 1 - Node GUID " << endl
         << "#   Field 2 - in port" << endl
         << "#   Field 3 - out port" << endl
         << "#   Other fields - map from SL to VL" << endl
         << "#     every hex digit is the VL for the current SL " << endl
         << "#     for example:" << endl
         << "#     0x43 0x21 0x01 0x23 0x01 0x23 0x01 0x23" << endl
         << "#     SL0 to VL4, SL1 to VL3, SL3 to VL2, SL4 to VL1 and so on " << endl
         << endl << endl;

    this->DumpSLVLFile(sout, retrieve_errors);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

#define IS_SUPPORT_SL_MAPPING(cap_mask) ((cap_mask) & 0x40)

bool IBDiag::HandleUnsupportedSLMapping(ofstream &sout,
                                        IBNode *p_node,
                                        phys_port_t port)
{
    IBPort *p_port = NULL;

    if (p_node->type == IB_SW_NODE)
        p_port = p_node->getPort(0);
    else
        p_port = p_node->getPort(port);

    if (!p_port)
        return true;

    struct SMP_PortInfo *p_port_info =
        this->fabric_extended_info.getSMPPortInfo(p_port->createIndex);

    if (!p_port_info)
        return true;

    if (IS_SUPPORT_SL_MAPPING(p_port_info->CapMsk))
        return false;

    char buffer[1024];
    if (p_node->type == IB_SW_NODE) {
        for (phys_port_t curr_out_port = 1 ;
             curr_out_port <= p_node->numPorts ; ++curr_out_port) {

            for (phys_port_t curr_in_port = 0 ;
                 curr_in_port <= p_node->numPorts ; ++curr_in_port) {

                if(curr_out_port == curr_in_port)
                    continue;

                snprintf(buffer, sizeof(buffer), U64H_FMT " %u %u"
                    " 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00\n",
                    p_node->guid_get(), curr_in_port, curr_out_port);

                for (uint8_t i = 0; i < IB_NUM_SL; ++i)
                    p_node->setSLVL(curr_in_port, curr_out_port, i, 0);

                sout << buffer;

            }

        }
    } else {
        snprintf(buffer, sizeof(buffer), U64H_FMT " 0 %u"
            " 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00\n",
            p_node->guid_get(),
            port);

        for (uint8_t i = 0; i < IB_NUM_SL; ++i)
            p_node->setSLVL(0, port, i, 0);

        sout << buffer;
    }

    return true;
}

int IBDiag::ReadCASLVL(ofstream &sout,
                       clbck_data_t &clbck_data,
                       SMP_SLToVLMappingTable &slvl_mapping,
                       IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    IBPort *p_port=NULL;

    for (phys_port_t port = 1 ; port <= p_node->numPorts ; port++) {
        p_port = p_node->getPort(port);
        if (p_port == NULL || p_port->p_remotePort == NULL)
            continue;

        if (HandleUnsupportedSLMapping(sout, p_node, port))
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_port);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to port=%s", p_port->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_node;
        clbck_data.m_data2 = (void *)(uintptr_t)0;
        clbck_data.m_data3 = (void *)(uintptr_t)port;

        this->ibis_obj.SMPSLToVLMappingTableGetByDirect(p_curr_direct_route,
                (phys_port_t)0, (phys_port_t)0, &slvl_mapping, &clbck_data);
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpSLVLFile(ofstream &sout,
                         list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    struct SMP_SLToVLMappingTable slvl_mapping;
    CLEAR_STRUCT(slvl_mapping);

    this->GetDiscoverFabricPtr()->numVLs = 1;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors, &sout);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPSLToVLMappingTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    u_int32_t nodesNumber = this->fabric_extended_info.getNodesVectorSize();

    for (u_int32_t i = 0; i < nodesNumber; ++i) {

        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_curr_node->appData1.val = 0;

        if (p_curr_node->type != IB_SW_NODE) {
            rc = ReadCASLVL(sout, clbck_data, slvl_mapping, p_curr_node);
            if (rc || ibDiagClbck.GetState())
                goto exit;
            continue;
        }

        if (HandleUnsupportedSLMapping(sout, p_curr_node, 0))
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        //char sLine[1024];
        // use unsigned int index to prevent overwrap
        for (unsigned int curr_out_port = 1 ;
             curr_out_port <= p_curr_node->numPorts ; ++curr_out_port) {

            for (unsigned int curr_in_port = 0 ;
                 curr_in_port <= p_curr_node->numPorts ; ++curr_in_port) {

                if(curr_out_port == curr_in_port)
                    continue;

                clbck_data.m_data1 = p_curr_node;
                clbck_data.m_data2 = (void *)(uintptr_t)curr_in_port;
                clbck_data.m_data3 = (void *)(uintptr_t)curr_out_port;

                progress_bar.push(p_curr_node);
                this->ibis_obj.SMPSLToVLMappingTableGetByDirect(p_curr_direct_route,
                        (phys_port_t)curr_out_port, (phys_port_t)curr_in_port, &slvl_mapping, &clbck_data);
            }

            if (ibDiagClbck.GetState())
                goto exit;
        }

    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WritePLFTFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("PLFT",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpPLFTInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpPLFTInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char buffer[2096];
    IBNode *p_curr_node = NULL;
    IBPort *p_port = NULL;

    sout << "File version: 1" << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            !p_curr_node->isPLFTEnabled())
            continue;

        snprintf(buffer, sizeof(buffer), "dump_plft: Switch " U64H_FMT ,
                 p_curr_node->guid_get());
        sout << buffer << endl;


        for (phys_port_t port = 0; port <= p_curr_node->numPorts ; ++port) {


            if (port !=0 ) {
                p_port = p_curr_node->getPort(port);
                //Skip ports that don't exist or are down or have undefined
                //state or are not in the scope of the fabric.
                if ( !p_port || !p_port->is_data_worthy())
                    continue;
            }

            p_curr_node->getPLFTMapping(port, buffer);
            sout << "rq: " << (int)port << " sl-plfft: "  << buffer <<endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteARFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("AR",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpARInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpARInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    sout << "File version: 2" << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            !p_curr_node->isAREnable())
            continue;

        stringstream ss;
        ss << "dump_ar: Switch " << PTR(p_curr_node->guid_get())
           << " fr_en: " << dec << (p_curr_node->isFREnabled() ? '1' : '0');

        ss << " en_sl: ";
        p_curr_node->getARActiveCfg(ss);
        ss << endl << "Groups Definition:" << endl;

        u_int16_t maxARGroupNumber = p_curr_node->getMaxARGroupNumber();

        for (u_int16_t groupNumber = 0;
              groupNumber <= maxARGroupNumber ; ++groupNumber)
        {
            ss << DEC(groupNumber) << ": ";
            p_curr_node->getARGroupCfg(groupNumber, ss);
            ss << endl;
        }

        ss << "LFT Definition:" << endl;
        u_int8_t maxPLFT =  p_curr_node->getMaxPLFT();
        for (u_int8_t pLFT = 0; pLFT <= maxPLFT ; ++pLFT) {
            ss << "PLFT_NUM: " << DEC(pLFT) << endl;
            ss << "LID    : static port : AR group" << endl;

            u_int16_t lfdbTop = p_curr_node->getLFDBTop(pLFT);

            for (lid_t dlid = 1; dlid <= lfdbTop; ++dlid) {
                u_int8_t static_port = p_curr_node->getLFTPortForLid(dlid, pLFT);
                if (static_port == IB_LFT_UNASSIGNED)
                    ss << PTR(dlid) << " :  UNREACHABLE  :  UNASSIGNED" << dec;
                else {
                    u_int16_t ports_group = p_curr_node->
                        getARLFTPortGroupForLid(dlid, pLFT);

                    if (ports_group == IB_AR_LFT_UNASSIGNED)
                        ss << PTR(dlid) << " :  " << DEC(static_port, 3, '0')
                           << "  :  UNASSIGNED" << dec;
                    else
                        ss << PTR(dlid) << " :  " << DEC(static_port, 3, '0')
                           << "  :  " <<  DEC(ports_group, 3, '0');
                }
                ss << endl;
            }
            ss << endl;
        }
        sout << ss.rdbuf() << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::WriteFARFile(AdditionalRoutingDataMap *p_routing_data_map,
                         const string &file_name,
                         bool isFLID)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (isFLID &&
            !this->discovered_fabric.globalMinFLID &&
            !this->discovered_fabric.globalMaxFLID)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    ofstream sout;
    int rc = this->OpenFile(isFLID ? "FLID and AR" : "Full AR",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpFARInfo(p_routing_data_map, sout, isFLID);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

void IBDiag::DumpPortsBitset(u_int64_t pgSubBlockElement,
                             phys_port_t portOffset,
                             ostream &sout)
{
    for (phys_port_t port = 0 ; port < 64 ; ++port) {
        if ( pgSubBlockElement & ((u_int64_t)1 << port) )
            sout << (int)(port+portOffset) << ",";
    }
}

void IBDiag::DumpPortgroup(ib_portgroup_block_element &pgBlockElement,
                           ostream &sout)
{
    if (pgBlockElement.SubGroup_0)
        DumpPortsBitset(pgBlockElement.SubGroup_0,
                      (phys_port_t)0, sout);
    if (pgBlockElement.SubGroup_1)
        DumpPortsBitset(pgBlockElement.SubGroup_1,
                     (phys_port_t)(1*64), sout);
    if (pgBlockElement.SubGroup_2)
        DumpPortsBitset(pgBlockElement.SubGroup_2,
                     (phys_port_t)(2*64), sout);
    if (pgBlockElement.SubGroup_3)
        DumpPortsBitset(pgBlockElement.SubGroup_3,
                     (phys_port_t)(3*64), sout);
}

int IBDiag::DumpFARInfo(AdditionalRoutingDataMap *p_routing_data_map,
                        ofstream &sout, bool isFLID)
{
    IBDIAG_ENTER;
    char buffer[2096];

    sout << "File version: " << (isFLID ? 1 : 4) << endl;

    for (AdditionalRoutingDataMap::iterator rnI = p_routing_data_map->begin();
            rnI != p_routing_data_map->end(); ++rnI) {
        AdditionalRoutingData *p_routing_data = &rnI->second;

        IBNode *p_curr_node = p_routing_data->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in AdditionalRoutingDataMap");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            (!p_curr_node->isAREnable() && !p_curr_node->isHBFEnable()))
            continue;

        stringstream ss;
        //Switch <GUID> ar_en: 1 fr_en: 1 rn_en: 1
        ss << "Switch "  << PTR(p_curr_node->guid_get())
           << " ar_en:"  << '1'
           << " fr_en: " << (p_curr_node->isFREnabled() ? '1' : '0')
           << " rn_en: " << (p_curr_node->isRnXmitEnabled() ? '1' : '0')
           << " whbf_en: " << (p_curr_node->isWeightsHBFEnabled() ? '1' : '0')
           << " pfrn_en: " << (p_curr_node->ispFRNEnabled() ? '1' : '0')
           << endl;

        //#switch-name=<name>
        ss << "#switch-name=" << p_curr_node->getName() << endl;

        //ar_en_sl: (0x0)
        ss << "ar_en_sl: " << R_BRACKETS(PTR(p_curr_node->getAREnabledSLMask(), 0));
        //hbf_sl_en: (0x0)
        ss << " hbf_sl: " << R_BRACKETS(PTR(p_curr_node->getHBFEnabledSLMask(), 0))
           << endl;

        //#HBF configuration: hash_type=XOR/CRC seed_type=Random/Configurable
        //seed_value=0x<value>/Reserved fields_enable=0x<value>
        ss << "#HBF configuration: ";
        const hbf_config *p_hbf = this->GetIBDMExtendedInfoPtr()->getHBFConfig(p_curr_node->createIndex);
        if (p_hbf)
            ss  << "hash_type="     << (p_hbf->hash_type ? "XOR" : "CRC")       << ' '
                << "seed_type="     << (p_hbf->seed_type ? "Random" : "Config") << ' '
                << "seed_value="    << PTR(p_hbf->seed)                         << ' '
                << "fields_enable=" << PTR(p_hbf->fields_enable)                << endl;
        else
            ss << "N/A" << endl;

        //ByTransportDisable: (0x5). UD, UC,
        int transpDisable = (int)p_curr_node->getTranspDisable();
        ss << "ByTransportDisable: (" << PTR(transpDisable,0) << "). ";
        if (transpDisable & 1)
            ss << "UD, ";
        if (transpDisable & 2)
            ss << "RC, ";
        if (transpDisable & 4)
            ss << "UC, ";
        if (transpDisable & 8)
            ss << "DCT";
        ss << endl << endl;

        ss << "Groups Definition:" << endl << endl;
        u_int16_t maxARGroupNumber = p_curr_node->getMaxARGroupNumber();
        u_int16_t groupNumber = 0;
        u_int16_t subGroupNumber = 0;

        ss << setfill(' ')
            << std::setw(10) << std::left << "Group"
            << std::setw(15) << std::left << "Sub Group"
            << std::setw(10) << std::left << "WHBF"
            << "Ports" << endl;
        ss << "--------------------------------------------------------------------------" << endl;

        //top_group_table_block might be 0 if no data was collected
        u_int16_t total_bloks =
            min((u_int16_t)(p_routing_data->top_group_table_block + 1),
                (u_int16_t) p_routing_data->group_table_vec.size());

        if (!total_bloks)
            ss << std::left << DEC(0, 10) << std::left << DEC(0, 15) << endl;
        else
            for (u_int16_t blockNumber = 0; blockNumber < total_bloks; ++blockNumber) {
                 for (u_int8_t blockEntry = 0; blockEntry < IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE;
                        blockEntry++) {

                      ss << std::left << DEC(groupNumber,10)
                         << std::left << DEC(subGroupNumber, 15);

                      //dump WHBF subgroup weight
                      if (!p_routing_data->ar_info.is_hbf_supported ||
                              !p_routing_data->ar_info.is_whbf_supported) {
                          ss << std::setw(10) << std::left << "N/A";
                      } else if (!p_routing_data->ar_info.whbf_en) {
                          ss << std::left << DEC(0, 10);
                      } else {
                          int weight = p_routing_data->GetSubGroupWeight(groupNumber, subGroupNumber);
                          if (weight > -1)
                              ss << std::left << DEC(weight, 10);
                          else
                              ss << std::setw(10) << std::left << "N/A";
                      }

                      DumpPortgroup(p_routing_data->group_table_vec[blockNumber].
                                    Group[blockEntry], ss);

                      ss << endl;

                      subGroupNumber++;
                      if (subGroupNumber > p_routing_data->ar_info.sub_grps_active) {
                          subGroupNumber = 0;
                          groupNumber++;
                      }
                      if (groupNumber > maxARGroupNumber)
                          break;
                 }
                 if (groupNumber > maxARGroupNumber)
                     break;
            }

        ss << endl <<"LFT Definition:" << endl;
        u_int8_t maxPLFT =  p_curr_node->getMaxPLFT();
        for (u_int8_t pLFT = 0; pLFT <= maxPLFT ; ++pLFT) {
            ss << std::right
               << endl << "PLFT_NUM: " << DEC(pLFT)
               << ", top: " << PTR(p_curr_node->getLFDBTop(pLFT), 0) << endl << endl
               << setfill(' ')
               << std::setw(10) << std::left << "LID"
               << std::setw(15) << std::left << "Static Port"
               << std::setw(15) << std::left << "Lid State"
               << "Group" << endl
               << "-----------------------------------------------" << endl;

            total_bloks =
                min((u_int16_t)(p_routing_data->top_ar_lft_table_block + 1),
                    (u_int16_t) p_routing_data->ar_lft_table_vec[pLFT].size());

            lid_t dlid = 0;
            for (u_int16_t blockNumber = 0;
                  blockNumber < total_bloks;
                  ++blockNumber) {
                for (u_int8_t blockEntry = 0;
                     blockEntry < IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX ;
                     blockEntry++) {

                    ib_ar_lft_block_element_sx &lft_entry =
                        p_routing_data->ar_lft_table_vec[pLFT][blockNumber].
                        LidEntry[blockEntry];

                    //skip local FLIDs for FLID_FAR file
                    if (isFLID && !this->discovered_fabric.isFLID(dlid)) {
                        dlid++;
                        continue;
                    }

                    if (lft_entry.DefaultPort != IB_LFT_UNASSIGNED) {
                        snprintf(buffer, sizeof(buffer), U16H_FMT, dlid);
                        ss << std::setw(10) << std::left << buffer
                           << std::setw(15) << std::left << DEC(lft_entry.DefaultPort);

                        ss << std::setw(15) << std::left;
                        switch(lft_entry.LidState) {
                            case AR_IB_LID_STATE_BOUNDED:
                                ss << "Bounded";
                                break;
                            case AR_IB_LID_STATE_FREE:
                                ss << "Free";
                                break;
                            case AR_IB_LID_STATE_STATIC:
                                ss << "Static";
                                break;
                            case AR_IB_LID_STATE_HBF:
                                ss << "HBF";
                                break;
                            default:
                                ss << "UNKNOWN";
                                break;
                        }

                        if (lft_entry.LidState == AR_IB_LID_STATE_STATIC &&
                            !p_routing_data->ar_info.fr_enabled)
                            ss << '0' << endl;
                        else
                            ss << DEC(lft_entry.GroupNumber) << endl;
                    }

                    dlid++;
                    if (dlid > p_curr_node->getLFDBTop(pLFT))
                        break;
                }
                if(dlid > p_curr_node->getLFDBTop(pLFT))
                    break;
            }
            ss << endl;
        }// pLFT
        sout << ss.rdbuf() << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteRNFile(list_p_fabric_general_err& errors,
                        AdditionalRoutingDataMap *p_routing_data_map,
                        const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("RN",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpRNInfo(errors, p_routing_data_map, sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

const char *IBDiag::RNDecisionToStr(u_int8_t decision){
    switch (decision) {
        case 0:
            return "Discard";
        case 1:
            return "Consume ARN";
        case 2:
            return "Consume ARN/FRN";
        case 3:
            return "Pass-on";
    }

    return "Invalid";
}

#define RN_XMIT_PORT_MASK_GENERATE_ARN 0X1
#define RN_XMIT_PORT_MASK_GENERATE_FRN 0X2
#define RN_XMIT_PORT_MASK_PASS_ON_RN   0X4

#define TAB_DEF std::setw(20) << std::left

int IBDiag::DumpRNInfo(list_p_fabric_general_err& errors,
                       AdditionalRoutingDataMap *p_routing_data_map,
                       ofstream &sout)
{
    IBDIAG_ENTER;
    char buffer[2096];

    set <u_int8_t> directionSet;
    set <u_int8_t>::iterator directionSetIter;
    set <u_int16_t> stringSet;

    u_int16_t maxSubGroups = (u_int16_t)
        (((AdditionalRoutingData::max_sub_group_direction_block + 1) *
         IBIS_IB_MAD_SMP_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE) + 1);

    // This field is calculated from ARInfo.DirectionNumSupported
    u_int8_t maxDirections = (u_int8_t)
        ((AdditionalRoutingData::max_direction_block + 1) *
         IBIS_IB_MAD_SMP_RN_GEN_STRING_TABLE_BLOCK_SIZE);

    // This field is calculated from ARInfo.StringWidthCap
    // it is relevant for num of valid bits in RNGenStringTable element
    // and for num of entries in RNRcvString
    u_int16_t maxStrings = (u_int16_t)
        ((AdditionalRoutingData::max_string_block + 1) *
         IBIS_IB_MAD_SMP_RN_RCV_STRING_BLOCK_SIZE);

    vector <u_int16_t> groupsHistogram(maxSubGroups, 0);
    vector <u_int16_t> sgHistogram(maxSubGroups, 0);
    vector <u_int16_t> directionHistogram(maxDirections, 0);
    vector <u_int16_t> stringHistogram(maxStrings, 0);

    sout << "File version: 1" << endl;

    for (AdditionalRoutingDataMap::iterator rnI = p_routing_data_map->begin();
            rnI != p_routing_data_map->end(); ++rnI) {
        AdditionalRoutingData *p_routing_data = &rnI->second;

        if (!p_routing_data->isRNSupported())
            continue; //continue to next sw

        directionSet.clear();
        stringSet.clear();

        snprintf(buffer, sizeof(buffer), "\n\ndump_rn: Switch " U64H_FMT ,
                 p_routing_data->p_node->guid_get());
        sout << buffer ;

        sout << "\n\nRN Sub Group Direction Table:" << endl;
        sout <<     "#============================\n" << endl;

        sout << TAB_DEF << "Group" <<
                TAB_DEF << "Sub Group" <<
                "Direction\n";
        sout << "#------------------------------------------------\n";

        u_int16_t blockSize =
            IBIS_IB_MAD_SMP_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE;

        unsigned group_num,sub_group_num;
        u_int16_t total_sg_num = 0;


        for (u_int16_t block = 0;
             block <= p_routing_data->top_sub_group_direction_block ; ++block) {

            for (unsigned blockEntry = 0;
                 blockEntry < blockSize; ++blockEntry) {

                group_num =  (blockEntry + block * blockSize) /
                    (p_routing_data->ar_info.sub_grps_active+1);
                sub_group_num = (blockEntry + block * blockSize) %
                    (p_routing_data->ar_info.sub_grps_active+1);

                if (group_num > p_routing_data->ar_info.group_top)
                    break;

                u_int8_t direction =
                    p_routing_data->sub_group_direction_table_vec[block].
                    direction_of_sub_group[blockEntry];

                sout << TAB_DEF << group_num <<
                        TAB_DEF << sub_group_num <<
                        (int)direction << endl;

                total_sg_num++;

                if (direction >= directionHistogram.size()) {

                    stringstream sstr;
                    sstr << "Invalid direction = " << +direction
                         << ". maxDirections = " << maxDirections;

                    errors.push_back(new FabricErrNodeWrongConfig(p_routing_data->p_node, sstr.str()));

                } else {

                    directionSet.insert(direction);
                    directionHistogram[direction]++;
                }
            }
        }

        if (total_sg_num < sgHistogram.size()) {
            sgHistogram[total_sg_num]++;

        } else {
            stringstream sstr;
            sstr << "total sub-group number ("
                 << total_sg_num
                 << ") doesn't fit in sub-group Histogram size ("
                 << sgHistogram.size()
                 << ")";

            errors.push_back(new FabricErrNodeWrongConfig(p_routing_data->p_node, sstr.str()));
        }

        sout << "\n\nRN Gen String Table:" << endl;
        sout <<     "#===================" << endl;

        blockSize =
            IBIS_IB_MAD_SMP_RN_GEN_STRING_TABLE_BLOCK_SIZE;

        unsigned direction;

        for (u_int8_t pLFT = 0; pLFT <= p_routing_data->p_node->getMaxPLFT(); ++pLFT) {

            sout << "\nPLFT_NUM: " << (int)pLFT << endl;
            sout << TAB_DEF << "Direction" <<
                    "Generated String\n";
            sout << "#-----------------------------------\n";

            for (directionSetIter = directionSet.begin();
                    directionSetIter != directionSet.end();
                    ++directionSetIter) {

                direction = *directionSetIter;

                u_int16_t block = (u_int16_t)(direction/blockSize);
                u_int16_t blockEntry = (u_int16_t)(direction%blockSize);

                u_int16_t genString =
                    p_routing_data->gen_string_table_vec[pLFT][block].
                    element[blockEntry].rn_gen_string;

                sout << TAB_DEF << direction <<
                    TAB_DEF << (int)genString << endl;

                if (genString >= stringHistogram.size()) {

                    stringstream sstr;
                    sstr << "Invalid genString = " << genString
                         << ". direction = " << direction
                         << ". pLFT = " << +pLFT
                         << ". maxStrings = " << maxStrings;

                    errors.push_back(new FabricErrNodeWrongConfig(p_routing_data->p_node, sstr.str()));

                } else {

                    stringSet.insert(genString);
                    stringHistogram[genString]++;
                }
            }
        }

        sout << "\n\nRN Gen By Sub Group Priority:" << endl;
        sout <<     "#============================\n" << endl;

        sout << TAB_DEF << "SG Priority" <<
            TAB_DEF << "gen_arn" << "gen_frn\n";
        sout << "#----------------------------------------------\n";

        for (unsigned sgPriority = 0;
             sgPriority <= p_routing_data->ar_info.sub_grps_active;
             ++sgPriority) {

            sout << TAB_DEF << sgPriority << TAB_DEF <<
                (int)p_routing_data->gen_by_sub_group_priority.
                element[sgPriority].gen_arn << TAB_DEF <<
                (int)p_routing_data->gen_by_sub_group_priority.
                element[sgPriority].gen_frn << endl;
        }

        sout << "\n\nRN Rcv String:" << endl;
        sout <<     "#=============\n" << endl;

        sout << TAB_DEF << "String" <<
            TAB_DEF << "Decision" <<
            TAB_DEF << "pLFT" <<
            "String to string\n";
        sout <<
            "#--------------------------------------------" <<
            "-------------------------------\n";

        blockSize = IBIS_IB_MAD_SMP_RN_RCV_STRING_BLOCK_SIZE;

        u_int32_t maxString =
            (0x1 << (p_routing_data->ar_info.string_width_cap)) - 1;

        unsigned stringNum;

        for (u_int16_t block = 0;
             block <= p_routing_data->top_string_block ; ++block) {

            for (unsigned blockEntry = 0;
                 blockEntry < blockSize; ++blockEntry) {

                stringNum =  (blockEntry + block * blockSize);

                if (stringNum > maxString)
                    break;

                sout << TAB_DEF << stringNum <<
                    TAB_DEF <<
                    RNDecisionToStr(p_routing_data->rcv_string_vec[block].
                                    element[blockEntry].decision) <<
                    TAB_DEF <<
                    (int)p_routing_data->rcv_string_vec[block].
                    element[blockEntry].plft_id <<
                    TAB_DEF <<
                    (int)p_routing_data->rcv_string_vec[block].
                    element[blockEntry].string2string << endl;
            }
        }

        if (p_routing_data->ar_info.rn_xmit_enabled) {
            sout << "\n\nRN Xmit Port Mask:" << endl;
            sout <<     "#=================\n" << endl;

            sout << TAB_DEF << "Port" << TAB_DEF << "Generate ARN" <<
                    TAB_DEF << "Generate FRN" << "Pass on RN\n";
            sout << "#-----------------------------------------" <<
                    "----------------------------\n";

            blockSize = IBIS_IB_MAD_SMP_RN_XMIT_PORT_MASK_BLOCK_SIZE;

            unsigned portNum;

            for (u_int16_t block = 0;
                    block <= p_routing_data->top_string_block ; ++block) {

                for (unsigned blockEntry = 0;
                        blockEntry < blockSize; ++blockEntry) {

                    portNum =  (blockEntry + block * blockSize);

                    if (portNum > p_routing_data->p_node->numPorts)
                        break;

                    u_int8_t element =
                            p_routing_data->xmit_port_mask_vec[block].element[blockEntry];

                    sout << TAB_DEF << portNum << TAB_DEF <<
                        (bool)(element & RN_XMIT_PORT_MASK_GENERATE_ARN) <<
                        TAB_DEF <<
                        (bool)(element & RN_XMIT_PORT_MASK_GENERATE_FRN) <<
                        TAB_DEF <<
                        (bool)(element & RN_XMIT_PORT_MASK_PASS_ON_RN) << endl;
                }
            }
        }

        sout << "\n\nStatistics:" << endl;
        sout <<     "#==========\n" << endl;

        //Statistics
        u_int16_t grpNum = p_routing_data->p_node->getNonEmptyARGroupNumber();
        groupsHistogram[grpNum]++;

        snprintf(buffer, sizeof(buffer), "Groups:%u    Sub Groups:%u"
                 "    Directions:%u    Strings:%u\n\n",
                 grpNum, total_sg_num,
                 (int)directionSet.size(), (int)stringSet.size());
        sout << buffer ;

        sout << "#*************************************************************************************\n";

    }
    sout << "#*************************************************************************************\n";

    sout << "\nHistograms:" << endl;
    sout <<   "#==========\n\n" << endl;

    sout << TAB_DEF << "Groups Number"<< "SW Number\n";
    sout << "#----------------------------\n";

    int size = (int)groupsHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (groupsHistogram[i])
            sout << TAB_DEF << i << TAB_DEF << (int)groupsHistogram[i] << endl;
    }

    sout << "\n" << TAB_DEF << "SG Number"<< "SW Number\n";
    sout << "#----------------------------\n";

    size = (int)sgHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (sgHistogram[i])
            sout << TAB_DEF << i << TAB_DEF << (int)sgHistogram[i] << endl;
    }

    sout << "\n" << TAB_DEF << "Direction"<<
        "SG Number\n";
    sout << "#----------------------------\n";

    size = (int)directionHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (directionHistogram[i])
            sout << TAB_DEF << i <<
            TAB_DEF << (int)directionHistogram[i] << endl;
    }

    sout << "\n" << TAB_DEF << "Gen String"<<
        "SW Number\n";
    sout << "#----------------------------\n";

    size = (int)stringHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (stringHistogram[i])
            sout << TAB_DEF << i <<
            TAB_DEF << (int)stringHistogram[i] << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteVL2VLFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("VL2VL",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpVL2VLInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpVL2VLInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_sw_line[1024];
    char curr_vl2vl_line[1024];
    IBNode *p_curr_node = NULL;

    sout << "File version: 1" << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (!p_curr_node->getInSubFabric() || p_curr_node->type == IB_CA_NODE)
            continue;

        p_curr_node->getSL2VLCfg(curr_vl2vl_line);
        if (curr_vl2vl_line[0] != '\0') {
            sprintf(curr_sw_line, "dump_vl2vl: Switch " U64H_FMT " ",
                    p_curr_node->guid_get());
            sout << curr_sw_line << curr_vl2vl_line << endl;
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int AdditionalRoutingData::GetSubGroupWeight(u_int16_t group, u_int16_t subgroup)
{
    IBDIAG_ENTER;

    if (group >= this->group_weights_vec.size())
        IBDIAG_RETURN(-1);

    IBDIAG_RETURN(this->group_weights_vec[group].get(subgroup));
}

void AdditionalRoutingData::AddSubGroupWeights(u_int8_t block_index, const whbf_config& weights_config)
{
    IBDIAG_ENTER;

    const size_t block_size = ARRAY_SIZE(whbf_config::group_weights_arr);
    this->group_weights_vec.resize((block_index + 1) * block_size);

    size_t base = block_index * block_size;
    for (size_t i = 0; i < block_size; i++)
        this->group_weights_vec[base + i].set(weights_config.group_weights_arr[i]);

    IBDIAG_RETURN_VOID;
}

int IBDiag::RetrieveWeightsHBFConfig(list_p_fabric_general_err& retrieve_errors)
{

    IBDIAG_ENTER;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPWeightsHBFConfigGetClbck>;

    whbf_config whbf = {0};
    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end() && rc == IBDIAG_SUCCESS_CODE; ++nI) {
        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //SM is able to set a wrong configuration when HBF is disabled but WHBF is enabled
        //but we will not report on this error.
        //The error: FabricErrWHBFConfiguration
        if (!p_curr_node->isHBFEnable() || !p_curr_node->isWeightsHBFEnabled())
                continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        //#TODO think of enabled
        const size_t size = ARRAY_SIZE(whbf_config::group_weights_arr);
        u_int8_t blocks_count = static_cast<u_int8_t>(p_curr_node->getARGroupTop() / size);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Swicth: %s group top: %u blocks to retrive: %u\n",
                p_curr_node->getName().c_str(), p_curr_node->getARGroupTop(), blocks_count);

        for (u_int8_t block_index = 0; block_index <= blocks_count; block_index++) {
            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = (void *)(uintptr_t)block_index;

            this->ibis_obj.SMPWHBFConfigGetSetByDirect(p_curr_direct_route,
                                                   IBIS_IB_MAD_METHOD_GET,
                                                   false,
                                                   block_index,
                                                   &whbf,
                                                   &clbck_data);

            rc = ibDiagClbck.GetState();
            if (rc) {
                this->SetLastError(ibDiagClbck.GetLastError());
                break;
            }
        }//sending blocks
    }//swicthes

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::RetrieveHBFConfig(list_p_fabric_general_err& retrieve_errors, unsigned int &supportedDev)
{
    IBDIAG_ENTER;

    supportedDev = 0;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPHBFConfigGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    hbf_config hbf = {0};
    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {
        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isHBFEnable())
            continue;

        supportedDev++;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.SMPHBFConfigGetSetByDirect(p_curr_direct_route,
                                                  IBIS_IB_MAD_METHOD_GET,
                                                  true,//until future notice we work in Global mode
                                                  0,
                                                  &hbf,
                                                  &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

     this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::Retrieve_pFRNConfig(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;
    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPpFRNConfigGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_pFRNConfig pfrn = { 0 };

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->ispFRNEnabled())
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.SMP_pFRNConfigGetSetByDirect(p_curr_direct_route,
                                                    IBIS_IB_MAD_METHOD_GET,
                                                    &pfrn,
                                                    &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

     this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::Dump_pFRNConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_PFRN_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,sl,mask_force_clear_timeout,mask_clear_timeout" << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;

        if (!p_curr_node || !p_curr_node->getInSubFabric() || !p_curr_node->ispFRNEnabled())
            continue;

        struct SMP_pFRNConfig *p_pfrn =
                this->GetIBDMExtendedInfoPtr()->getpFRNConfig(p_curr_node->createIndex);

        if (!p_pfrn)
            continue;

        sstream.str("");

        sstream << PTR(p_curr_node->guid_get())       << ","
                << +p_pfrn->sl                        << ","
                << p_pfrn->mask_force_clear_timeout   << ","
                << p_pfrn->mask_clear_timeout         << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PFRN_CONFIG);

    IBDIAG_RETURN(rc);
}

//N2NClassPortInfo
int IBDiag::Retrieve_N2NClassPortInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;
    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::N2NClassPortInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct IB_ClassPortInfo class_port_info = { 0 };

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.ClassCPortInfoGet(p_curr_node->getFirstLid(),
                                         DEFAULT_SL,
                                         &class_port_info,
                                         &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::Dump_N2NClassPortInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_N2N_CLASS_PORT_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,BaseVersion,ClassVersion,CapabilityMask,CapabilityMask2,"
            << "RespTimeValue,TrapGID,TrapTC,TrapSL,TrapFL,TrapLID,"
            << "TrapP_Key,TrapHL,TrapQP,TrapQ_Key" << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;

        if (!p_curr_node || !p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        struct IB_ClassPortInfo *p_cpi =
                this->GetIBDMExtendedInfoPtr()->getN2NClassPortInfo(p_curr_node->createIndex);

        if (!p_cpi)
            continue;

        sstream.str("");

        sstream << PTR(p_curr_node->guid_get())     << ","
                << +p_cpi->BaseVersion              << ","
                << +p_cpi->ClassVersion             << ","
                << PTR(p_cpi->CapMsk)               << ","
                << PTR(p_cpi->CapMsk2)              << ","
                << +p_cpi->RespTimeValue            << ",";

        // Redirection fields are disabled.

        // TrapGID
        sstream << "0x"
                << HEX(p_cpi->TrapGID[0])
                << HEX(p_cpi->TrapGID[1])
                << HEX(p_cpi->TrapGID[2])
                << HEX(p_cpi->TrapGID[3])
                << ",";

        sstream << +p_cpi->TrapTC                   << ","
                << +p_cpi->TrapSL                   << ","
                << +p_cpi->TrapFL                   << ","
                << +p_cpi->TrapLID                  << ","
                << +p_cpi->TrapPKey                 << ","
                << +p_cpi->TrapHL                   << ","
                << +p_cpi->TrapQP                   << ","
                << +p_cpi->TrapQKey                 << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_N2N_CLASS_PORT_INFO);

    IBDIAG_RETURN(rc);
}

// Neighbors Info
int IBDiag::Retrieve_NeighborsInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;
    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::NeighborsInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct NeighborsInfo neighbors_info = { 0 };

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        p_curr_node->appData1.val = 0;
        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        // neighbor_record per port, 14 records in block
        u_int32_t num_of_blocks = p_curr_node->numPorts / NEIGHBORS_IN_BLOCK + 1;

        for (u_int32_t block = 0; block < num_of_blocks; ++block) {

            clbck_data.m_data2 = (void *)(uintptr_t)block;

            this->ibis_obj.ClassCNeighborsInfoGet(p_curr_node->getFirstLid(),
                                                  DEFAULT_SL,
                                                  block,
                                                  &neighbors_info,
                                                  &clbck_data);
        }

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::Dump_NeighborsInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_NEIGHBORS_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,record,node_type,lid,key" << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;

        if (!p_curr_node || !p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        for (u_int32_t record = 1; record <= p_curr_node->numPorts; ++record) {

            struct neighbor_record *p_record =
                    this->GetIBDMExtendedInfoPtr()->getNeighborRecord(p_curr_node->createIndex, record);

            // skip records for irrelevant ports
            if (!p_record || !p_record->node_type)
                continue;

            sstream.str("");

            sstream << PTR(p_curr_node->guid_get())    << ","
                    << +record                         << ","
                    << +p_record->node_type            << ","
                    << +p_record->lid                  << ","
                    << PTR(p_record->key)              << endl;

            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_NEIGHBORS_INFO);

    IBDIAG_RETURN(rc);
}

// N2N Key Info
int IBDiag::Retrieve_N2NKeyInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;
    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::N2NKeyInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct Class_C_KeyInfo key_info = { 0 };

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.ClassCKeyInfoGet(p_curr_node->getFirstLid(),
                                        DEFAULT_SL,
                                        &key_info,
                                        &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::Dump_N2NKeyInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_N2N_KEY_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,Key,KeyProtectBit,KeyLeasePeriod,KeyViolations,NodeKeyViolations" << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;

        if (!p_curr_node || !p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        struct Class_C_KeyInfo *p_key_info =
                this->GetIBDMExtendedInfoPtr()->getN2NKeyInfo(p_curr_node->createIndex);

        if (!p_key_info)
            continue;

        sstream.str("");

        sstream << PTR(p_curr_node->guid_get())      << ","
                << PTR(p_key_info->Key)              << ","
                << +p_key_info->KeyProtectBit        << ","
                << +p_key_info->KeyLeasePeriod       << ","
                << +p_key_info->KeyViolations        << ","
                << +p_key_info->NodeKeyViolations    << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_N2N_KEY_INFO);

    IBDIAG_RETURN(rc);
}

int IBDiag::pFRNSupportAndTrapsValidation(list_p_fabric_general_err &pfrn_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    u_int32_t pfrn_suppported = 0; // use as bit set
    set < lid_t > trap_lids;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;

        if (!p_curr_node)
            continue;

        pfrn_suppported |= 1 << p_curr_node->ispFRNSupported();

        if (!p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        struct IB_ClassPortInfo *p_cpi =
                this->GetIBDMExtendedInfoPtr()->getN2NClassPortInfo(p_curr_node->createIndex);

        struct Class_C_KeyInfo *p_key_info =
                this->GetIBDMExtendedInfoPtr()->getN2NKeyInfo(p_curr_node->createIndex);

        // check Trap LID only if:
        // 1. pFRN Enabled
        // 2. key is set (not 0)

        if (!p_cpi ||
                !p_key_info ||
                !p_curr_node->ispFRNEnabled() ||
                !p_key_info->Key)
            continue;

        trap_lids.insert(p_cpi->TrapLID);
    }

    // check if all switches support pFRN
    if (pfrn_suppported > 2) { // 2 bits are on means some support and some not
        pFRNErrPartiallySupported *p_curr_err =
                new pFRNErrPartiallySupported("Not all switches in fabric support pFRN");

        p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);
        pfrn_errors.push_back(p_curr_err);
    }

    // check if all switches send pFRN traps to the same LID
    if (trap_lids.size() > 1)
        pfrn_errors.push_back(
                new pFRNErrDiffTrapLIDs("Not all switches in fabric send pFRN traps to the same LID"));

    // check if all switches send pFRN traps to the SM LID
    sm_info_obj_t *cur_sm_master = NULL;

    for (list_p_sm_info_obj::iterator it =
            this->GetIBDMExtendedInfoPtr()->getSMPSMInfoListRef().begin();
            it != this->GetIBDMExtendedInfoPtr()->getSMPSMInfoListRef().end();
            ++it) {

        if ((*it)->smp_sm_info.SmState == IBIS_IB_SM_STATE_MASTER) {
            cur_sm_master = *it;
            break;
        }
    }

    if (trap_lids.size() == 1 &&
            cur_sm_master &&
            cur_sm_master->p_port->base_lid != *(trap_lids.begin())) {
        pFRNErrTrapLIDNotSM *p_curr_err =
                new pFRNErrTrapLIDNotSM("Switches don't report pFRN traps to master SM LID");

        p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);
        pfrn_errors.push_back(p_curr_err);
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::pFRNNeighborsValidation(list_p_fabric_general_err &pfrn_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;

        if (!p_curr_node || !p_curr_node->getInSubFabric() || !p_curr_node->ispFRNSupported())
            continue;

        for (u_int32_t record = 1; record <= p_curr_node->numPorts; ++record) {

            struct neighbor_record *p_record =
                    this->GetIBDMExtendedInfoPtr()->getNeighborRecord(p_curr_node->createIndex, record);

            // skip records for irrelevant ports
            if (!p_record || !p_record->node_type)
                continue;

            IBPort* p_neighbor_port = this->discovered_fabric.getPortByLid(p_record->lid);

            // check if neighbor exist
            if (!p_neighbor_port || !p_neighbor_port->p_node) {
                pFRNErrNeighborNotExist *p_curr_err =
                        new pFRNErrNeighborNotExist(p_curr_node, record);

                p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);
                pfrn_errors.push_back(p_curr_err);

                continue;
            }

            // check if neighbor is switch
            if (p_record->node_type != IB_SW_NODE ||
                    p_neighbor_port->p_node->type != IB_SW_NODE) {

                pFRNErrNeighborNotSwitch *p_curr_err =
                        new pFRNErrNeighborNotSwitch(p_curr_node, record);

                p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);
                pfrn_errors.push_back(p_curr_err);
            }
        } // records
    } // nodes

    IBDIAG_RETURN(rc);
}

int IBDiag::ARGroupsUniformValidation(list_p_fabric_general_err &pfrn_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (u_int8_t plft = 0; plft < MAX_PLFT_NUM; ++plft) {

        for (lid_t dlid = 1; dlid <= IB_MAX_UCAST_LID; ++dlid) {

            set <u_int16_t> ar_groups;

            for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
                    nI != this->discovered_fabric.Switches.end(); ++nI) {

                IBNode *p_curr_node = *nI;

                if (!p_curr_node || !p_curr_node->getInSubFabric() ||
                        !p_curr_node->isAREnable() ||
                        !p_curr_node->ispFRNEnabled() ||
                        plft > p_curr_node->getMaxPLFT() ||
                        dlid > p_curr_node->getLFDBTop(plft))
                    continue;

                u_int16_t ar_group = p_curr_node->getARLFTPortGroupForLid(dlid, plft);

                if (ar_group)
                    ar_groups.insert(ar_group);
            } // Switches

            if (ar_groups.size() > 1) {
                stringstream cur_err;
                cur_err << "Found different AR Groups ID for DLID " << dlid
                        << " for PLFT " << +plft << ". Groups: ";

                for (set <u_int16_t>::iterator it = ar_groups.begin(); it != ar_groups.end(); it++) {
                    cur_err << *it << "  ";
                }

                DifferentARGroupsIDForDLIDErr *p_curr_err =
                        new DifferentARGroupsIDForDLIDErr(rtrim(cur_err.str()));

                p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);
                pfrn_errors.push_back(p_curr_err);
            }
        } // lids
    } // PLFTs

    IBDIAG_RETURN(rc);
}

// Rail Filter
int IBDiag::BuildRailFilterDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarPorts progress_bar;
    struct SMP_RailFilterConfig rail_filter;
    CLEAR_STRUCT(rail_filter);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRailFilterGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsRailFilterSupported))
            continue;

        u_int8_t max_blocks = (u_int8_t)(p_curr_node->numPorts / RAIL_FILTER_EGRESS_BLOCK_SIZE + 1);

        // go over all ports of this switch:
        for (phys_port_t i = 1; i <= p_curr_node->numPorts; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);

            // Skip ports that don't exist or are down or have undefined state
            // or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            // resize RailFilter vector on switch port
            p_curr_port->railFilterData.outPorts.resize(max_blocks * RAIL_FILTER_EGRESS_BLOCK_SIZE);

            //get direct route for this port
            direct_route_t *p_curr_direct_route = this->GetDR(p_curr_port);
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to port=%s",
                        p_curr_port->getName().c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            progress_bar.push(p_curr_port);

            for (u_int8_t block = 0; block < max_blocks; ++block) {

                clbck_data.m_data1 = p_curr_port;
                clbck_data.m_data2 = (void *)(uintptr_t)block;

                this->ibis_obj.SMPRailFilterConfigGetByDirect(p_curr_direct_route,
                                                              i,     // port
                                                              0,     // ingress block - ignored
                                                              block, // block
                                                              &rail_filter,
                                                              &clbck_data);

            } // egress_blocks

            if (ibDiagClbck.GetState())
                goto exit;

        } // ports
    } // switches

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::WriteRailFilterFile(const string &file_name)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Rail Filter",
                            file_name,
                            sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpRailFilterInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpRailFilterInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    sout << "File version: " << RAIL_FILTER_FILE_VERSION << endl;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsRailFilterSupported))
            continue;

        stringstream ss;

        ss << "Switch "  << PTR(p_curr_node->guid_get()) << endl
           << "#switch-name=" << p_curr_node->getName() << endl << endl
           << "Rail Filter DB:" << endl
           << setfill(' ')
           << std::setw(10) << std::left << "In Port"
           << std::setw(25) << std::left << "VLs"
           << std::setw(10) << std::left << "MCEnable"
           << std::setw(10) << std::left << "UCEnable"
           << "Out Ports List" << endl
           << "--------------------------------------------------------------------------" << endl;

        for (phys_port_t in_p = 1; in_p <= p_curr_node->numPorts; ++in_p) {

            IBPort *p_curr_port = p_curr_node->getPort(in_p);

            // Skip ports that don't exist or are down or have undefined state
            // or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            stringstream out_VLs;
            stringstream out_ports;
            string delimiter = "";

            for (int vl = 0; vl < 15 ; ++vl) {

                if ((p_curr_port->railFilterData.VLMask >> vl) & 1) {
                    out_VLs << delimiter << vl;
                    delimiter = ",";
                }
            } // VLs

            delimiter = "";
            for (phys_port_t out_p = 1; out_p <= p_curr_node->numPorts; ++out_p) {

                if (p_curr_port->railFilterData.outPorts[out_p]) {
                    out_ports << delimiter << +out_p;
                    delimiter = ", ";
                }
            } // out_ports

            // check if any out_port is on
            if (!out_ports.str().empty()) {
                ss << std::setw(10) << std::left << +in_p
                   << std::setw(25) << std::left << out_VLs.str()
                   << std::setw(10) << std::left << p_curr_port->railFilterData.MCEnable
                   << std::setw(10) << std::left << p_curr_port->railFilterData.UCEnable
                   << out_ports.str()
                   << endl;
            }

        } // in_ports

        sout << ss.rdbuf() << endl << endl;

    } // Switches

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// Entry Plane Filter
int IBDiag::BuildEntryPlaneFilterDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarPorts progress_bar;
    struct SMP_EntryPlaneFilterConfig entry_plane_filter;
    CLEAR_STRUCT(entry_plane_filter);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPEntryPlaneFilterGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsEntryPlaneFilterSupported))
            continue;

        ib_extended_node_info* p_ext_node_info =
                        this->fabric_extended_info.getSMPExtNodeInfo(p_curr_node->createIndex);
        if (!p_ext_node_info)
            continue;

        u_int8_t max_planes = p_ext_node_info->asic_max_planes;
        u_int8_t max_blocks = (u_int8_t)(p_curr_node->numPorts / ENTRY_PLANE_FILTER_EGRESS_BLOCK_SIZE + 1);

        // resize EPF vector on switch for in_ports
        p_curr_node->EPF.resize(p_curr_node->numPorts + 1);

        // go over all ports of this switch:
        for (phys_port_t i = 1; i <= p_curr_node->numPorts; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);

            // Skip ports that don't exist or are down or have undefined state
            // or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            // resize EPF vector on switch for planes
            p_curr_node->EPF[i].resize(max_planes + 1);

            //get direct route for this port
            direct_route_t *p_curr_direct_route = this->GetDR(p_curr_port);
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to port=%s",
                        p_curr_port->getName().c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            progress_bar.push(p_curr_port);

            for (u_int8_t plane = 1; plane <= max_planes; ++plane) {

                // resize EPF vector on switch for out_ports
                p_curr_node->EPF[i][plane].resize(max_blocks * ENTRY_PLANE_FILTER_EGRESS_BLOCK_SIZE);

                for (u_int8_t block = 0; block < max_blocks; ++block) {

                    clbck_data.m_data1 = p_curr_port;
                    clbck_data.m_data2 = (void *)(uintptr_t)plane;
                    clbck_data.m_data3 = (void *)(uintptr_t)block;

                    this->ibis_obj.SMPEntryPlaneFilterConfigMadGetByDirect(p_curr_direct_route,
                                                                           i,     // port
                                                                           plane, // plane
                                                                           block, // block
                                                                           &entry_plane_filter,
                                                                           &clbck_data);

                } // egress_blocks
            } // planes

            if (ibDiagClbck.GetState())
                goto exit;

        } // ports
    } // switches

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::WriteEPFFile(const string &file_name)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Entry Plane Filter",
                            file_name,
                            sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpEPFInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpEPFInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    sout << "File version: " << EPF_FILE_VERSION << endl;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsEntryPlaneFilterSupported))
            continue;

        ib_extended_node_info* p_ext_node_info =
                        this->fabric_extended_info.getSMPExtNodeInfo(p_curr_node->createIndex);
        if (!p_ext_node_info)
            continue;

        u_int8_t max_planes = p_ext_node_info->asic_max_planes;

        stringstream ss;

        ss << "Switch "  << PTR(p_curr_node->guid_get()) << endl
           << "#switch-name=" << p_curr_node->getName() << endl << endl
           << "Entry Plane Filter DB:" << endl
           << setfill(' ')
           << std::setw(10) << std::left << "In Port"
           << std::setw(10) << std::left << "Plane"
           << "Out Ports List" << endl
           << "--------------------------------------------------------------------------" << endl;

        for (phys_port_t in_p = 1; in_p <= p_curr_node->numPorts; ++in_p) {

            IBPort *p_curr_port = p_curr_node->getPort(in_p);

            // Skip ports that don't exist or are down or have undefined state
            // or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            for (u_int8_t plane = 1; plane <= max_planes; ++plane) {

                stringstream out_ports;
                string delimiter = "";

                for (phys_port_t out_p = 1; out_p <= p_curr_node->numPorts; ++out_p) {

                    if (p_curr_node->EPF[in_p][plane][out_p]) {
                        out_ports << delimiter << +out_p;
                        delimiter = ", ";
                    }
                } // out_ports

                // check if any out_port is on
                if (!out_ports.str().empty()) {
                    ss << std::setw(10) << std::left << +in_p
                       << std::setw(10) << std::left << +plane
                       << out_ports.str()
                       << endl;
                }

            } // planes
        } // in_ports

        sout << ss.rdbuf() << endl << endl;

    } // Switches

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// Next table define how EntryPlaneFilter should work
// at all possible scenarios where in_port is either 1,2,4
// and out_port is also either 1,2,4

// in: 1x (or legacy) out: 1x (or legacy)
static bool EPFTable_1_1[4][4] = {1, 0, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0};
// in: 2x out: 1x (or legacy)
static bool EPFTable_2_1[4][4] = {1, 0, 0, 0,
                                  1, 0, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0};
// in: 4x out: 1x (or legacy)
static bool EPFTable_4_1[4][4] = {1, 0, 0, 0,
                                  1, 0, 0, 0,
                                  1, 0, 0, 0,
                                  1, 0, 0, 0};
// in: 1x (or legacy) out: 2x
static bool EPFTable_1_2[4][4] = {1, 1, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0};
// in: 2x out: 2x
static bool EPFTable_2_2[4][4] = {1, 0, 0, 0,
                                  0, 1, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0};
// in: 4x out: 2x
static bool EPFTable_4_2[4][4] = {1, 0, 0, 0,
                                  0, 1, 0, 0,
                                  1, 0, 0, 0,
                                  0, 1, 0, 0};
// in: 1x (or legacy) out: 4x
static bool EPFTable_1_4[4][4] = {1, 1, 1, 1,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0};
// in: 2x out: 4x
static bool EPFTable_2_4[4][4] = {1, 0, 1, 0,
                                  0, 1, 0, 1,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0};
// in: 4x out: 4x
static bool EPFTable_4_4[4][4] = {1, 0, 0, 0,
                                  0, 1, 0, 0,
                                  0, 0, 1, 0,
                                  0, 0, 0, 1};


bool IBDiag::isAvailableByEPF(IBPort *in_port, IBPort *out_port)
{
    IBDIAG_ENTER;

    int in_num_of_planes = 0;
    int out_num_of_planes = 0;
    int in_plane_num = 0;
    int out_plane_num = 0;

    if (in_port->p_aport) {
        in_num_of_planes = in_port->p_port_hierarchy_info->m_num_of_planes;
        in_plane_num = in_port->p_port_hierarchy_info->m_plane;
    } else { // legacy
        in_num_of_planes = 1;
        in_plane_num = 1;
    }

    if (out_port->p_aport) {
        out_num_of_planes = out_port->p_port_hierarchy_info->m_num_of_planes;
        out_plane_num = out_port->p_port_hierarchy_info->m_plane;
    } else { // legacy
        out_num_of_planes = 1;
        out_plane_num = 1;
    }

    switch (in_num_of_planes) {
        case 1:
            switch (out_num_of_planes) {
                case 1:  return EPFTable_1_1[in_plane_num - 1][out_plane_num - 1];
                case 2:  return EPFTable_1_2[in_plane_num - 1][out_plane_num - 1];
                case 4:  return EPFTable_1_4[in_plane_num - 1][out_plane_num - 1];
            }
            break;

        case 2:
            switch (out_num_of_planes) {
                case 1:  return EPFTable_2_1[in_plane_num - 1][out_plane_num - 1];
                case 2:  return EPFTable_2_2[in_plane_num - 1][out_plane_num - 1];
                case 4:  return EPFTable_2_4[in_plane_num - 1][out_plane_num - 1];
            }
            break;

        case 4:
            switch (out_num_of_planes) {
                case 1:  return EPFTable_4_1[in_plane_num - 1][out_plane_num - 1];
                case 2:  return EPFTable_4_2[in_plane_num - 1][out_plane_num - 1];
                case 4:  return EPFTable_4_4[in_plane_num - 1][out_plane_num - 1];
            }
    }

    IBDIAG_RETURN(false);
}


int IBDiag::EntryPlaneFilterValidation(list_p_fabric_general_err& validation_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    bool epf_check_failed = false;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsEntryPlaneFilterSupported)) {
            if (p_curr_node->EPF.size()) {
                validation_errors.push_back(new EntryPlaneFilterUnexpected(p_curr_node));
                epf_check_failed = true;
            }

            continue;
        }

        ib_extended_node_info* p_ext_node_info =
                        this->fabric_extended_info.getSMPExtNodeInfo(p_curr_node->createIndex);
        if (!p_ext_node_info)
            continue;

        u_int8_t max_planes = p_ext_node_info->asic_max_planes;

        if (!p_curr_node->CheckEPFSize(max_planes)) {
            validation_errors.push_back(new EntryPlaneFilterInvalidSize(p_curr_node));
            epf_check_failed = true;
            continue;
        }

        for (phys_port_t in_p = 1; in_p <= p_curr_node->numPorts; ++in_p) {

            IBPort *p_in_port = p_curr_node->getPort(in_p);
            if (!p_in_port || !p_in_port->is_data_worthy())
                continue;

            for (phys_port_t out_p = 1; out_p <= p_curr_node->numPorts; ++out_p) {

                IBPort *p_out_port = p_curr_node->getPort(out_p);
                if (!p_out_port || !p_out_port->is_data_worthy())
                    continue;

                // skip same port
                if (p_in_port == p_out_port)
                    continue;

                bool expected = isAvailableByEPF(p_in_port, p_out_port);
                bool actual   = p_curr_node->getEPFFromAllPlanes(in_p, out_p, max_planes);

                if (expected != actual) {
                    validation_errors.push_back(new EntryPlaneFilterMismatch(p_curr_node,
                                                                             in_p,
                                                                             out_p,
                                                                             expected,
                                                                             actual));
                    epf_check_failed = true;
                }

            } // out_ports
        } // in_ports
    } // switches

    this->discovered_fabric.SetIsEPFValid(!epf_check_failed);

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildEndPortPlaneFilterDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;
    struct SMP_EndPortPlaneFilterConfig end_pplane_filter;
    CLEAR_STRUCT(end_pplane_filter);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPEndPortPlaneFilterGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (IBNode *p_curr_node : this->discovered_fabric.HCAs) {
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in HCAs");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsEndPortPlaneFilterConfigSupported))
            continue;

        direct_route_t *p_curr_route = this->GetDR(p_curr_node);
        if (!p_curr_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // resize EndPortPlaneFilter
        p_curr_node->EndPortPlaneFilter.resize(4 + 1);

        progress_bar.push(p_curr_node);
        this->ibis_obj.SMPEndPortPlaneFilterConfigMadGetByDirect(p_curr_route, &end_pplane_filter,
                            &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::WriteEndPortPlaneFilterFile(const string &file_name)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("End Port Plane Filter",
                            file_name,
                            sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpEndPortPlaneFilterInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

int IBDiag::EndPortPlaneFilterValidation(list_p_fabric_general_err& validation_errors)
{
    IBDIAG_ENTER;

    for (IBNode *p_curr_node : this->discovered_fabric.HCAs) {

        if (!p_curr_node) {
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsEndPortPlaneFilterConfigSupported)) {
            if (p_curr_node->EndPortPlaneFilter.size()) {
                validation_errors.push_back(new EndPortPlaneFilterUnexpected(p_curr_node));
            }

            continue;
        }

        for (size_t i = 1; i < p_curr_node->EndPortPlaneFilter.size(); ++i) {
            // skip 0 entries
            if (!p_curr_node->EndPortPlaneFilter[i])
                continue;

            IBPort *p_end_port = this->discovered_fabric.getPortByLid(
                                    p_curr_node->EndPortPlaneFilter[i]);
            if (!p_end_port) {
                validation_errors.push_back(
                    new EndPortPlaneFilterInvalidLID(p_curr_node, i));
                continue;
            }

            IBPort *p_local_port = p_curr_node->getPort(phys_port_t(i));
            if (!p_local_port)
                continue;


            if (p_end_port->p_node->getSpecialNodeType() != IB_SPECIAL_PORT_AN) {
                validation_errors.push_back(
                    new EndPortPlaneFilterInvalidNodeType(p_curr_node, i));
                continue;
            }

            if (!p_end_port->p_remotePort || !p_local_port->p_remotePort ||
                    (p_end_port->p_remotePort->p_node != p_local_port->p_remotePort->p_node) ) {
                validation_errors.push_back(
                    new EndPortPlaneFilterWrongLID(p_curr_node, i));
            }
        } // EndPortPlaneFilter LIDs
    } // HCAs

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpEndPortPlaneFilterInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    sout << "File version: " << END_PORT_PLANE_FILTER_FILE_VERSION << endl;

    for (IBNode *p_curr_node : this->discovered_fabric.HCAs) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // check if supported
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsEndPortPlaneFilterConfigSupported))
            continue;
        if (!p_curr_node->EndPortPlaneFilter.size())
            continue;

        stringstream ss;

        ss << "HCA "  << PTR(p_curr_node->guid_get()) << endl
           << "#hca-name=" << p_curr_node->getName() << endl << endl
           << "End Port Plane Filter DB:" << endl
           << setfill(' ')
           << std::setw(10) << std::left << "Plane"
           << "LID"<< endl
           << "--------------------------------------------------------------------------" << endl;

        for (size_t i = 1; i < p_curr_node->EndPortPlaneFilter.size(); ++i) {
           ss << std::setw(10) << std::left << DEC(i)
              << PTR(p_curr_node->EndPortPlaneFilter[i])
              << endl;
        }

        sout << ss.rdbuf() << endl << endl;

    } // HCAs

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::StaticRoutingSymmetricLinkValidation(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        set_ports checked_ports;

        for (u_int8_t pLFT = 0; pLFT <= p_curr_node->getMaxPLFT(); ++pLFT) {

            u_int16_t lfdbTop = (u_int16_t)p_curr_node->LFT[pLFT].size();

            if (p_curr_node->isPLFTEnabled())
                lfdbTop = p_curr_node->getLFDBTop(pLFT);

            for (lid_t lid = 1; lid <= lfdbTop; ++lid) {

                phys_port_t port = p_curr_node->getLFTPortForLid(lid, pLFT);

                if (checked_ports.count(port))
                    continue;

                checked_ports.insert(port);

                IBPort *p_port = p_curr_node->getPort(port);

                // check if this aport is asymmetric
                if (p_port &&
                        p_port->p_aport &&
                        !p_port->p_aport->isSymmetric()) {

                    errors.push_back(new StaticRoutingAsymmetricLink(p_curr_node,
                                                                     p_port,
                                                                     lid,
                                                                     pLFT));
                }
            } // lids
        } // pLFTs
    } //Switches

    IBDIAG_RETURN(rc);
}

int IBDiag::AdaptiveRoutingSymmetricLinkValidation(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isAREnable() )
            continue;

        set_ports checked_ports;

        for (u_int8_t pLFT = 0; pLFT <= p_curr_node->getMaxPLFT(); ++pLFT) {

            u_int16_t lfdbTop = (u_int16_t)p_curr_node->LFT[pLFT].size();

            if (p_curr_node->isPLFTEnabled())
                lfdbTop = p_curr_node->getLFDBTop(pLFT);

            for (lid_t lid = 1; lid <= lfdbTop; ++lid) {

                list_phys_ports ports;
                p_curr_node->getLFTPortListForLid(lid, pLFT, true, ports);

                for (phys_port_t port : ports) {

                    if (checked_ports.count(port))
                        continue;

                    checked_ports.insert(port);

                    IBPort *p_port = p_curr_node->getPort(port);

                    // check if this aport is asymmetric
                    if (p_port &&
                            p_port->p_aport &&
                            !p_port->p_aport->isSymmetric()) {

                        errors.push_back(new AdaptiveRoutingAsymmetricLink(p_curr_node,
                                                                           p_port,
                                                                           lid,
                                                                           pLFT));
                    }
                } // ports from group
            } // lids
        } // pLFTs
    } //Switches

    IBDIAG_RETURN(rc);
}

// SMP Linear Forwarding Table Split
int IBDiag::BuildLFTSplit(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;
    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::SMP_LFTSplitGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_LinearForwardingTableSplit lft_split = { 0 };

    for (IBNode* p_node : this->discovered_fabric.Switches) {

        if (!p_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_node->getInSubFabric())
            continue;

        // check cap bit
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                    p_node, EnSMPCapIsLinearForwardingTableSplitSupported))
            continue;

        // get direct route
        direct_route_t *p_curr_direct_route = this->GetDR(p_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_node;
        progress_bar.push(p_node);

        this->ibis_obj.SMPLinearForwardingTableSplitGetByDirect(p_curr_direct_route,
                                                                &lft_split,
                                                                &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpLFTSplitCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_LINEAR_FORWARDING_TABLE_SPLIT))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "global_lid_range_start,"
               "global_lid_range_cap,"
               "global_lid_range_top,"
               "alid_range_start,"
               "alid_range_cap,"
               "alid_range_top,"
               "local_plane_lid_range_start,"
               "local_plane_lid_range_cap,"
               "local_plane_lid_range_top"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_node : this->discovered_fabric.Switches) {

        if (!p_node)
            continue;

        struct SMP_LinearForwardingTableSplit *p_data =
                this->fabric_extended_info.getSMP_LFTSplit(p_node->createIndex);

        if (!p_data)
            continue;

        sstream.str("");
        sstream << PTR(p_node->guid_get())              << ','
                << p_data->global_lid_range_start       << ','
                << p_data->global_lid_range_cap         << ','
                << p_data->global_lid_range_top         << ','
                << p_data->alid_range_start             << ','
                << p_data->alid_range_cap               << ','
                << p_data->alid_range_top               << ','
                << p_data->local_plane_lid_range_start  << ','
                << p_data->local_plane_lid_range_cap    << ','
                << p_data->Local_plane_lid_range_top    << endl;

        csv_out.WriteBuf(sstream.str());

    } // nodes

    csv_out.DumpEnd(SECTION_LINEAR_FORWARDING_TABLE_SPLIT);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
