/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _ACC_REG_KEY_H_
#define _ACC_REG_KEY_H_

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include "../internal_packets/internal_packets_layouts.h"

class AccRegKey {

public:
    const uint64_t node_guid;

    AccRegKey(uint64_t ng) : node_guid(ng) {}

    virtual ~AccRegKey(){}

    virtual void DumpKeyData(ostream &stream) const = 0;
    virtual void DumpKeyHeader(ostream &stream) const = 0;

    virtual bool operator < (const AccRegKey &other) const = 0;
};


class AccRegKeyPortLane : public AccRegKey {

public:
    const uint64_t port_guid;
    const uint8_t port_num;
    const uint8_t lane;
    const uint8_t idx_in_lane;

    virtual bool operator < (const AccRegKey &other) const {
        if (node_guid < other.node_guid)
            return true;
        if (node_guid == other.node_guid &&
            port_guid < ((AccRegKeyPortLane&)other).port_guid)
            return true;
        if (node_guid == other.node_guid &&
            port_guid == ((AccRegKeyPortLane&)other).port_guid &&
          port_num < ((AccRegKeyPortLane&)other).port_num)
            return true;
        if (node_guid == other.node_guid &&
            port_guid == ((AccRegKeyPortLane&)other).port_guid &&
            port_num == ((AccRegKeyPortLane&)other).port_num &&
            lane < ((AccRegKeyPortLane&)other).lane)
            return true;
        if (node_guid == other.node_guid &&
            port_guid == ((AccRegKeyPortLane&)other).port_guid &&
            port_num == ((AccRegKeyPortLane&)other).port_num &&
            lane == ((AccRegKeyPortLane&)other).lane &&
            idx_in_lane < ((AccRegKeyPortLane&)other).idx_in_lane)
            return true;

        return false;
    }

    AccRegKeyPortLane(uint64_t ng,
                      uint64_t pg,
                      uint8_t pn,
                      uint8_t ln,
                      uint8_t idx = 0) : AccRegKey(ng),
                                         port_guid(pg),
                                         port_num(pn),
                                         lane(ln),
                                         idx_in_lane(idx) {}

    virtual ~AccRegKeyPortLane(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};

class AccRegKeyNode : public AccRegKey {

public:

    virtual bool operator < (const AccRegKey &other) const {
        if (node_guid < other.node_guid)
            return true;
        return false;
    }

    AccRegKeyNode(uint64_t ng) : AccRegKey(ng) {}
    virtual ~AccRegKeyNode(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};

class AccRegKeyPort : public AccRegKey {

public:
    const uint64_t port_guid;
    const uint8_t port_num;

    virtual bool operator < (const AccRegKey &other) const {
        if (node_guid < other.node_guid)
            return true;
        if (node_guid == other.node_guid &&
            port_guid < ((AccRegKeyPort&)other).port_guid)
            return true;
        if (node_guid == other.node_guid &&
            port_guid == ((AccRegKeyPort&)other).port_guid &&
            port_num < ((AccRegKeyPort&)other).port_num)
            return true;
        return false;
    }
    AccRegKeyPort(uint64_t ng,
                  uint64_t pg,
                  uint8_t pn) : AccRegKey(ng),
                                port_guid(pg),
                                port_num(pn) {}

    virtual ~AccRegKeyPort(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};

//  ** AccRegKeyPortIndex **
// todo : check if should be derived from AccRegKeyPort
class AccRegKeyPortIndex : public AccRegKey {

public:
    const uint64_t port_guid;
    const uint8_t  port_num;
    const uint64_t  index;

    virtual bool operator < (const AccRegKey &other) const {

        if (node_guid != other.node_guid)
            return node_guid < other.node_guid;
        if (port_guid != ((AccRegKeyPortIndex&)other).port_guid)
            return port_guid < ((AccRegKeyPortIndex&)other).port_guid;
        if (port_num != ((AccRegKeyPortIndex&)other).port_num)
            return port_num < ((AccRegKeyPortIndex&)other).port_num;
        if (index != ((AccRegKeyPortIndex&)other).index)
            return index < ((AccRegKeyPortIndex&)other).index;

        return false;
    }

    AccRegKeyPortIndex(uint64_t ng, uint64_t pg, uint8_t pn, uint64_t  idx) : AccRegKey(ng),
                                                                              port_guid(pg),
                                                                              port_num(pn),
                                                                              index(idx){}

    virtual ~AccRegKeyPortIndex(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};

class AccRegKeyGroup : public AccRegKey {

public:
    const uint8_t group_num;

    virtual bool operator < (const AccRegKey &other) const {
        if (node_guid < other.node_guid)
            return true;
        if (node_guid == other.node_guid &&
            group_num < ((AccRegKeyGroup&)other).group_num)
            return true;
        return false;
    }
    AccRegKeyGroup(uint64_t ng, uint8_t gr) : AccRegKey(ng),
                                              group_num(gr) {}

    virtual ~AccRegKeyGroup(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};

class AccRegKeyNodeSensor : public AccRegKey {

public:
    const uint8_t sensor_id;

    virtual bool operator < (const AccRegKey &other) const {
        if (node_guid < other.node_guid)
            return true;
        if (node_guid == other.node_guid &&
            sensor_id < ((AccRegKeyNodeSensor&)other).sensor_id)
            return true;
        return false;
    }

    AccRegKeyNodeSensor(uint64_t node_guid, uint8_t sensor_id) : AccRegKey(node_guid),
                                                                 sensor_id(sensor_id) {}

    virtual ~AccRegKeyNodeSensor(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};


class AccRegKeyTypeGroup : public AccRegKey {

public:
    const uint8_t group_type;
    const uint8_t group_num;
    const uint16_t start_index;
    const uint16_t num_of_indices;

    virtual bool operator < (const AccRegKey &other) const {

        if (node_guid != other.node_guid)
            return node_guid < other.node_guid;
        if (group_type != ((AccRegKeyTypeGroup&)other).group_type)
            return group_type < ((AccRegKeyTypeGroup&)other).group_type;
        if (group_num != ((AccRegKeyTypeGroup&)other).group_num)
            return group_num < ((AccRegKeyTypeGroup&)other).group_num;
        if (start_index != ((AccRegKeyTypeGroup&)other).start_index)
            return start_index < ((AccRegKeyTypeGroup&)other).start_index;

        return false;
    }

    AccRegKeyTypeGroup(uint64_t ng,
                       uint8_t grp_type,
                       uint8_t grp_num,
                       uint16_t start_idx,
                       uint16_t indices_num) : AccRegKey(ng),
                                               group_type(grp_type),
                                               group_num(grp_num),
                                               start_index(start_idx),
                                               num_of_indices(indices_num) {}

    virtual ~AccRegKeyTypeGroup(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};


class AccRegKeyDPN : public AccRegKey {

public:
    const IBPort *p_port;
    const uint8_t depth;
    const uint8_t pci_idx;
    const uint8_t pci_node;

    virtual bool operator < (const AccRegKey &other) const {

        if (node_guid != other.node_guid)
            return node_guid < other.node_guid;
        if (pci_idx != ((AccRegKeyDPN&)other).pci_idx)
            return pci_idx < ((AccRegKeyDPN&)other).pci_idx;
        if (depth != ((AccRegKeyDPN&)other).depth)
            return depth < ((AccRegKeyDPN&)other).depth;
        if (pci_node != ((AccRegKeyDPN&)other).pci_node)
            return pci_node < ((AccRegKeyDPN&)other).pci_node;

        return false;
    }

    AccRegKeyDPN(const IBPort *port,
                 uint8_t depth,
                 uint8_t pci_idx,
                 uint8_t pci_node) : AccRegKey(port->p_node->guid_get()),
                                     p_port(port),
                                     depth(depth),
                                     pci_idx(pci_idx),
                                     pci_node(pci_node) {}

    virtual ~AccRegKeyDPN(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};

class AccRegKeyPCIPortLane : public AccRegKeyPortLane {

public:
    AccRegKeyPCIPortLane(uint64_t ng,
                         uint8_t local_port,
                         uint8_t lane) : AccRegKeyPortLane(ng, 0, local_port, lane) {}

    virtual ~AccRegKeyPCIPortLane(){}

    virtual void DumpKeyData(ostream &stream) const;
    virtual void DumpKeyHeader(ostream &stream) const;
};


#endif // _ACC_REG_KEY_H_
