/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


/*
 * Subnet Utilities:
 *
 * The file holds a set of utilities to be run on the subnet to mimic OpenSM
 * initialization and analyze the results:
 *
 * Assign Lids: SubnMgtAssignLids
 * Init min hop tables: SubnMgtCalcMinHopTables
 * Perform Enhanced LMC aware routing: SubnMgtOsmEnhancedRoute
 * Perform standard routing: SubnMgtOsmRoute
 * Verify all CA to CA routes: SubnMgtVerifyAllCaToCaRoutes
 */


#include <set>
#include <algorithm>
#include <iomanip>

#include <infiniband/misc/regexp/regexp.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/TraceRoute.h>
#include <infiniband/ibdm/ARTraceRoute.h>
#include <unordered_set>
#include <vector>

extern int g_clearedCache;

// saved for a specific node, this struct records data that regards dfsSendRecvMemMCG.
// it records which root_node the DFS was started from, which planes was this nodes 
// discovered from, and from which planes in the current node the DFS already discovered.
class IBNodeVisitsRecord {
public:
    IBNode            *root_node;
    set<int>          entry_planes;
    // record exit port, to avoid using the same port twice.
    set<phys_port_t>  exit_ports;
    IBNodeVisitsRecord() : root_node(nullptr) {};
    IBNodeVisitsRecord(IBNode *root_node, int entry_planes);
};

IBNodeVisitsRecord::IBNodeVisitsRecord(IBNode *root_node, int entry_plane)
        : root_node(root_node)
{
    entry_planes.insert(entry_plane);
}

class NodesVisits {
public:
    map<IBNode *, IBNodeVisitsRecord> visits;

    bool isVisited(IBNode *p_node, IBPort *entry_port = nullptr);
    bool addVisit(IBNode *p_node, IBNode *root_node, IBPort *entry_port = nullptr);
    bool isAlreadyExited(IBNode *p_node, phys_port_t exit_pn);
    bool addExit(IBNode *p_node, phys_port_t exit_pn);
    bool addAPortVisit(APort *p_aport, IBNode *root_node);
    IBNode *getRootNode(IBNode *p_node) const;
};

// Checks if a node was already visited. if IBPort is given,
// checks whether the node was visited from the plane of the given IBPort.
bool NodesVisits::isVisited(IBNode *p_node, IBPort *entry_port) {
    int entry_plane = entry_port ? entry_port->get_plane_number() : -1;
    const auto& it = visits.find(p_node);
    if (it == visits.end())
        return false;
    if (entry_plane == -1 || it->second.entry_planes.count(-1))
        return true;

    return it->second.entry_planes.count(entry_plane);
}

// true if new node visit added, false if colliding visit already recorded
bool NodesVisits::addVisit(IBNode *p_node, IBNode *root_node, IBPort *entry_port) {
    int entry_plane = entry_port ? entry_port->get_plane_number() : -1;
    auto it = visits.find(p_node);
    if (it == visits.end()) {
        visits[p_node] = IBNodeVisitsRecord(root_node, entry_plane);
        return true;
    }
    // if current/prev visits were non-planarized, return false.
    if (entry_plane == -1 || it->second.entry_planes.count(-1))
        return false;

    // attempt to insert and return true if new element was inserted
    return (it->second.entry_planes.insert(entry_plane).second);
}

// true if new exit port number was added, false otherwise.
bool NodesVisits::addExit(IBNode *p_node, phys_port_t exit_pn) {
    auto it = visits.find(p_node);
    if (it == visits.end())
        return false;

    return (it->second.exit_ports.insert(exit_pn).second);
}

bool NodesVisits::isAlreadyExited(IBNode *p_node, phys_port_t exit_pn) {
    auto it = visits.find(p_node);
    if (it == visits.end())
        return false;

    return (it->second.exit_ports.count(exit_pn));
}

bool NodesVisits::addAPortVisit(APort *p_aport, IBNode *root_node) {
    bool success = true;
    for (IBPort *p_port : p_aport->ports) {
        if (!p_port)
            continue;
        if (!addVisit(p_port->p_node, root_node, p_port))
            success = false;
    }

    return success;
}

IBNode * NodesVisits::getRootNode(IBNode *p_node) const {
    auto it = visits.find(p_node);
    if (it == visits.end())
        return nullptr;

    return it->second.root_node;
}

// every multicast send only node should eventually reach a send/recv multicast tree.
// this class is saved for each send only node. It records which send/recv multicast trees
// are reached from each port of the node. Multicast trees are indexed by 'root nodes'.
// it also saves which multicast trees were reached per plane, to allow checking if a tree
// was reached twice from the same plane.
class NodeReachPerPort{
public:
    // reach is saved per each port and for each plane (aggregation of all ports of that plane)
    map<phys_port_t, set<IBNode *>> reach_per_port;
    // -1 is for no plane (legacy), -2 is for all planarized together
    map<int, set<IBNode *>> reach_per_plane;

    /**
     * @brief This function checks if a given mcast tree is can be reached from a port.
     *        More specifically, this function checks if the mcast tree can be reached by any
     *        port of the same plane as the given port.
     * 
     * @param mcast_root - the root node of the send/recv mcast tree.
     * @param p_port - output port from the current node.
     * @return if p_port is planarized, check if the tree can be reached from ports 
     *         with p_port's plane or from a non-planarized port.
     *         if p_port is legacy, check if tree can be reached from any port.
     */
    bool isReached(IBNode *mcast_root, IBPort *p_port) const;
    bool addReached(IBNode *p_node, IBPort *p_port);
    bool reachedAny() const;
};

bool NodeReachPerPort::isReached(IBNode *mcast_root, IBPort *p_port) const {
    // search can be done per plane alone, as all nodes should be there.
    int plane_num = p_port->get_plane_number();
    if (plane_num == -1) {
        // search non-planarized and all planes
        auto it = reach_per_plane.find(plane_num);
        if (it != reach_per_plane.end() && it->second.count(mcast_root))
            return true;
        it = reach_per_plane.find(-2);
        if (it != reach_per_plane.end() && it->second.count(mcast_root))
            return true;
    } else {
        // search non-planarized and the given plane
        auto it = reach_per_plane.find(plane_num);
        if (it != reach_per_plane.end() && it->second.count(mcast_root))
            return true;
        it = reach_per_plane.find(-1);
        if (it != reach_per_plane.end() && it->second.count(mcast_root))
            return true;
    }

    return false;
}

bool NodeReachPerPort::addReached(IBNode *p_node, IBPort *p_port) {
    if (isReached(p_node, p_port))
        return false;

    reach_per_port[p_port->num].insert(p_node);

    int plane_num = p_port->get_plane_number();
    reach_per_plane[plane_num].insert(p_node);
    if (plane_num != -1)
        reach_per_plane[-2].insert(p_node);

    return true;
}

bool NodeReachPerPort::reachedAny() const {
    for (auto& it : reach_per_port) {
        if (!it.second.empty())
            return true;
    }

    return false;
}

class SONodesReach {
public:
    map<IBNode *, NodeReachPerPort> reach_by_node;
    /**
     * @brief Adds to the reach of p_node nodes from the reach of p_remote.
     *        Nodes are only added if they are reached from ports that are accessible from
     *        the given exit port (in terms of EPF).
     * 
     * @param p_node - a node whose reach would be expanded
     * @param p_remote - node connected to p_node
     * @param p_exit_port - port on p_node that connects both nodes.
     * @return false if a tree that can already be reached by p_node is attempted to be added.
     *         otherwise - true.
     */
    bool addRemoteReach(IBNode *p_node, IBNode *p_remote, IBPort *p_exit_port);
    NodeReachPerPort& operator[](IBNode *p_node) { return reach_by_node[p_node]; };
};

bool SONodesReach::addRemoteReach(IBNode *p_node, IBNode *p_remote, IBPort *p_exit_port) {
    bool ret = true;

    // go over all the remote node's ports with recorded reaches.
    // add the remote node's reaches to the current node and check duplicates.
    for (auto& pn_reach_pair : reach_by_node[p_remote].reach_per_port ) {
        phys_port_t remote_out = pn_reach_pair.first;
        set<IBNode *>& remote_out_reach = pn_reach_pair.second;

        if (!p_exit_port->p_remotePort)
            continue;

        // do not add to reach if the path is blocked by Rail Filter
        if (!p_exit_port->p_remotePort->isPassingRailFilter(remote_out))
            continue;

        if (remote_out_reach.empty())
            continue;
        
        for (IBNode *p_remote_reach : remote_out_reach) {
            if (!reach_by_node[p_node].addReached(p_remote_reach, p_exit_port))
                ret = false;
        }
    }

    return ret;
}

bool IsRankAssigned(const IBNode* p_node) {
    if (p_node->rank == IBNODE_UNASSIGNED_RANK) {
        cout << "-E- Rank wasn't assigned for: " << p_node->name
		     << endl;
	    return false;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Assign lids given the start NodePort
// NOTE: does not support a different LMC for SW and CA

// separate for clarity
static int
AssignNodeLids(IBPort *p_port, lid_t lid,  uint8_t lmc, lid_t numLidsPerPort)
{
    IBNode *p_node = p_port->p_node;
    IBFabric* p_fabric = p_node->p_fabric;
    if ( p_node->type != IB_SW_NODE ) {
        p_port->base_lid = lid;
        p_port->lmc = lmc;
        for ( lid_t l = lid ; l < lid + numLidsPerPort; l++ )
            p_fabric->setLidPort(l, p_port);
    } else {
        // go over all ports of the node:
        // use unsigned int index to prevent overwrap
        for ( unsigned int pn = 0; pn <= p_node->numPorts; pn++ ) {
            IBPort *p_curr_port = p_node->getPort((phys_port_t)pn);
            if ( p_curr_port ) {
                p_curr_port->base_lid = lid;
                p_curr_port->lmc = lmc;
                for ( lid_t l = lid ; l < lid + numLidsPerPort; l++ )
                    p_fabric->setLidPort(l, p_curr_port);
            }
        }
    }
    return(0);
}

int
SubnMgtAssignLids (IBPort *p_smNodePort, uint8_t lmc)
{
  list<IBPort *> thisStepPorts;
  list<IBPort *> nextStepNodePorts;
  set<IBPort *> visited;
  phys_port_t pn;
  IBFabric *p_fabric;
  IBPort *p_port;
  IBNode *p_node;
  lid_t numLidsPerPort = GET_LID_STEP(lmc);

  if (!p_smNodePort ||
      !p_smNodePort->p_node ||
      !p_smNodePort->p_node->p_fabric) {
    cout << "-E- Assign LIDs called with invalid object " << endl;
    return 1;
  }

  if ((p_smNodePort->p_node->type == IB_SW_NODE) &&
      (p_smNodePort->num != 0)) {
    cout << "-E- SM Port is on switch thus must be port 0!" << endl;
    return 1;
  }

  p_fabric = p_smNodePort->p_node->p_fabric;

  lid_t lid = (lid_t)(1<<lmc);
  int step = 0;
  p_fabric->minLid = lid;

  thisStepPorts.push_back(p_smNodePort);
  // NOTE: visited is marked when inserting to next step
  visited.insert(p_smNodePort);

  // BFS Style ...
  while (thisStepPorts.size() > 0) {
    nextStepNodePorts.clear();
    step++;

    // go over all this step ports
    while (! thisStepPorts.empty()) {
      p_port = thisStepPorts.front();
      thisStepPorts.pop_front();

      // get the node
      p_node = p_port->p_node;

      AssignNodeLids(p_port, lid, lmc, numLidsPerPort);
      // do not forget to increment the lids
      lid = (lid_t)(lid + numLidsPerPort);

      // based on the node type we do the recursion and assignment
      phys_port_t fPortNum, tPortNum;
      if (p_node->type != IB_SW_NODE) {
	// only consider this port
	fPortNum = p_port->num;
	tPortNum = p_port->num;
      } else {
	// for a switch consider all ports
	fPortNum = 1;
	tPortNum = p_node->numPorts;
      }

      // go over all ports of the node:
      for (pn = fPortNum; pn <= tPortNum; pn++) {
	IBPort *p_nPort = p_node->getPort(pn);
	if (!p_nPort || !p_nPort->p_remotePort) continue;

	// if we have a remote port that is not visited
	IBPort *p_remPort = p_nPort->p_remotePort;
	IBNode *p_remNode = p_remPort->p_node;
	IBPort *p_repRemNodePort = p_remPort;

	// The below assumes a switch node always have port 0
	if (p_remNode->type == IB_SW_NODE) {
	  p_repRemNodePort = p_remNode->getPort(0);
	  if (!p_repRemNodePort) {
	    cout << "SubnMgtAssignLids: BUG no port 0 for switch: " << p_remNode->name << endl;
	    exit(1);
	  }
	}
	// if the port representing the remote node was not visited and not included in
	if (visited.find(p_repRemNodePort) == visited.end()) {
	  visited.insert(p_repRemNodePort);
	  nextStepNodePorts.push_back(p_repRemNodePort);
	}
      }
    }
    thisStepPorts = nextStepNodePorts;
  }

  p_fabric->maxLid = (lid_t)(lid - 1);
  p_fabric->caLmc = lmc;
  p_fabric->swLmc = lmc;
  cout << "-I- Assigned " << p_fabric->maxLid - p_fabric->minLid + 1
       << " LIDs (lmc=" << (unsigned int)lmc
       << ") in " << step << " steps" << endl;
  return 0;
}


///////////////////////////////////////////////////////////////////////////////
// OpenSM style relaxation algorithm:
// First step is to mark your own or neighbors
// Loop updating from neighbors hops untill no update
int
SubnMgtCalcMinHopTables (IBFabric *p_fabric)
{
    IBNode *p_node;
    map_str_pnode::iterator nI;
    set_pnode::iterator swI;
    lid_t base_lid;
    lid_t lid;
    uint8_t lmc;

    // we should not assign hops for non SW nodes.
    // first step - update self on switches.
    for (nI = p_fabric->NodeByName.begin(); nI != p_fabric->NodeByName.end(); nI++) {
        p_node = (*nI).second;

        if (p_node->type == IB_SW_NODE) {
            p_node->getLidAndLMC(0, base_lid, lmc);
            for (lid = base_lid; lid < base_lid + ((lid_t)1 << lmc); lid++)
                p_node->setHops(NULL, lid, 0); // assign all ports value
        } else {
            // a non switch node might have connections on both ports
            // and also we just want to update the switch on the other side
            int conPorts = 0;
            for (unsigned int i = 1; i <= p_node->numPorts; i++) {
                IBPort *p_port = p_node->getPort((phys_port_t)i);
                if (p_port &&
                        p_port->p_remotePort &&
                        p_port->p_remotePort->p_node->type == IB_SW_NODE) {
                    p_node->getLidAndLMC(p_port->num, base_lid, lmc);
                    conPorts++;
                    // update the switch:
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-V- Setting switch:" << p_port->p_remotePort->p_node->name
                        << " as neighbor of CA port: " << p_port->getName()
                        << " lid: " << base_lid << endl;

                    for (lid = base_lid; lid < base_lid + ((lid_t)1<<lmc); lid++) {
                        p_port->p_remotePort->p_node->setHops(p_port->p_remotePort, lid, 1);
                    }
                }
            }
            if (!conPorts)
                cout << "-W- CA with no connected ports:" << p_node->name << endl;
        }
    }

    // second - loop until nothing to update:
    int anyUpdate;
    int loop = 0;
    do {
        loop++;
        anyUpdate = 0;

        // go over all switch nodes:
        for (swI = p_fabric->Switches.begin(); swI != p_fabric->Switches.end(); swI++) {
            p_node = *swI;

            // go over all lids (base) on this switch:
            for (lid_t lid = 1; lid <= p_fabric->maxLid; lid++) {
                // go over all connected ports
                for (phys_port_t pn = 1; pn <= p_node->numPorts; pn++) {
                    IBPort *p_port = p_node->getPort(pn);

                    // do we have a port on the other side ? is it a SW ?
                    if (p_port &&
                            p_port->p_remotePort &&
                            p_port->p_remotePort->p_node->type == IB_SW_NODE) {
                        // the min we have for this lid is:
                        int minHops = p_node->getHops(p_port, lid);

                        // we need to update the local port hops only if
                        // they will be made smaller by this step.
                        // i.e. the remote port hops value + 1 is < hops
                        uint8_t remNodeHops = p_port->p_remotePort->p_node->getHops(NULL, lid);

                        if (remNodeHops + 1 < minHops) {
                            // need to update:
                            p_node->setHops(p_port, lid, (uint8_t)(remNodeHops + 1));
                            anyUpdate++;
                        }
                    }
                }
            }
        }
        // cout << "-I- Propagated:" << anyUpdate << " updates" << endl;
    } while (anyUpdate);
    cout << "-I- Init Min Hops Tables in:" << loop << " steps" << endl;

    // simply check that one can reach to all lids from all switches:
    // also we build a historgram of the number of ports one can use
    // to get to any lid.
    vec_int numPathsHist(50,0);
    int anyUnAssigend = 0;
    for( nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE)
            continue;

        // go over all the lids.
        for (lid_t i = 1; i <= p_fabric->maxLid; i++ ) {
            // skip lids that are not mapped to a port:
            IBPort *port = p_fabric->getPortByLid(i);
            if (! port || (port->base_lid != i))
                continue;

            int minHops = p_node->getHops(NULL, i);
            if (minHops == IB_HOP_UNASSIGNED) {
                cout << "-W- Found - un-assigned hops for node:"
                        << p_node->name << " to lid:" << i << ")" << endl;
                anyUnAssigend++;
            } else {
                // count all ports that have min hops to that lid (only HCAs count)
                IBPort *p_targetPort = p_fabric->getPortByLid(i);
                if (p_targetPort && (p_targetPort->p_node->type != IB_SW_NODE)) {
                    int numMinHopPorts = 0;
                    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                        IBPort *p_port = p_node->getPort((phys_port_t)pn);
                        if (p_port && (p_node->getHops(p_port, i) == minHops))
                            numMinHopPorts++;
                    }
                    numPathsHist[numMinHopPorts]++;
                }
            }
        }
    }

    cout << "------------------ NUM ALTERNATE PORTS TO CA HISTOGRAM --------------------" << endl;
    cout << "Describes how many out ports on every switch have the same Min Hop to each " << endl;
    cout << "target CA. Or in other words how many alternate routes are possible at the " << endl;
    cout << "switch level. This is useful to show the symmetry of the cluster.\n" << endl;
    cout << "OUT-PORTS NUM-SW-LID-PAIRS" << endl;
    for (int b = 0; b < 50 ; b++)
        if (numPathsHist[b])
            cout << setw(4) << b << "   " << numPathsHist[b] << endl;
    cout << "---------------------------------------------------------------------------\n" << endl;

    if (anyUnAssigend) {
        cout << "-W- Found - un-reachable lids." << endl;
        return 1;
    }

    // report the worst case hops count found.
    int maxHops = 0;
    IBNode *p_worstHopNode  = NULL;
    IBPort *p_worstHopPort  = NULL;
    lid_t worstHopLid = 0;
    IBNode *p_caNode;
    IBPort *p_caPort;
    vec_int maxHopsHist(50,0);

    for( nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_caNode = (*nI).second;
        if (p_caNode->type != IB_CA_NODE)
            continue;

        // find the switch connected to the HCA port
        for (unsigned int n = 1; n <= p_caNode->numPorts; n++) {
            p_caPort = p_caNode->getPort((phys_port_t)n);
            // ignore the port if does not exist or not connected
            if (!p_caPort || !p_caPort->p_remotePort)
                continue;

            p_node = p_caPort->p_remotePort->p_node;

            // go over all lids found and get the min hop
            for (lid_t i = 1; i <= p_fabric->maxLid; i++) {
                // please ignore non CA lids and ourselves
                IBPort *p_port = p_fabric->getPortByLid(i);
                if (!p_port ||
		    (p_port->base_lid != i) ||
		    (p_port->p_node->type == IB_SW_NODE) ||
		    (p_caPort == p_port))
		  continue;

		int minHops = p_node->getHops(NULL, i);
		if (IB_HOP_UNASSIGNED != minHops) {
		  maxHopsHist[minHops]++;
		  if (minHops > maxHops) {
		    p_worstHopPort = p_node->getFirstMinHopPort(i);
		    p_worstHopNode = p_node;
		    worstHopLid = i;
		    maxHops = minHops;
		  }
		} else {
		  cout << "-W- Found - un-assigned hop for node:"
		       << p_node->name << " to port:" << p_port->p_node->name
		       << "/" << (unsigned int)p_port->num
		       << " (lid:" << i << ")" << endl;
		}
	    }
        }
    }

    // print the histogram:
    cout << "---------------------- CA to CA : MIN HOP HISTOGRAM -----------------------" << endl;
    cout << "The number of CA pairs that are in each number of hops distance." << endl;
    cout << "The data is based on topology only - even before any routing is run.\n" << endl;
    cout << "HOPS NUM-CA-CA-PAIRS" << endl;
    for (int b = 0; b <= maxHops ; b++)
        if (maxHopsHist[b])
            cout << setw(3) << b+1 << "   " << maxHopsHist[b] << endl;
    cout << "---------------------------------------------------------------------------\n" << endl;

    if (p_worstHopNode) {
        IBPort *p_worst_hop_lid_port = p_fabric->getPortByLid(worstHopLid);
        if (!p_worst_hop_lid_port)
            cout << "-E- fail to find port for worst hop lid:" << worstHopLid << endl;
        else
            cout << "-I- Found worst min hops:" << maxHops + 1 << " at node:"
                 << p_worstHopNode->name << " to node:"
                 << p_worst_hop_lid_port->p_node->name << endl;

        // the worst hop node lid
        TraceRouteByMinHops(p_fabric, p_worstHopPort->base_lid, worstHopLid);
    }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Fill in the FDB tables in an Extended OpesnSM style routing
// which is switch based, uses number of routes per port
// profiling and treat LMC assigned lids sequentialy
// Rely on running the SubnMgtCalcMinHopTables beforehand
// We added the notion of selecting the other system or node
// if same hop and profile
int
SubnMgtOsmEnhancedRoute(IBFabric *p_fabric)
{
    IBNode *p_node;
    cout << "-I- Using Enhanced OpenSM Routing" << endl;

    // we want to collect port subscriptions statistics:
    vec_int subscHist(10000,0);

    // also track the selections used:
    int numSelByOtherSys = 0;
    int numSelByOtherNode = 0;
    int numSelByMinSubsc = 0;

    // go over all nodes in the fabric
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;

        // if not a switch cont
        if (p_node->type != IB_SW_NODE)
            continue;

        // define port profiles
        vec_int portsSubscriptions(p_node->numPorts,0);

	// track previous port to know if on same port (in lmc)
	IBPort *pPrevPort = 0;

	// We track the Systems we already went through:
	set<IBSystem *> goThroughSystems;

	// we also track the nodes we went through:
	set <IBNode *> goThroughNodes;

	// go over all lids on this switch:
	for (lid_t lid = 1; lid <= p_fabric->maxLid; lid++) {
            int targetIsHCA;
            IBPort *pTargetPort = p_fabric->getPortByLid(lid);
            if (pTargetPort && (pTargetPort->p_node->type == IB_SW_NODE))
                targetIsHCA = 0;
            else
                targetIsHCA = 1;

            int minHop = p_node->getHops(NULL, lid);

	    // if the min is 0 just map to port 0
	    if (minHop == 0) {
	      p_node->setLFTPortForLid(lid, 0);
	      continue;
	    }

	    // if we change dest we clear the tracking
	    if (pTargetPort != pPrevPort) {
	      goThroughSystems.clear();
	      goThroughNodes.clear();
	    }
	    pPrevPort = pTargetPort;

	    // we are going to track the min profile under three
	    // possible cases:
	    int minSubsShared = 100000;
	    int minSubsDiffNodes = 100000;
	    int minSubsDiffSystems = 100000;
	    phys_port_t minPortNumShared = 0;
	    phys_port_t minPortNumDiffNodes = 0;
	    phys_port_t minPortNumDiffSystems = 0;
	    int minSubs;

	    if (minHop < 255) {
	      // look for the port with min profile
	      for (unsigned int i = 1; i <= p_node->numPorts; i++) {
              phys_port_t pn = (phys_port_t)i;
		IBPort *p_port = p_node->getPort(pn);
		if (! p_port)
		  continue;
		if (! p_port->p_remotePort)
		  continue;

		// the hops should match the min
		if (p_node->getHops(p_port,lid) == minHop) {
		  minSubs = portsSubscriptions[pn-1];
		  IBNode   *p_remNode = p_port->p_remotePort->p_node;
		  IBSystem *p_system = p_remNode->p_system;

		  if (goThroughSystems.find(p_system) == goThroughSystems.end()) {
		    if (minSubsDiffSystems > minSubs) {
		      minSubsDiffSystems = minSubs;
		      minPortNumDiffSystems = pn;
		    }
		  }
		  if (goThroughNodes.find(p_remNode) == goThroughNodes.end()) {
		    if (minSubsDiffNodes > minSubs) {
		      minSubsDiffNodes = minSubs;
		      minPortNumDiffNodes = pn;
		    }
		  }

		  if (minSubsShared > minSubs) {
		    minSubsShared = minSubs;
		    minPortNumShared = pn;
		  }
		} // hop = min hops
	      } // all ports

	      // we always select the system over node over shared:
	      if (minPortNumDiffSystems) {
		minPortNumShared = minPortNumDiffSystems;
		numSelByOtherSys++;
	      } else if (minPortNumDiffNodes) {
		numSelByOtherNode++;
		minPortNumShared = minPortNumDiffNodes;
	      } else {
		numSelByMinSubsc++;
	      }

	      // so now we need to have the port number or error
	      if (!minPortNumShared) {
		cout << "-E- Cound not find min hop port!" << endl;
		return(1);
	      }

	      // Track used systems and nodes
	      IBPort *p_bestPort = p_node->getPort(minPortNumShared);
	      if (!p_bestPort) {
	          cout << "-E- Best port is NULL!" << endl;
	          return(1);
	      }

	      IBNode *p_remNode = p_bestPort->p_remotePort->p_node;
	      IBSystem *p_system = p_node->p_system;

	      goThroughSystems.insert(p_system);
	      goThroughNodes.insert(p_remNode);

	    } else {
	      // there is no path to that lid...
	      minPortNumShared = 255;
	    }

	    // track subscriptions:
	    if (targetIsHCA)
	      portsSubscriptions[minPortNumShared-1]++;

	    // assign the fdb table.
	    p_node->setLFTPortForLid(lid, minPortNumShared);
	} // all lids

        // we want to get some histogram of subsriptions.
        for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
            IBPort *p_port = p_node->getPort((phys_port_t)pn);
            if (p_port && p_port->p_remotePort) {
                if (portsSubscriptions[pn-1] == 0) {
                    cout << "-W- Unused port:" << p_port->getName() << endl;
                }
                subscHist[portsSubscriptions[pn-1]]++;
            }
        }
    } // all nodes

    cout << "-I- Enhanced selection by Sys:" << numSelByOtherSys
            << " Node:" << numSelByOtherNode
            << " Subscription:" << numSelByMinSubsc << endl;
   return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Fill in the FDB tables in an OpesnSM style routing
// which is switch based, uses number of routes per port
// profiling and treat LMC assigned lids sequentialy
// Rely on running the SubnMgtCalcMinHopTables beforehand
int
SubnMgtOsmRoute(IBFabric *p_fabric)
{
  IBNode *p_node;
  cout << "-I- Using standard OpenSM Routing" << endl;

  // we want to collect port subscriptions statistics:
  vec_int subscHist(10000,0);

  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end();
       nI++) {
    p_node = (*nI).second;

    // if not a switch cont
    if (p_node->type != IB_SW_NODE)
      continue;

    // define port profiles
    vec_int portsSubscriptions(p_node->numPorts,0);

    // go over all valid lid values (i.e. base lids )
    for (lid_t lid = p_fabric->minLid; lid <= p_fabric->maxLid; lid++) {
      int targetIsHCA;
      IBPort *pTargetPort = p_fabric->getPortByLid(lid);
      if (pTargetPort && (pTargetPort->p_node->type == IB_SW_NODE))
        targetIsHCA = 0;
      else
        targetIsHCA = 1;

      // get the minimal hop count from this port:
      int minHop = p_node->getHops(NULL,lid);

      // if the min is 0 just map to port 0
      if (minHop == 0) {
        p_node->setLFTPortForLid(lid, 0);
        continue;
      }

      // initialize the min subsription to a huge number:
      int minSubsc = 100000;
      phys_port_t minSubsPortNum = 0;
#if 1
      // look for the port with min profile
      if (minHop != 255) {
        for (unsigned int i = 1; i <= p_node->numPorts; i++) {
            phys_port_t pn = (phys_port_t)i;
          IBPort *p_port = p_node->getPort(pn);

          if (! p_port)
            continue;

          // the hops should match the min
          if (p_node->getHops(p_port, lid) == minHop) {
            // Standard OpenSM Routing:
            // is it the lowest subscribed port:
            if (portsSubscriptions[pn-1] < minSubsc) {
              minSubsPortNum = pn;
              minSubsc = portsSubscriptions[pn-1];
            }
          }
        }
      } else {
	// it is possible to have "gaps" in LID assignment
        minSubsPortNum = 255;
      }
#else
      // do a random selection
      {
        vector<unsigned int> minHopOutPorts;
        for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
          IBPort *p_port = p_node->getPort((phys_port_t)pn);

          if (! p_port)
            continue;

          // the hops should match the min
          if (p_node->getHops(p_port, lid) == minHop) {
            minHopOutPorts.push_back(pn);
          }
        }
        double portRand = 1.0*minHopOutPorts.size()*rand()/RAND_MAX;
        unsigned int portIdx = int(portRand);
        minSubsPortNum = minHopOutPorts[portIdx];
      }
#endif
      // so now we need to have the port number or error
      if (!minSubsPortNum) {
        cout << "-E- Cound not find min hop port for lid:" << lid << " on node:" << p_node->name  << endl;
	p_node->repHopTable();
        return(1);
      }

      // track subscriptions only if target is not a switch:
      if (targetIsHCA)
        portsSubscriptions[minSubsPortNum-1]++;

      // assign the fdb table.
      p_node->setLFTPortForLid(lid, minSubsPortNum);
      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
	cout << "-V- Setting switch:" << p_node->name
	     <<  " LFT(" << lid << ") = " << (unsigned int)minSubsPortNum << endl;
    } // all lids

    // we want to get some histogram of subsriptions.
    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
      IBPort *p_port = p_node->getPort((phys_port_t)pn);
      if (p_port && p_port->p_remotePort) {
        if (portsSubscriptions[pn-1] == 0) {
          cout << "-W- Unused port:" << p_port->getName() << endl;
        }
        subscHist[portsSubscriptions[pn-1]]++;
      }
    }
  } // all nodes

  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Given a list of root nodes mark them with a zero rank
// Then BFS and rank min
// note we use the provided map of IBNode* to int for storing the rank
int
SubnRankFabricNodesByRootNodes(IBFabric *p_fabric, list_pnode rootNodes)
{
    list_pnode curNodes, nextNodes;
    curNodes = rootNodes;
    uint8_t rank = 0;

    // rank by zero the starting nodes:
    for (list_pnode::iterator nI = rootNodes.begin();
            nI != rootNodes.end();
            nI++) {
        IBNode *p_node = (*nI);
        p_node->rank = 0;
    }

    // ok so now we BFS
    while (curNodes.size()) {
        nextNodes.clear();
        rank++;

        // go over cur step nodes
        for (list_pnode::iterator lI = curNodes.begin();
                lI != curNodes.end(); lI++) {
            IBNode *p_node = *lI;
            // go over all ports to find unvisited remote nodes
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);
                if (! p_port)
                    continue;
                // might have remote port
                if (p_port->p_remotePort) {
                    // was it visited?
                    IBNode *p_remNode = p_port->p_remotePort->p_node;
                    if (p_remNode->rank == IBNODE_UNASSIGNED_RANK) {
                        // add it
                        nextNodes.push_back(p_remNode);
                        // mark it:
                        p_remNode->rank = (uint8_t)rank;
                    }
                }
            }
        }
        curNodes = nextNodes;
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-I- Max nodes rank=" << rank  << endl;
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Given a regular expression for nodes mark them with a zero rank
// Then BFS and rank min
// note we use the provided map of IBNode* to int for storing the rank
int
SubnRankFabricNodesByRegexp(IBFabric *p_fabric, const char * nodeNameRex)
{
    regExp nodeRex(nodeNameRex);
    rexMatch *p_rexRes;
    list_pnode rootNodes;

    // go over all nodes of the fabric;
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end(); nI++) {
        // match rex ?
        p_rexRes = nodeRex.apply((*nI).first.c_str());
        if (p_rexRes) {
            cout  << "-I- Starting UpDown Routing from node:"
                    << (*nI).second->name << endl;
            rootNodes.push_back((*nI).second);
            delete p_rexRes;
        }
    }
    return SubnRankFabricNodesByRootNodes(p_fabric, rootNodes);
}


///////////////////////////////////////////////////////////////////////////////
// Clasify the routes for the same port to port
// Given two list of nodes provide the number of shared
// Systems and Shared nodes:
int
SubnFindPathCommonality(vec_pnode *p_path1, vec_pnode *p_path2,
                        vec_phys_ports *p_portsList1,
                        vec_phys_ports *p_portsList2,
                        int *commonSystems, int *commonNodes,
                        int *commonLinks)
{
    map_pnode_int nodesIntersection;
    map_pport_int portsIntersection;
    map_psystem_int systemIntersection;
    IBSystem *p_system;
    IBNode *p_node;
    IBPort *p_port;

    // we got it now:
    *commonNodes = 0;
    *commonSystems = 0;
    *commonLinks = 0;

    vec_phys_ports::const_iterator portsI = p_portsList1->begin();
    int numPorts = (int)p_portsList1->size();
    int curr_port = 1;
    // Go over all nodes in first path add them to the map:
    for (vec_pnode::const_iterator lI = p_path1->begin();
            lI != p_path1->end(); lI++) {
        p_node = *lI;
        nodesIntersection[p_node] = 1;
        p_system = p_node->p_system;
        systemIntersection[p_system] = 1;

        p_port = p_node->getPort(*portsI);
        if (p_port && curr_port < numPorts)
            portsIntersection[p_port] = 1;
        ++portsI;
        curr_port++;
    }

    // We only need to count the number of pre-existing nodes.
    // we do it once for each nodes.
    portsI = p_portsList2->begin();
    for (vec_pnode::const_iterator lI = p_path2->begin();
            lI != p_path2->end(); lI++) {
        // we only count once for each node.
        p_node = *lI;
        map_pnode_int::iterator mI = nodesIntersection.find(p_node);
        if ( (mI != nodesIntersection.end()) &&
                ((*mI).second == 1) ) {
            (*commonNodes)++;
            // we increase it for next time
            (*mI).second++;
        }
        if (portsI != p_portsList2->end()) {
            p_port = p_node->getPort(*portsI);
            map_pport_int::iterator mpI = portsIntersection.find(p_port);
            if ( (mpI != portsIntersection.end()) &&
                    ((*mpI).second == 1) ) {
                (*commonLinks)++;
                // we increase it for next time
                (*mpI).second++;
            }
            ++portsI;
        }

        p_system = p_node->p_system;
        map_psystem_int::iterator sI = systemIntersection.find(p_system);
        if ( (sI != systemIntersection.end()) &&
                ((*sI).second == 1) ) {
            (*commonSystems)++;
            // we increase it for next time
            (*sI).second++;
        }
    }

    //cout << "P1:" << p_path1->size()
    //  << " P2:"<< p_path2->size()
    //  << " Intersaction:" << *commonNodes << "/" << *commonSystems<<endl;
    return(0);
}

#define MAX_COMMON_SYSTEMS_REPORT 10
#define MAX_COMMON_LINKS_REPORT   10

///////////////////////////////////////////////////////////////////////////////
// Verify point to point connectivity
int
SubnMgtVerifyAllCaToCaRoutes(IBFabric *p_fabric, const char *outDir)
{
  bool anyLmcUsed = false;
  bool anyCommonLinks = false;

  int anyError = 0, paths = 0;
  vec_int maxHopsHist(50,0);
  // we want to track common Nodes and Systems
  // based on the original path length
  int CommonNodes[50][16];
  int CommonSystems[50][16];
  int CommonLinks[50][16];

  unsigned int maxDepth = 0;
  int maxLinkSubscriptions = 0;
  int maxDlidPerOutPort = 0;
  int commonSystemsReport = 0;
  int commonLinksReport = 0;

  cout << "-I- Verifying all CA to CA paths ... " << endl;

  // initialize the histograms:
  memset(CommonNodes, 0 , 50*16 * sizeof(int));
  memset(CommonSystems, 0 , 50*16 * sizeof(int));
  memset(CommonLinks, 0, 50*16 * sizeof(int));

  unsigned int hops, maxHops = 0;
  vec_pnode path1, path2, *p_path;
  vec_phys_ports portsList1, portsList2, *p_portsList;

  // NOTE: since it is OK to have just one of the DLIDs assigned to a single
  // port be available as path from any given source we do not complain
  // about dead paths unless all of them from sPort to dPort are not avaiable

  // go over all ports in the fabric
  for (lid_t dLid = p_fabric->minLid; dLid <= p_fabric->maxLid; dLid++) {

    IBPort *p_dstPort = p_fabric->getPortByLid(dLid);
    IBVPort *p_dstVPort = NULL;
    lid_t pathsPerDlid = 1;

    if (p_dstPort) {
      if (p_dstPort->base_lid != dLid || p_dstPort->p_node->type == IB_SW_NODE) continue;

      // go over all LMC values of the destination:
      pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
      if (pathsPerDlid > 1)
        anyLmcUsed = true;
    } else { // check if the dlid is a vport lid
        p_dstVPort = p_fabric->getVPortByLid(dLid);
        if (!p_dstVPort) continue;
    }

    // go over all the rest of the ports - but only once - so skip non base lids
    for (lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++ ) {
      IBPort *p_srcPort = p_fabric->getPortByLid(sLid);

      if (!p_srcPort || p_srcPort->base_lid != sLid)
        continue;

      if (!p_srcPort->p_node ||
            p_srcPort->p_node->skipRoutingChecks)
        continue;

      // Avoid tracing to itself
      if (sLid == dLid)
        continue;
      if (p_srcPort->p_node->type == IB_SW_NODE)
        continue;
      if (p_dstVPort && p_dstVPort->getIBPortPtr()->base_lid == sLid)
        continue;

      bool anyPathToDest = false;

      // go over all LMC combinations:
      for (lid_t l = 0; l < pathsPerDlid; l++) {
        paths++;

        // we track the path nodes in lists but need to know which one
        // to use:
        if (l == 0) {
          p_path = &path1;
          p_portsList = &portsList1;
        } else {
          p_path = &path2;
          p_portsList = &portsList2;
        }
        p_path->clear();
        p_portsList->clear();

        // now go and verify the path:
        if (!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l), &hops,
                                              p_path, p_portsList, false)) {
          anyPathToDest = true;
          // track the hops histogram
          maxHopsHist[hops]++;
          if (hops > maxHops) maxHops = hops;

          // populate the number of the out ports along the path
          // but ignore the last element
          vec_phys_ports::const_iterator pI = p_portsList->begin();
          for (vec_pnode::const_iterator lI = p_path->begin();
               lI != p_path->end() && pI != p_portsList->end() ; lI++, pI++) {

            // init the ports count vector if needed
            IBNode *pNode = (*lI);
            if(!pNode->routing_cache.paths.size())
            {
                vec_int tmp(pNode->numPorts + 1,0);
                pNode->routing_cache.paths.swap(tmp);
            }

            // init the marking of any path through the port if needed:
            if(!pNode->routing_cache.any_paths.size())
            {
                vec_int tmp(pNode->numPorts + 1,0);
                pNode->routing_cache.any_paths.swap(tmp);
            }


            vec_pnode::const_iterator nlI = lI;
            nlI++;
            if (nlI != p_path->end()) {
              unsigned int outPort = *pI;
              int v = ++pNode->routing_cache.paths[outPort];

              if (maxLinkSubscriptions < v)
                maxLinkSubscriptions = v;

              ++pNode->routing_cache.any_paths[outPort];
            }
          }

          // Analyze the path against the previous path:
          if (l != 0) {
            static int commonSystems, commonNodes, commonLinks;

            // we only care about paths longer then 2 hops since the two switches
            //  must be identical
            if (path1.size()) {
              SubnFindPathCommonality(&path1, &path2,
                                      &portsList1, &portsList2,
                                      &commonSystems, &commonNodes, &commonLinks);


              // track the max depth
              if (path1.size() > maxDepth)
                  maxDepth = (unsigned int)path1.size();

              if (maxDepth > 15) {
                cout << "-E- Found a path length > 15. Need to recompile with larger"
                     << " Histogram size!" << endl;
                exit(1);
              }

              if (!path1.size()) {
                cout << "-W- Zero size path1 ???" << endl;
                continue;
              }

              // store the statistics:
              CommonSystems[commonSystems][0]++;
              CommonSystems[commonSystems][path1.size()]++;
              CommonNodes[commonNodes][0]++;
              CommonNodes[commonNodes][path1.size()]++;
              if(commonLinks) {
                  CommonLinks[commonLinks][0]++;
                  CommonLinks[commonLinks][path1.size()-1]++;
                  anyCommonLinks = true;

                  if(commonLinksReport < MAX_COMMON_LINKS_REPORT) {
                    cout << "---- COMMON LINKS PATH ----- " << endl;
                    cout << "From:"
                         << p_srcPort->p_node->name << "/" << (int) p_srcPort->num
                         << " to:" << p_dstPort->p_node->name << "/" << (int) p_dstPort->num
                         << endl;
                    cout << "Path 1 (LID " << dLid << ")" << endl;
                    vec_phys_ports::iterator portI = portsList1.begin();
                    for (vec_pnode::iterator lI = path1.begin();
                         lI!= path1.end();++lI, ++portI)
                      cout << "." << (*lI)->name.c_str() <<
                        "/" << (int) *portI << endl;

                    cout << "Path 2 (LID " << dLid + l << ")" << endl;
                    portI = portsList2.begin();
                    for (vec_pnode::iterator lI = path2.begin();
                         lI != path2.end();
                         ++lI, ++portI)
                      cout << "." << (*lI)->name.c_str() <<
                        "/" << (int) *portI << endl;

                    commonLinksReport++;
                  } else if (commonLinksReport == MAX_COMMON_LINKS_REPORT) {
                      cout << "... COMMON LINKS PATH TRUNCATED "<<
                          "(more than " << (int) MAX_COMMON_LINKS_REPORT <<
                          ")." << endl;
                      commonLinksReport++;
                  }


              }

              if (commonSystems > 5) {
                  if(commonSystemsReport < MAX_COMMON_SYSTEMS_REPORT) {
                    cout << "---- MORE THEN 5 COMMON SYSTEMS PATH ----- " << endl;
                    cout << "From:"
                         << p_srcPort->p_node->name << "/" << (int) p_srcPort->num
                         << " to:" << p_dstPort->p_node->name << "/" << (int) p_dstPort->num
                         << endl;
                    cout << "Path 1 (LID " << dLid << ")" << endl;
                    for (vec_pnode::iterator lI = path1.begin();
                         lI!= path1.end();lI++)
                      cout << "." << (*lI)->name.c_str() << endl;

                    cout << "Path 2 (LID " << dLid + l << ")" << endl;
                    for (vec_pnode::iterator lI = path2.begin();
                         lI != path2.end();
                         lI++ )
                      cout << "." << (*lI)->name.c_str() << endl;

                    commonSystemsReport++;
                  } else if (commonSystemsReport == MAX_COMMON_SYSTEMS_REPORT) {
                      cout << "... MORE THEN 5 COMMON SYSTEMS PATH TRUNCATED " <<
                          "(more than " << (int) MAX_COMMON_SYSTEMS_REPORT <<
                          ")." << endl;
                      commonSystemsReport++;
                  }

              }
            }
            // cleanup :
            // path2.clear();
          }

        } // route exists

      } // all lmc values

      if (!anyPathToDest) {
          if (anyError < 1000) {
            cout << "-E- Fail to find a path from:"
                 << p_srcPort->p_node->name << "/"
                 << (unsigned int)p_srcPort->num;
            if (p_dstPort) {
                cout << " to:" << p_dstPort->p_node->name << "/"
                     << (unsigned int)p_dstPort->num;
            } else { // dest is vport
                cout << " to " << p_dstVPort->getIBPortPtr()->p_node->name << "/"
                     << (unsigned int)p_dstVPort->getIBPortPtr()->num
                     << ", vnode: " << p_dstVPort->m_p_vnode->getDescription()
                     << ", vport: " << p_dstVPort->getVPortNum();
            }
            cout << endl;
          }
        anyError++;
      }

      //isPLFTEnabled check VL15 connectivity
      if (p_fabric->isPLFTEnabled()) {
          anyPathToDest = false;
          for (lid_t l = 0; l < pathsPerDlid; l++) {
              if(!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l),
                                               NULL, NULL, NULL, true))
                  anyPathToDest = true;
          }
          if (!anyPathToDest) {
            cout << "-E- Fail to find a VL15 path from:"
                 << p_srcPort->p_node->name << "/"
                 << (unsigned int)p_srcPort->num;
            if (p_dstPort) {
                cout << " to:" << p_dstPort->p_node->name << "/"
                     << (unsigned int)p_dstPort->num;
            } else { // dest is vport
                cout << " to " << p_dstVPort->getIBPortPtr()->p_node->name << "/"
                     << (unsigned int)p_dstVPort->getIBPortPtr()->num
                     << ", vnode: " << p_dstVPort->m_p_vnode->getDescription()
                     << ", vport: " << p_dstVPort->getVPortNum();
            }
            cout << endl;
            anyError++;
          }
      }
    } // all src lids

    // add to dlid per port vector
    for(set_pnode::iterator itr = p_fabric->Switches.begin();
        itr != p_fabric->Switches.end(); ++itr)
    {
        IBNode *pNode = *itr;
        if(pNode->routing_cache.any_paths.size())
        {
            // init if required
            if(!pNode->routing_cache.lids.size())
            {
                vec_int tmp(pNode->numPorts + 1,0);
                pNode->routing_cache.lids.swap(tmp);
            }

            for (unsigned int i = 1; i <= pNode->numPorts; i++)
                if (pNode->routing_cache.any_paths[i])
                {
                    int v = ++pNode->routing_cache.lids[i];
                    if (v > maxDlidPerOutPort)
                        maxDlidPerOutPort = v;

                }
            // clear any_paths for next dlid iteration
            pNode->routing_cache.any_paths.clear();
        }
    }
  } // all dlids

  // cleanup TraceRouteByLFT output lists
  // (they are also cleaned before every call to TraceRouteByLFT)
  path1.clear();
  path2.clear();
  portsList1.clear();
  portsList2.clear();

  cout << "---------------------- CA to CA : LFT ROUTE HOP HISTOGRAM -----------------" << endl;
  cout << "The number of CA pairs that are in each number of hops distance." << endl;
  cout << "This data is based on the result of the routing algorithm.\n" << endl;
  cout << "HOPS NUM-CA-CA-PAIRS" << endl;
  for (int b = 0; b <= (int)maxHops ; b++)
    if (maxHopsHist[b])
      cout << setw(3) << b+1 << "   " << maxHopsHist[b] << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;

  if (anyLmcUsed) {

    if (anyCommonLinks) {
        cout << "------------------ LMC BASED ROUTING :COMMON LINKS HISTOGRAM -----------------" << endl;

        cout << "Describes the distribution of the number of common links between the " << endl;
        cout << "different LMC paths of all the CA to CA paths," << endl;
        cout << "not including the first and last links (from/to the CA).\n"
            << endl << endl;
        cout << "PATH HOPS|";
        unsigned int maxLinks = maxDepth - 1;

        for (unsigned int d = 1; d <= maxLinks; d++)
          cout <<  setw(6) << d << "|";
        cout << endl;

        for (unsigned int b = 0; b <= maxLinks ; b++)
          if (CommonLinks[b][0] != 0) {
            cout << "COMM="<< setw(4) << b << "|";
            for (unsigned int d = 1; d <= maxLinks; d++)
              cout << setw(6) << CommonLinks[b][d] << "|";
            cout << endl;
          }
        cout << "---------------------------------------------------------------------------\n" << endl;
    }


    cout << "------------------ LMC BASED ROUTING :COMMON NODES HISTOGRAM -----------------" << endl;

    cout << "Describes the distribution of the number of common nodes between the " << endl;
    cout << "different LMC paths of all the CA to CA paths.\n" << endl;
    cout << "COMMON-NODES NUM-CA-CA-PAIRS" << endl;
    cout << "PATH DPT|";
    for (unsigned int d = 1; d <= maxDepth; d++)
      cout <<  setw(6) << d << "|";
    cout << endl;

    for (unsigned int b = 0; b <= maxHops ; b++)
      if (CommonNodes[b][0] != 0) {
        cout << "COMM="<< setw(3) << b << "|";
        for (unsigned int d = 1; d <= maxDepth; d++)
          cout << setw(6) << CommonNodes[b][d] << "|";
        cout << endl;
      }
    cout << "---------------------------------------------------------------------------\n" << endl;

    cout << "---------------- LMC BASED ROTING :COMMON SYSTEMS HISTOGRAM ---------------" << endl;
    cout << "The distribution of the number of common systems between the " << endl;
    cout << "different LMC paths of all the CA to CA paths.\n" << endl;
    cout << "COMMON-SYSTEM NUM-CA-CA-PAIRS" << endl;
    cout << "PATH DPT|";
    for (unsigned int d = 1; d <= maxDepth; d++)
      cout <<  setw(6) << d << "|";
    cout << endl;
    for (unsigned int b = 0; b <= maxHops ; b++)
      if (CommonSystems[b][0] != 0) {
        cout << "COMM=" << setw(3) << b << "|";
        for (unsigned int d = 1; d <= maxDepth; d++)
          cout << setw(6) << CommonSystems[b][d] << "|";
        cout << endl;
      }
    cout << "---------------------------------------------------------------------------\n" << endl;
  }

  string outDirStr;
  if (outDir == NULL)
      outDirStr = string(SEPERATOR) + "tmp" + SEPERATOR;
  else
      outDirStr = string(outDir);

  //---------------------------------------------------------------------
  string err_message;
  // report the link over subscription histogram and dump out the
  // num paths per switch out port
  string fileName_linkUsage = outDirStr + "ibdm.sw_out_port_num_paths";
  ofstream linkUsage;
  int rc_linkUsage = IBFabric::OpenFile(fileName_linkUsage.c_str(), linkUsage, false, err_message);
  if (rc_linkUsage)
      cout << err_message << endl;
  else
      linkUsage << "# NUM-PATHS PORT-NAME " << endl;

  vec_int linkSubscriptionHist(maxLinkSubscriptions + 1,0);

  //---------------------------------------------------------------------

  // now do the DLID per out port:
  string fileName_portDlidsUsage = outDirStr + "ibdm.sw_out_port_num_dlids";
  ofstream portDlidsUsage; //(outDirStr.c_str());
  int rc_portDlidsUsage = IBFabric::OpenFile(fileName_portDlidsUsage.c_str(), portDlidsUsage, false, err_message);
  if (rc_portDlidsUsage)
      cout << err_message << endl;
  else
       portDlidsUsage << "# NUM-DLIDS PORT-NAME " << endl;

  vec_int dlidsSubscriptionHist(maxDlidPerOutPort + 1,0);
  //---------------------------------------------------------------------



  for(set_pnode::iterator itr = p_fabric->Switches.begin();
      itr != p_fabric->Switches.end(); ++itr) {
    IBNode *pNode = *itr;
    bool pathes_cond = pNode->routing_cache.paths.size();
    bool lids_cond = pNode->routing_cache.lids.size();

    if(pathes_cond || lids_cond)
      for (unsigned int pn = 1; pn <= pNode->numPorts; pn++) {
        IBPort *pPort = pNode->getPort((phys_port_t)pn);
        if (pPort && pPort->p_remotePort &&
           (pPort->p_remotePort->p_node->type == IB_SW_NODE)) {
             if (pathes_cond) {
                 if (!rc_linkUsage)
                    linkUsage << setw(6) << pNode->routing_cache.paths[pn]  << " " << pPort->getName() << endl;
                 linkSubscriptionHist[pNode->routing_cache.paths[pn]]++;
             }
             if(lids_cond) {
                 if (!rc_portDlidsUsage)
                   portDlidsUsage << setw(6) << pNode->routing_cache.lids[pn]  << " " << pPort->getName() << endl;
                 dlidsSubscriptionHist[pNode->routing_cache.lids[pn]]++;
             }
        }
      }
  }

  if (!rc_linkUsage) linkUsage.close();
  if (!rc_portDlidsUsage) portDlidsUsage.close();

  cout << "---------- LFT CA to CA : SWITCH OUT PORT - NUM PATHS HISTOGRAM -----------" << endl;
  cout << "Number of actual paths going through each switch out port considering" << endl;
  cout << "all the CA to CA paths. Ports driving CAs are ignored (as they must" << endl;
  cout << "have = Nca - 1). If the fabric is routed correctly the histogram" << endl;
  cout << "should be narrow for all ports on same level of the tree." << endl;
  cout << "A detailed report is provided in: " << outDirStr << "ibdm.sw_out_port_num_paths \n" << endl;
  cout << "NUM-PATHS NUM-SWITCH-PORTS" << endl;
  for (int b = 0; b <= maxLinkSubscriptions; b++)
    if (linkSubscriptionHist[b])
      cout << setw(8) << b << "   " << linkSubscriptionHist[b] << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;

  cout << "---------- LFT CA to CA : SWITCH OUT PORT - NUM DLIDS HISTOGRAM -----------" << endl;
  cout << "Number of actual Destination LIDs going through each switch out port considering" << endl;
  cout << "all the CA to CA paths. Ports driving CAs are ignored (as they must" << endl;
  cout << "have = Nca - 1). If the fabric is routed correctly the histogram" << endl;
  cout << "should be narrow for all ports on same level of the tree." << endl;
  cout << "A detailed report is provided in: " << outDirStr << "ibdm.sw_out_port_num_dlids \n" << endl;
  cout << "NUM-DLIDS NUM-SWITCH-PORTS" << endl;
  for (int b = 0; b <= maxDlidPerOutPort; b++)
    if (dlidsSubscriptionHist[b])
      cout << setw(8) << b << "   " << dlidsSubscriptionHist[b] << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;

  if (anyError)
    cout << "-E- Found " << anyError << " missing paths"
         << " out of:" << paths << " paths" << endl;
  else
    cout << "-I- Scanned:" << paths << " CA to CA paths " << endl;

  cout << "---------------------------------------------------------------------------\n" << endl;
  return anyError;
}

///////////////////////////////////////////////////////////////////////////////
int
SubnMgtVerifyAllRoutes(IBFabric *p_fabric)
{
    int anyError = 0, paths = 0;

    cout << "-I- Verifying all paths ... " << endl;
    unsigned int hops, maxHops = 0;
    //list_pnode path;

    // go over all ports in the fabric
    for (lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++) {
      IBPort *p_srcPort = p_fabric->getPortByLid(sLid);

      if (!p_srcPort || (p_srcPort->base_lid != sLid))
	continue;

      // go over all the rest of the ports:
      for (lid_t dLid = p_fabric->minLid; dLid <= p_fabric->maxLid; dLid++) {
	IBPort *p_dstPort = p_fabric->getPortByLid(dLid);

	if (!p_dstPort || p_dstPort->base_lid != dLid)
	  continue;

	// Avoid tracing to itself
	if (dLid == sLid)
	  continue;

	bool anyPathToDest = false;
	for (lid_t l = 0; l < ((lid_t)1<<p_dstPort->lmc); l++) {
	  paths++;

	  // now go and verify the path:
	  if (!TraceRouteByLFT(p_fabric, sLid, (lid_t)(dLid + l), &hops,
                           NULL, NULL, false)) {
	    if (hops > maxHops)
	      maxHops = hops;
	    anyPathToDest = true;
	  }
	  //path.clear();
	}
	if (!anyPathToDest) {
	  cout << "-E- Fail to find a path from:"
	       << p_srcPort->p_node->name << "/"
           << (unsigned int)p_srcPort->num
	       << " to:" << p_dstPort->p_node->name << "/"
           << (unsigned int)p_dstPort->num
	       << endl;
	  anyError++;
	}
    //isPLFTEnabled check VL15 connectivity
    if (p_fabric->isPLFTEnabled()) {
        anyPathToDest = false;
        lid_t pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
        for (lid_t l = 0; l < pathsPerDlid; l++) {
            if(!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l),
                                             NULL, NULL, NULL, true))
                anyPathToDest = true;
        }
        if (!anyPathToDest) {
          cout << "-E- Fail to find a VL15 path from:"
               << p_srcPort->p_node->name << "/"
               << (unsigned int)p_srcPort->num
               << " to:" << p_dstPort->p_node->name << "/"
               << (unsigned int)p_dstPort->num
               << endl;
          anyError++;
        }
    }

      }
    }

    if (anyError)
        cout << "-E- Found " << anyError << " missing paths"
               << " out of:" << paths << " paths" << endl;
    else
        cout << "-I- Scanned:" << paths << " paths " << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;
    return anyError;
}


///////////////////////////////////////////////////////////////////////////////
// Analyze the fabric to see if it we can recognize the root nodes
// Return the list of root nodes found.
// We limit our auto root recognition too symmetrical fat trees only.
// in this case every switch should either connect to level N-1 or N+1.
// We use this and check for it during the BFS.
typedef list <IBNode *> list_pnode;

list_pnode
SubnMgtFindTreeRootNodes(IBFabric *p_fabric)
{
    list_pnode nextNodes;
    list_pnode curNodes;
    list_pnode rootNodes;
    list_pnode emptyRes;
    IBNode *p_node;
    int rank = 0;

    cout << "-I- Automatically recognizing the tree root nodes ..."
            <<  endl;

    // find all non switch nodes and add them to the current step nodes
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;
        if (p_node->type == IB_SW_NODE)
            continue;

        curNodes.push_back(p_node);
    }

    // BFS:
    while (! curNodes.empty()) {
        rank++;
        nextNodes.clear();

        // the last group should be our roots nodes.
        rootNodes = curNodes;

        if (0) {
            cout << "-V- Level:" << rank << " nodes:" << endl;
            int nInLine = 1;
            for (list_pnode::iterator nI = curNodes.begin();
                    nI != curNodes.end();
                    nI++) {
                if (nInLine == 16) {
                    nInLine = 0;
                    cout << endl;
                }
                nInLine++;
                cout << (*nI)->name << " ";
            }
        }

        // go over all this step ports
        while (! curNodes.empty()) {
            p_node = curNodes.front();
            curNodes.pop_front();

            // go over all ports of the node
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);

                // if there is a connection:
                if (p_port && p_port->p_remotePort) {
                    IBNode *p_remNode = p_port->p_remotePort->p_node;

                    // we ignore non SW nodes:
                    if (p_remNode->type != IB_SW_NODE)
                        continue;

                    // if already marked
                    if (p_remNode->rank == IBNODE_UNASSIGNED_RANK) {
                        int remNodeRank = p_remNode->rank;

                        // as we use the rank for signing visited status
                        // we might have the remote node be marked
                        // either with rank - 1 or rank + 1
                        if ((remNodeRank != rank + 1) && (remNodeRank != rank - 1)) {
                            cout << "-E- Given topology is not a pure levelized tree:" << endl;
                            cout << "    Node:" << p_remNode->name << " rank:" << remNodeRank
                                    << " accessed from node:" << p_node->name
                                    << " rank:" << rank << endl;
                            return emptyRes;
                        }
                    } else {
                        // first we mark the node as rank + 1
                        p_remNode->rank = (rank_t)(rank + 1);

                        // now we push it to the next level list:
                        nextNodes.push_back(p_remNode);
                    }
                }
            }
        }
        curNodes = nextNodes;
    }
    return rootNodes;
}


///////////////////////////////////////////////////////////////////////////////
// This routine is based on the min hop tables and
// the fact the statistics of the min hops changes for the
// root nodes:
typedef vector< int > vec_int;

list_pnode
SubnMgtFindRootNodesByMinHop(IBFabric *p_fabric)
{
    list_pnode rootNodes;
    unsigned int minHop;
    int numCas = 0;

    // go over all switch nodes and print the statistics:
    cout << "-I- Automatically recognizing the tree root nodes ..." <<  endl;

    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE) numCas++;
    }

    // find all non switch nodes and add them to the current step nodes
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE)
            continue;

        // the min hop table should exist on the node:
        unsigned int maxHops = 0;

        // go over all nodes that are CA and calc the histogram of min hops to cas
        vec_int swToCaMinHopsHist(50, 0);

        // go over all ports in the fabric
        for (lid_t bLid = p_fabric->minLid; bLid <= p_fabric->maxLid; bLid++ ) {
            IBPort *p_port = p_fabric->getPortByLid(bLid);

            if (!p_port || (p_port->p_node->type == IB_SW_NODE))
                continue;

	    if (bLid != p_port->base_lid)
	      continue;

            // get the min hops to this port:
            minHop = p_node->getHops(NULL, bLid);

            swToCaMinHopsHist[minHop]++;

            // track the max:
            if (minHop > maxHops) maxHops = minHop;
        } // all lids

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
            // print out the hist for now:
            cout << " CA MIN HOP HISTOGRAM:" <<  p_node->name;
            for (unsigned int b = 0; b <= maxHops ; b++)
                cout << " " << setw(4) << swToCaMinHopsHist[b] ;
            cout << endl;
        }

        // we recognize spines by requiring one bar to be above 90% of the
        // number of CAs
        int numHopBarsOverThd1 = 0;
        int numHopBarsOverThd2 = 0;
        double thd1 = numCas * 0.9;
        double thd2 = numCas * 0.05;
        for (unsigned int b = 0; b <= maxHops ; b++) {
            if (swToCaMinHopsHist[b] > thd1) numHopBarsOverThd1++;
            if (swToCaMinHopsHist[b] > thd2) numHopBarsOverThd2++;
        }
        if ((numHopBarsOverThd1 == 1) && (numHopBarsOverThd2 == 1))
            rootNodes.push_back(p_node);
    } // all switches
    return rootNodes;
}

///////////////////////////////////////////////////////////////////////////////
// Find any routes that exist in the FDB's from CA to CA and do not adhare to
// the up/down rules. Report any crossing of the path.
int
SubnReportNonUpDownCa2CaPaths(IBFabric *p_fabric)
{
  vec_pnode path;
  vec_phys_ports portsList;

  int anyError = 0, paths = 0, numBadPaths = 0;
  unsigned int hops;
  string firstChangeMsg;

  cout << "-I- Tracing all CA to CA paths for Credit Loops potential ..." << endl;

  // go over all ports in the fabric
  for (lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++) {
    IBPort *p_srcPort = p_fabric->getPortByLid(sLid);

    // avoid too many errors...
    if (numBadPaths > 100)
      break;

    if (!p_srcPort || (p_srcPort->p_node->type == IB_SW_NODE))
      continue;

    if (sLid != p_srcPort->base_lid)
      continue;

    // go over all the rest of the ports:
    for (lid_t dLid = p_fabric->minLid; dLid <= p_fabric->maxLid; dLid++) {
      IBPort *p_dstPort = p_fabric->getPortByLid(dLid);
      if (!p_dstPort || (p_dstPort->p_node->type == IB_SW_NODE))
        continue;
      if (dLid != p_dstPort->base_lid)
        continue;

      // avoid too many errors...
      if (numBadPaths > 100)
        break;

      // Avoid tracing to itself
      if (dLid == sLid)
        continue;

      paths++;
      // go over all possible dLids
      bool anyPathToDest = false;
      for (lid_t pb = 0; pb < (lid_t)(1 << p_dstPort->lmc); pb++) {
        // now go and verify the path:
        if (!TraceRouteByLFT(p_fabric, sLid, (lid_t)(dLid+pb),
                                              &hops, &path, &portsList, false)) {
          int numChanges = 0;
          int prevGoingUp = 1, goingUp;
          int prevRank = 99, rank;
          IBNode *p_prevNode = NULL;
          anyPathToDest = true;

	  // Go through the path and check for up down transitions.
	  for (vec_pnode::iterator lI = path.begin(); lI!= path.end(); lI++) {
	    IBNode *p_node = *lI;

	    if (!IsRankAssigned(p_node))
            exit(1);

	    rank = p_node->rank;
	    // we are going up if
	    if (rank < prevRank)
	      goingUp = 1;
	    else
	      goingUp = 0;

	    // now look for dir change.
	    if (prevGoingUp != goingUp) {
	      // we need to report if not first change.
	      if (numChanges) {
		// header only once:
		if (numChanges == 1) {
		  cout << "-E- Potential Credit Loop on Path from:"
		       << p_srcPort->p_node->name << "/" << p_srcPort->num
		       << " to:" << p_dstPort->p_node->name
		       << "/" << p_dstPort->num
		       << endl;
		  cout << firstChangeMsg << endl;
		  numBadPaths++;
		}
		// now the point
		if (goingUp) {
		  cout << "  Going:Up ";
		} else {
		  cout << "  Going:Down ";
		}
		cout << "from:" << p_prevNode->name
		     << " to:" << p_node->name << endl;
	      } else {
		// first change so:
		firstChangeMsg = string("  Going:Down from:") +
		  p_prevNode->name + string(" to:") + p_node->name;
	      }
	      numChanges++;
	    }
	    prevRank = rank;
	    prevGoingUp = goingUp;
	    p_prevNode = p_node;
	  }
	}
        path.clear();
        portsList.clear();

      }

      if (!anyPathToDest) {
        cout << "-E- Fail to find a path from:"
             << p_srcPort->p_node->name << "/"
             << (unsigned int)p_srcPort->num
             << " to:" << p_dstPort->p_node->name << "/"
             << (unsigned int)p_dstPort->num
             << endl;
        anyError++;
      }
      //isPLFTEnabled check VL15 connectivity
      if (p_fabric->isPLFTEnabled()) {
          anyPathToDest = false;
          lid_t pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
          for (lid_t l = 0; l < pathsPerDlid; l++) {
              if(!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l),
                                               NULL, NULL, NULL, true))
                  anyPathToDest = true;
          }
          if (!anyPathToDest) {
            cout << "-E- Fail to find a VL15 path from:"
                 << p_srcPort->p_node->name << "/"
                 << (unsigned int)p_srcPort->num
                 << " to:" << p_dstPort->p_node->name << "/"
                 << (unsigned int)p_dstPort->num
                 << endl;
            anyError++;
          }
      }

    }
  }

  if (numBadPaths) {
    if (numBadPaths > 100)
      cout << "-W- Stopped checking for loops after 100 errors" << endl;
    cout << "-E- Found:" << numBadPaths
         << " CA to CA paths that can cause credit loops." << endl;
  } else
    cout << "-I- No credit loops found in:" << paths
         << " CA to CA paths" << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;
  return(numBadPaths + anyError);
}


///////////////////////////////////////////////////////////////////////////////
int
SubnMgtUpDnIsLegelStep(IBNode *curNode, IBNode *nxtNode)
{
    int curDir,nxtDir;
    //cout << "-V- B4 __IsLegelStep check" <<endl;
    curDir = (int) curNode->appData2.val;
    nxtDir = (int) nxtNode->appData2.val;
    //  cout << "-V- In  __IsLegelStep check , curDir="<<curDir<<" nxtDir="<<nxtDir<<"\n" << endl;
    if (curDir == 1 && nxtDir == 0)
        return 1;
    else
        return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Do BFS and set Min Hops for the current port , based on previous defined ranking
int
SubnMgtUpDnBFSFromPort(lid_t lid, IBFabric *p_fabric)
{
    list_pnode      CurState , NextState;
    IBPort          *tmpPort, *self;
    IBNode          *self_node;

    tmpPort = p_fabric->getPortByLid(lid);
    if (!tmpPort) {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- There is no port associated with the lid: " << lid << endl;
        
        return 0;
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-V- BFS for lid="<<lid<<", type of port= " << tmpPort->p_node->type << endl;
        tmpPort->p_node->repHopTable();
    }

    if (tmpPort->p_node->type == IB_SW_NODE) {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- This is a switch ..." << endl;

        self = tmpPort;
        self_node = tmpPort->p_node;
        if (!self_node) {
            cout << "-E- The port with guid: " << tmpPort->guid_get()
                 << " is related to NULL node"  << endl;
                 exit(1);
        }

        // Update in MinHop Table port 0 with 1 hop
        //tmpPort = self_node->getPort(0);
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- tmpPort is : " << tmpPort << endl;

        if (self->base_lid == lid)
            self_node->setHops(tmpPort,lid,0);
        else
            self_node->setHops(tmpPort,lid,1);

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            self_node->repHopTable();
    } else { // CA node
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
            cout << "-V- This is an HCA ..." << endl;
            cout << "-V- tmpPort :" << tmpPort << " remote Port : " << tmpPort->p_remotePort << endl;
        }

        self = tmpPort->p_remotePort;
        if (!self) {
            cout << "-E- The port: " << tmpPort->getName()
                 << " is not connected to any ramote port"  << endl;
                 exit(1);
        }
        self_node = self->p_node;
        if (!self_node) {
            cout << "-E- The port with guid: " << self->guid_get()
                 << " is related to NULL node"  << endl;
                 exit(1);
        }

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- After assignment of self " << self << " && self_node " << self_node << endl;

        // Check to see non switchable subnet
        if (self_node->type != IB_SW_NODE) {
            cout << "-W- This is a non switch subnet could not perform algorithm" << endl;
            return 1;
        }

        // Assign current lid (HCA) to the remote node
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            self_node->repHopTable();
        // If we BFS through base_lid the same lid we can reach it by 0 hops
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- base lid : " <<self->base_lid << " BFS lid : " << lid << endl;

        self_node->setHops(self, lid, 1);

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            self_node->repHopTable();
    }

    // We use the current node appData2.val to store direction
    self->p_node->appData2.val = 0UL;
    // Push into list the Current Item
    CurState.push_back(self->p_node);
    // Iterate over all nodes in CurState list till empty
    //  cout << "-V- b4 while .." << endl;
    while (CurState.size()) {
        NextState.clear();
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Iterating in while : size of CurState is : "
                  << CurState.size() << " size of NextState is : " << NextState.size() <<  endl;

        // go over cur step nodes
        for (list_pnode::iterator lI = CurState.begin(); lI != CurState.end(); lI++) {
            IBNode *p_node = *lI;

            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                cout << "-V- Current switch handeled is : " << p_node->name << endl;

            // go over all ports to find unvisited remote nodes
            for (phys_port_t pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort(pn);

                // Only if current port is NULL or not connected skip it
                if (!p_port || !p_port->p_remotePort)
                    continue;

                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- Handling port num : " << pn << " of node :" << p_node->name
                        << " p_type="<<p_node->type<<"\n"<<endl;

                IBPort *p_rem_port = p_port->p_remotePort;
                IBNode *p_rem_node = p_rem_port->p_node;
                uint8_t minHop = 0;

                // Only if it is a switch then update its table
                if (p_rem_node->type == IB_SW_NODE) {
                    // First put UP / DOWN label to remote node
                    //IsRankAssigned will print a warning message
                    if (!IsRankAssigned(p_node) || !IsRankAssigned(p_rem_node))
                        cout << "-W- try to continue to create UpDown MinHop table "
                             << "inspite of rank issue" << endl;

                    if (p_node->rank < p_rem_node->rank)
                        // This is DOWN
                        p_rem_node->appData2.val = 1UL;
                    else
                        // This is UP
                        p_rem_node->appData2.val = 0UL;
                    // cout << "-V- b4 Validity check of UP/DOWN"<<endl;

                    // Check validity of direction
                    if (SubnMgtUpDnIsLegelStep(p_node, p_rem_node))
                        continue;

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-V- Current step is legal ..." << endl;

                    // Update remote port MinHopTable according to Current MinHopTable
                    // Set minHop with the minimum hop count for this lid through current node
                    minHop = p_node->getHops(NULL, lid);

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                        cout << "-V- lid=" <<lid << " p_node="<<p_node<<"\n"<<endl;
                        cout << "-V- MinHopTable of current node at lid=" <<lid
                                << " port=0 is : " << minHop<<endl;
                        //cout << "-V- After minHop :"<<minHop<<"\n" <<endl;
                        cout << "-V- B4 Check of current_node(minHop)=" << minHop
                                << " and p_rem_node(minHop)="
                                << p_rem_node->getHops(p_rem_port, lid) << "\n"<<endl;
                    }

                    // we will use the remote node min hops to know if we
                    // already have visited this node
                    int remNodeHops = p_rem_node->getHops(NULL, lid);

                    // Check min hop count if better insert into NextState
                    // list && update the remote node Min Hop Table
                    if (minHop + 1 <= p_rem_node->getHops(p_rem_port, lid)) {
                        // Update MinHopTable of remote node and add it to NextState
                        p_rem_node->setHops(p_rem_port, lid, (uint8_t)(minHop + 1));

                        // only if the remote node did not have this hop we
                        // need to visit it next step
                        if (remNodeHops > minHop + 1)
                            NextState.push_back(p_rem_node);

                        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                            cout << "-V- Updating MinHopTable of node : " << p_rem_node->name
                                    << " for lid :" <<  lid  << " Value is : " << minHop << "\n" << endl;
                            // Lets see the last update for the current node
                            p_rem_node->repHopTable();
                        }
                    }
                }
            }
        }

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Start another iteration for NextState with : "
                  << NextState.size() << " elements" << endl;

        CurState = NextState;
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- End of Iterations ..." << endl;
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Calc Min Hop Table using UP-DOWN algorithm
int
SubnMgtCalcUpDnMinHopTbls(IBFabric *p_fabric)
{
    // go over all the lids and init their Min Hop Tables
    for (lid_t i = p_fabric->minLid; i <= p_fabric->maxLid; i++) {
        IBPort *p_cur_port = p_fabric->getPortByLid(i);
        if (p_cur_port == NULL)
            continue;

        IBNode *p_cur_node = p_cur_port->p_node;

        if (p_cur_node->type == IB_SW_NODE) {
            // assign all ports of the current node/lid initial value
            p_cur_node->setHops(NULL, 0, IB_HOP_UNASSIGNED);
        }
    }

    // Now do the UP-Down Min Hop propagation
    for (lid_t i = p_fabric->minLid; i <= p_fabric->maxLid; i++)
        if (SubnMgtUpDnBFSFromPort(i, p_fabric))
	        return 1;

    // dump the min hops
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
                nI != p_fabric->NodeByName.end();
                nI++) {
            IBNode *p_node = (*nI).second;
            if (p_node->type != IB_SW_NODE)
                continue;
            p_node->repHopTable();
        }
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// do up down given a node name regular expression to find the
// root nodes
int
SubnMgtCalcUpDnMinHopTblsByRootNodesRex(IBFabric *p_fabric,
                                        const char * rootNodesNameRex)
{
    // rank the fabric
    SubnRankFabricNodesByRegexp(p_fabric, rootNodesNameRex);

    // do the actual route
    SubnMgtCalcUpDnMinHopTbls(p_fabric);
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
//
// Multicast Check Code
//
struct bfsEntry {
    IBNode *pNode;
    uint8_t inPort;
};

// return true if remote side is a switch and it points back to the given port
// by its MFT
bool isRemSwPortPointingBackByMFT( IBPort *pPort, uint16_t mlid)
{
    if ( !pPort || !pPort->p_remotePort ) return(false);
    IBPort *pRemPort = pPort->p_remotePort;
    IBNode *pRemNode = pRemPort->p_node;
    if ( pRemNode->type != IB_SW_NODE ) return(false);

    list_phys_ports mftPNs = pRemNode->getMFTPortsForMLid(mlid);
    return(find(mftPNs.begin(), mftPNs.end(), pRemPort->num) != mftPNs.end());
}

// look through all the remote switch of the given port if
// it has any port pointing back to the original port node
IBPort * getAnyPortPointingBackByMFT(IBPort *pPort, uint16_t mlid)
{
    if ( !pPort || !pPort->p_remotePort ) return(NULL);
    IBPort *pRemPort = pPort->p_remotePort;
    IBNode *pRemNode = pRemPort->p_node;
    if ( pRemNode->type != IB_SW_NODE ) return(NULL);

    list_phys_ports mftPNs = pRemNode->getMFTPortsForMLid(mlid);
    for ( list_phys_ports::const_iterator pnI = mftPNs.begin();
        pnI != mftPNs.end(); pnI++ ) {
        IBPort *pAnyRemPort = pRemPort->p_node->getPort(*pnI);
        if ( pAnyRemPort &&
             pAnyRemPort->p_remotePort &&
             pAnyRemPort->p_remotePort->p_node == pPort->p_node ) {
            return(pAnyRemPort);
        }
    }
    return(NULL);
}

bool
isAggregationNodeInList(const list< IBNode *>& nodes_list)
{
    for (IBNode *p_node : nodes_list) {
        if (p_node->isSpecialNode())
            return true;
    }

    return false;
}

// Check that if a single plane appears in the MFT, all other APort planes are
// in the MFT as well.
int
checkFabricAPortMFTSymmetry(IBFabric *p_fabric, lid_t mlid,
                    const list< IBNode *>& nodes_list)
{
    int errs = 0;
    // make sure the mlid is in range:
    if (mlid < 0xc000) {
        cout << "-E- checkFabricAPortMFTSymmetry : Given lid:" << mlid
                << " is out of range" << endl;
        return 1;
    }

    ibport_accessor_t<bool> is_in_mft = [mlid] (const IBPort *p_port) {
        lid_t idx = (lid_t)(mlid - 0xc000);
        if (!p_port || !p_port->p_node || p_port->p_node->MFT.size() <= idx)
            return false;
        return p_port->p_node->MFT[idx].test(p_port->num);
    };
    unordered_set<uint64_t> visited_systems;

    for (IBNode *p_node : nodes_list) {
        if (!p_node || visited_systems.count(p_node->system_guid_get()))
            continue;

        const auto& aports_entry = p_fabric->APortsBySysGuid.find(p_node->system_guid_get());
        if (aports_entry == p_fabric->APortsBySysGuid.end())
            continue;

        for (APort *p_aport : aports_entry->second) {
            if (!p_aport)
                continue;

            if (!p_aport->isEqualAttribute(is_in_mft)) {
                cout << "-E- APort " << p_aport->getName() 
                     << " has inconsistent MFT entries in differet planes for MLID "
                     << mlid << endl;
                errs++;
            }
        }

        visited_systems.insert(p_node->system_guid_get());
    }

    return errs;
}


// DFS Full Member MFTs
int dfsSendRecvMemMCG(IBNode *sw, IBPort *inPort, uint16_t mlid,
                  NodesVisits &swGroup, IBNode *root_node,
                  bool onlyBiDir, int level)
{
    int anyErr = 0;
    char buffer[128];
    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        if ( inPort ) {
            cout << "-V- DFS FullMembers Visiting node:" << sw->name
            << " through port:" << (unsigned int)inPort->num
            << " for mlid:" << buffer << " level:" << level << endl;
        } else {
            cout << "-V- DFS FullMembers starting node:" << sw->name
            << " for mlid:" << buffer << endl;
        }
    }

    // get the list of MC group ports for this mlid of this node
    list_phys_ports portNums = sw->getMFTPortsForMLid(mlid);

    // go over all MC output ports of the current node ignoring the input
    list_phys_ports::iterator pnI;
    for ( pnI = portNums.begin(); pnI != portNums.end(); pnI++ ) {
        phys_port_t pn = (*pnI);

        IBPort *pPort = sw->getPort(pn);
        if ( ! pPort || ! pPort->p_remotePort ) continue;

        // ignore the port we got here through.
        if ( pPort == inPort ) continue;

        // ignore FNM ports, checked later
        if (pPort->isFNMPort()) continue;

        // get the remote node
        IBNode *pRemNode = pPort->p_remotePort->p_node;
        if ( !pRemNode ) continue;

        // we ignore remote HCAs
        if ( pRemNode->type != IB_SW_NODE ) continue;

        // ignore ports that are blocked by Rail Filter
        if ( inPort && !inPort->isPassingRailFilter(pn) )
            continue;

        // igonre ports that were already exited in previous times
        // (e.g. legacy ports on Crocodile)
        if ( swGroup.isAlreadyExited(sw, pn) )
            continue;

        swGroup.addExit(sw, pn);

        bool allowToProceed = true;

        // no need to go here if already visited
        if ( swGroup.isVisited(pRemNode, pPort->p_remotePort) ) {
            // we only report loops on cases of !onlyBiDir
            if ( !onlyBiDir && (swGroup.getRootNode(pRemNode) == root_node) ) {
                cout << "-E- Found a loop on MLID:" << buffer
                << " got to node:" << pRemNode->name
                << " again through port:" << (unsigned int)pPort->p_remotePort->num
                << " connected to:" << sw->name << " port:"
                << (unsigned int)pn;
                if ( pPort->p_remotePort->get_plane_number() != -1 )
                    cout << " plane: " << pPort->p_remotePort->get_plane_number();
                cout << endl;
                anyErr++;
            }
            allowToProceed = false;
        }

        // criteria for visiting is based on onlyBiDir
        if ( onlyBiDir ) {
            // if the remote node does not point back to this one
            // (i.e. the port bit is not set in the MFT do not go through ...
            if ( !isRemSwPortPointingBackByMFT(pPort, mlid) ) {
                cout << "-W- Found a non-bidir MC route on FullMember tree of MLID:"
                << buffer
                << "\n    From port:" << pPort->getName()
                << " to port:" << pPort->p_remotePort->getName()
                << " which does not point back." << endl;
                anyErr++;
                // do not go there
                allowToProceed = false;
            }
        }

        if ( allowToProceed ) {
            swGroup.addVisit(pRemNode, root_node, pPort->p_remotePort);
            anyErr += dfsSendRecvMemMCG(pRemNode, pPort->p_remotePort, mlid,
                                    swGroup, root_node, onlyBiDir, level+1);
        }
    } // all ports
    return anyErr;
}

// DFS Sender Only Tree
int dfsSenderOnlyMCG(IBNode *sw, IBPort *inPort, uint16_t mlid,
                     NodesVisits &soGroup,
                     SONodesReach &swPassingSenders,
                     NodesVisits &swGroup,
                     IBNode *p_root_node,
                     int level)
{
    int anyErr = 0;
    char buffer[128];
    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        if ( inPort ) {
            cout << "-V- DFS SenderOnlyMembers Visiting node:" << sw->name
            << " through port:" << (unsigned int)inPort->num
            << " for mlid:" << buffer << " level:" << level << endl;
        } else {
            cout << "-V- DFS SenderOnlyMembers starting node:" << sw->name
            << " for mlid:" << buffer << endl;
        }
    }

    // get the list of MC group ports for this mlid of this node
    list_phys_ports portNums = sw->getMFTPortsForMLid(mlid);
    if (!portNums.size()) {
        cout << "-E- Empty MFT. Got to dead end from host :" << sw->name
            << " for mlid:" << buffer << endl;
        anyErr++;
    } else if (portNums.size() > 1 && !APort::isSameAPort(sw, portNums))
        cout << "-W- Sender passing node " << sw->name
             << " has more than one MFT forwarding port"
             << " for mlid:" << buffer << endl;

    // go over all MC output ports of the current node
    list_phys_ports::iterator pnI;
    for ( pnI = portNums.begin(); pnI != portNums.end(); pnI++ ) {
        phys_port_t pn = (*pnI);
        IBPort *pPort = sw->getPort(pn);
        if ( ! pPort || ! pPort->p_remotePort ) continue;

        // get the remote node
        IBNode *pRemNode = pPort->p_remotePort->p_node;

        //check for bi-direction link to inPort from MFT port
        if (pPort == inPort) {
            cout << "-E- Bi-Direction link for sender only from port: "
                 << pPort->getName() << "to port " << pRemNode->name << endl;
            anyErr++;
            continue;
        }

        if (pPort->isFNMPort()) {
            cout << "-E- FNM Port " << pPort->getName()
                 << " isn't allowed on send only multicast switch" << endl;
            anyErr++;
            continue;
        }

        // we ignore remote HCAs
        if ( pRemNode->type != IB_SW_NODE ) {
            cout << "-E- forward multicast package from sw "
                 << sw->name << " to non member HCA "
                 << pRemNode->name << " mlid: "
                 << buffer << endl;
            continue;
        }

        // ignore ports that are blocked by Rail Filter
        if ( inPort && !inPort->isPassingRailFilter(pn) )
            continue;

        // igonre ports that were already exited in previous times
        // (e.g. legacy ports on Crocodile)
        if ( soGroup.isAlreadyExited(sw, pn) )
            continue;

        soGroup.addExit(sw, pn);

        // stop condition, if this remote node is a full member
        if ( swGroup.isVisited(pRemNode, pPort->p_remotePort) ) {
            //we reached the first full member in dfs
            //insert the full member root the list of full members connecting
            //to this dfs branch.
           if (!swPassingSenders[sw].addReached(swGroup.getRootNode(pRemNode), pPort)) {
                cout << "-E- duplicate multicast package from switch " << sw->name
                     << " in port " << (unsigned int)pn
                     << " to mlid: " << buffer << endl;
                anyErr++;
            }
            
            continue;
        }

        if (soGroup.isVisited(pRemNode, pPort->p_remotePort)) {
            // if the remote switch has the same root as the sw we have a loop
            if (p_root_node == soGroup.getRootNode(pRemNode)) {
                cout << "-E- Found a loop on MLID:" << buffer
                    << " got to node:" << pRemNode->name
                    << " again through port:" << (unsigned int)pPort->p_remotePort->num
                    << " connected to:" << sw->name << " port:" << (unsigned int)pn;
                if (pPort->p_remotePort->p_aport)
                    cout << " on plane: " << pPort->p_remotePort->get_plane_number();
                cout << endl;
                anyErr++;
                continue;
            }

            // add FullMembers that are reached by the remote node to the current node's reach
            // check duplicated links (FullMembers reached twice)
            if (!swPassingSenders.addRemoteReach(sw, pRemNode, pPort)) {
                cout << "-E- duplicate multicast package from host " << sw->name
                    << " in port " << (unsigned int)pn
                    <<" to mlid: " << buffer << endl;
                anyErr++;
            }

            continue;
        }

        // if we got here, remote node isn't visited or full member,
        // proceeding with the DFS
        soGroup.addVisit(pRemNode, p_root_node, pPort->p_remotePort);
        anyErr += dfsSenderOnlyMCG(pRemNode, pPort->p_remotePort, mlid,
                                soGroup, swPassingSenders, swGroup, p_root_node, level+1);

        // add FullMembers that are reached by the remote node to the current node's reach
        // check duplicated links (FullMembers reached twice)
        if (!swPassingSenders.addRemoteReach(sw, pRemNode, pPort)) {
            cout << "-E- duplicate link from host " << sw->name
                    << " to host " << pRemNode->name
                    <<" on multicast group: " << buffer << endl;
            anyErr++;
        }
    } // all ports
    return anyErr;
}

// If the given node is a Black Mamba node,
// check that the node has valid internal FNM connections
// in the MFT.
int
CheckBlackMambaNodeFNMPorts(uint16_t mlid, IBNode *p_node,
                            int min_member_plane,
                            int max_member_plane)
{
    // Nodes should be connected in a "line" of FNM ports,
    // with all nodes connected to each member node via MFT.
    // []--[]--[]--[]
    // 1   2   3   4
    // For exmaple:
    // if node 1 is a member:
    // []<-[]<-[]<-[]
    //(1)  2   3   4
    // if both node 1 and node 3 are members:
    // []<->[]<->[]<-[]
    //(1)   2   (3)  4

    int errs = 0;
    int curr_plane = p_node->getSuitablePlane();
    if (curr_plane <= 0 || min_member_plane <= 0 || min_member_plane <= 0)
        return errs;
    IBPort *forward_fnm = p_node->getFNMPort(FNM_FORWARD_PORT);
    IBPort *backward_fnm = p_node->getFNMPort(FNM_BACKWARD_PORT);
    if (!forward_fnm || !backward_fnm)
        return errs;

    list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);
    // if there are members in this system with smaller plane number, backwards FNM should be on.
    bool backward_mft = count(portNums.begin(), portNums.end(), backward_fnm->num);
    bool should_backward_mft = curr_plane > min_member_plane;
    bool forward_mft = count(portNums.begin(), portNums.end(), forward_fnm->num);
    bool should_forward_mft = curr_plane < max_member_plane;

    if (forward_mft && !should_forward_mft) {
        cout << "-E- FNM Port " << forward_fnm->getName()
             << " shouldn't be in the multicast tree" << endl;
        errs++;
    }
    else if (!forward_mft && should_forward_mft) {
        cout << "-E- FNM Port " << forward_fnm->getName()
             << " missing from the multicast tree" << endl;
        errs++;
    }

    if (backward_mft && !should_backward_mft) {
        cout << "-E- FNM Port " << backward_fnm->getName()
             << " shouldn't be in the multicast tree" << endl;
        errs++;
    }
    else if (!backward_mft && should_backward_mft) {
        cout << "-E- FNM Port " << backward_fnm->getName()
             << " missing from the multicast tree" << endl;
        errs++;
    }

    return errs;
}

// Check multicast group given a list of Full Members Ports and a list
// of sender only ports
int
SubnMgtCheckMCGrpByMemPortLists(IBFabric *p_fabric,
                                uint16_t mlid,
                                list<IBPort *> &sendRecvMemPorts,
                                list<IBPort *> &sendOnlyPorts)
{
    char buffer[128];
    bool onlyBiDir = true;
    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);

    // the algorithm searches for loops, connection groups for both full and
    // send only semantics. Also report switches with garbadge entries

    // to overcome the missing port 0 in ibdm if listed port is of a switch it
    // means port 0

    //          | GROUP CAN BE | GROUP CAN'T BE
    //  ---------------------------------------
    //   SENDER | SENDER-ONLY- | SENDER-ONLY
    //          | FULL-MEMBER  |
    //          | (BIT 1000)   | (BIT 0100)
    //  ---------------------------------------
    // SEND/RECV| FULL-MEMBER  | NON-MEMBER
    //          | (BIT 0001)   | (BIT 0010)
    //
    // Data model:
    // VISITED[node] marked node visited in order to find loops
    // swGroup[node] = full member sw root.
    // soGroup[sw] = senderOnly sw root, from which we started path to the full member tree
    // swPassingSenders[sw] = list of, root full mems, reached for this sender (which comes from swGroup[first SW We found that is in the fullMember tree])
    //
    // SenderOnly Example:
    // (HCA1-SenderOnly) (HCA2-SenderOnly)
    //     |                   |
    //  [ SW-A ]            [ SW-B ]
    //      \               /
    //    [ SW-C ]    [ SW-D ]
    //        \         /
    //       [ SW-E - FullMemeberTreeRoot ]
    //         |
    //        [ ... ]
    //
    // After running the senderOnly algorithem we'll get this:
    // soGroup[SW-A] = SW-A
    // soGroup[SW-C] = SW-A
    // soGroup[SW-B] = SW-B
    // soGroup[SW-D] = SW-B
    //
    // swPassingSenders[SW-A] = < E >
    // swPassingSenders[SW-C] = < E >
    // swPassingSenders[SW-B] = < E >
    // swPassingSenders[SW-D] = < E >
    //
    // Algorithm:
    // NOTE: due to the requirement that full member connections
    //       will only utilize bi-dir links we need to search for
    //       loops with a simpler algorithm that is directional
    //       so loop detection is simpler and require a slight change
    //       in the dfs
    //
    // # Full Members
    // foreach full(send/recv) port
    //   if a CA port
    //     verify connected switch has entry to it or error
    //     get to the sw
    //
    //   if !swGroup[sw]
    //      swGroup[sw] = sw
    //      dfsSendRecvMemMCG(mlid, sw, swGroup, sw, onlyBiDir) - traverse the switches tree
    //
    //   if !VISITED[sw]
    //       VISITED[sw] = sw
    //       dfsSendRecvMemMCG(mlid, sw, VISITED, sw, !onlyBiDir)
    //
    //   if CA port swGroup[ca] = swGroup[sw]
    //
    // # Sender only
    // foreach sender port
    //   if a CA port
    //     get to the connected sw
    //
    //   swPassingSenders[port] = dfsSenderOnlyMCG(mlid, port, sw, sw)
    //
    // # Summary
    // foreach switch in the fabric
    //  if has MFT for mlid
    //    if !swGroup[sw] or !SENDERS[sw]
    //      report as garbadge MFT
    //    if swGroup[sw] collect GROUP_MEMS[swGroup[sw]]
    // if GROUP_MEMS.size() > 1
    //   Report members in each group
    // foreach sender port
    //   if !SENDER_GROUPS[port]
    //     Error disconnected sender
    //   else
    //     verbose report groups GROUP_MEMS[GROUP[SENDER_GROUPS[port]]]

    // Data structures:
    NodesVisits swGroup;
    NodesVisits swVisited;
    NodesVisits soGroup;
    SONodesReach swPassingSenders;
    map<IBNode*, set<IBNode*> > groupMembers;

    // Algorithm:
    int anyErr = 0;
    if (sendRecvMemPorts.size() == 1 && sendOnlyPorts.empty()) {
        list<IBPort *>::iterator mI = sendRecvMemPorts.begin();
        IBPort *pPort = *mI;
        cout << "-I- FullMember node:" << pPort->p_node->getName()
             << " is a single member for mlid " << buffer
             << endl;
        return(0);
    }

    // # Full Members
    for ( list<IBPort *>::iterator mI = sendRecvMemPorts.begin();
        mI != sendRecvMemPorts.end(); mI++ ) {
        IBPort *pPort = *mI;
        IBPort *p_start_port = nullptr;
        APort *p_start_aport = nullptr;

        IBNode *startSw = NULL;
        IBNode *root_node = NULL;

        if ( pPort->p_node->type != IB_SW_NODE ) {
            IBPort *pRemPort = pPort->p_remotePort;
            if ( !pRemPort ) {
                cout << "-E- FullMember port:" << pPort->getName()
                << " is not connected" << endl;
                anyErr++;
            } else {
                IBNode *pRemNode = pRemPort->p_node;

                // make sure there is a remote switch and it has MFT pointing here
                if ( !isRemSwPortPointingBackByMFT(pPort, mlid) ) {
                    cout << "-E- FullMember port:" << pPort->getName()
                    << " is not connected to MFT on attached switch:"
                    << pRemNode->name << endl;
                    anyErr++;
                }
                startSw = pRemNode;
                p_start_port = pRemPort;
                p_start_aport = p_start_port->p_aport;
            }
        } else {
            // starting from a switch
            startSw = pPort->p_node;
            p_start_port = pPort;
            p_start_aport = pPort->p_aport;
        }

        if ( !startSw ) continue;
        root_node = startSw;
        // checking for connection groups
        if ( !swGroup.isVisited(startSw) ) {
            swGroup.addVisit(startSw, root_node, p_start_port);
            anyErr += dfsSendRecvMemMCG(startSw, NULL, mlid, swGroup,
                                            root_node, onlyBiDir, 0);
            // make sure all planes are a part of the DFS
            if (p_start_aport) {
                for (IBPort *p_plane : p_start_aport->ports) {
                    if (!p_plane || !p_plane->p_node)
                        continue;
                    if (p_plane == p_start_port)
                        continue;
                    if (swGroup.isVisited(p_plane->p_node, p_plane))
                        continue;
                    swGroup.addVisit(p_plane->p_node, root_node, p_plane);

                    // no need to start the same DFS from the same node
                    if (p_plane->p_node == startSw)
                        continue;

                    anyErr += dfsSendRecvMemMCG(p_plane->p_node, NULL, mlid, swGroup,
                                                    root_node, onlyBiDir, 0);
                }
            }
        } else {
            root_node = swGroup.getRootNode(startSw);
        }

        if ( pPort->p_node->type != IB_SW_NODE ) {
            if (pPort->p_aport)
                // APort MLID symmetry checked earlier,
                // all planes in APort can be marked as visited
                swGroup.addAPortVisit(pPort->p_aport, root_node);
            else
                swGroup.addVisit(pPort->p_node, root_node);
        }

        // checking for loops
        if ( !swVisited.isVisited(startSw) ) {
            swVisited.addVisit(startSw, root_node, p_start_port);
            anyErr += dfsSendRecvMemMCG(startSw, NULL, mlid, swVisited,
                                    root_node, !onlyBiDir, 0);
            if (p_start_aport) {
                for (IBPort *p_plane : p_start_aport->ports) {
                    if (!p_plane || !p_plane->p_node)
                        continue;
                    if (p_plane == p_start_port)
                        continue;
                    if (swVisited.isVisited(p_plane->p_node, p_plane)) {
                        cout << "-E- Found a plane loop on MLID:" << mlid
                             << " on node:" << p_plane->p_node->name
                             << " reached plane " << p_plane->p_port_hierarchy_info->m_plane
                             << " from other planes." << endl;
                        anyErr++;
                        continue;
                    }
                    swVisited.addVisit(p_plane->p_node, root_node, p_plane);

                    if (p_plane->p_node == startSw)
                        continue;

                    swVisited.addVisit(p_plane->p_node, root_node, p_plane);
                    anyErr += dfsSendRecvMemMCG(p_plane->p_node, NULL, mlid, swVisited,
                                    root_node, !onlyBiDir, 0);
                }
            }
        }

    } // all full(send/recv) members

    // # Sender only
    // foreach sender port
    //   if a CA port
    //     get to the connected sw
    //
    //   SENDER_GROUPS[port] = dfsSenderOnlyMCG(mlid, port, sw, sw)
    for ( list<IBPort *>::iterator mI = sendOnlyPorts.begin();
        mI != sendOnlyPorts.end(); mI++ ) {
        IBPort *pPort = *mI;
        IBPort *p_start_port = nullptr;

        IBNode *startSw = NULL;
        if ( pPort->p_node->type != IB_SW_NODE ) {
            IBPort *pRemPort = pPort->p_remotePort;
            if ( !pRemPort ) {
                cout << "-E- Send onlyMember port:" << pPort->getName()
                << " is not connected" << endl;
                anyErr++;
            } else {
                IBNode *pRemNode = pRemPort->p_node;
                if (isRemSwPortPointingBackByMFT(pPort, mlid) ) {
                    cout << "-E- SenderOnlyMember port:" << pPort->getName()
                        << " is connected Bi-Direct on attached switch:"
                        << pRemNode->name << endl;
                    anyErr++;
                }
                startSw = pRemNode;
                p_start_port = pRemPort;
            }
        } else {
            // starting from a switch
            startSw = pPort->p_node;
            p_start_port = pPort;
        }

        if ( !startSw ) continue;

        // if the sw is in the swgroup , meaning that sw is a fullMember
        // or part of the fullMember tree(send/recv non member)
        if (swGroup.isVisited(startSw)) {
            // if the senderOnly port is also part of the fullMember tree
            // no need to dfs him.
            continue;
        }

        if (soGroup.isVisited(startSw)) {
            continue;
        }

        soGroup.addVisit(startSw, startSw, p_start_port);
        dfsSenderOnlyMCG(startSw, NULL, mlid,
                            soGroup, swPassingSenders,
                            swGroup, startSw, 0);
        // checking for disconnect
        if (!swPassingSenders[startSw].reachedAny()) {
            cout << "-E- sender only node " << startSw->name
                << " is disconnected from the multicast for mlid: "
                << buffer << endl;
        }
        if (p_start_port->p_aport) {
            for (IBPort *p_plane : p_start_port->p_aport->ports) {
                if (!p_plane || !p_plane->p_node)
                    continue;
                if (p_plane == p_start_port)
                    continue;
                if (soGroup.isVisited(p_plane->p_node, p_plane)) {
                    cout << "-E- Found a plane loop on MLID:" << mlid
                            << " on node:" << p_plane->p_node->name
                            << " reached plane " << p_plane->p_port_hierarchy_info->m_plane
                            << " from other planes." << endl;
                    anyErr++;
                    continue;
                }
                soGroup.addVisit(p_plane->p_node, startSw, p_plane);

                // no need to start DFS from the same node twice
                if (p_plane->p_node == startSw)
                    continue;

                dfsSenderOnlyMCG(p_plane->p_node, NULL, mlid,
                            soGroup, swPassingSenders,
                            swGroup, startSw, 0);
            }
        }

    } // all sendonly members


    // # FNM Ports
    //
    // Go over all full member ports. If the full member port is on a
    // Black Mamba switch, check that switch's FNM port connections
    // with CheckBlackMambaNodeFNMPorts.
    // Report an error on all FNM ports in the MFT that aren't on a valid BlackMama node.
    //
    // NOTE: currently, FNM checks will only work for BlackMamba switches.
    // A more general test would check that for all planarized systems with a member node,
    // all other nodes in the same system can reach the member node via FNM. it should
    // also check that no redundant FNM ports are active in the MFT.

    // For each planarized system, the minimum and maximum plane numbers
    // of full-member nodes (used in CheckBlackMambaNodeFNMPorts).
    map<IBSystem *, pair<int, int>> sys_min_max_planes;
    for (IBPort *p_port : sendRecvMemPorts) {
        if ( p_port->p_node->type != IB_SW_NODE || p_port->num != 0 )
            continue;

        // make sure this node has FNM port
        if (!p_port->p_node->hasFNMPort())
            continue;

        // make sure this node is a valid BlackMamba node
        int curr_node_plane = p_port->p_node->getSuitablePlane();
        if (p_port->p_node->isPrismaSwitch() || curr_node_plane == -1)
            continue;

        auto it = sys_min_max_planes.find(p_port->p_node->p_system);
        if (it == sys_min_max_planes.end()) {
            sys_min_max_planes[p_port->p_node->p_system] = {curr_node_plane, curr_node_plane};
        } else if (it->second.first > curr_node_plane) {
            it->second.first = curr_node_plane;
        } else if (it->second.second < curr_node_plane) {
            it->second.second = curr_node_plane;
        }
    }

    for (auto& system_ports_it : sys_min_max_planes) {
        IBSystem *p_system = system_ports_it.first;
        pair<int, int>& min_max_planes = system_ports_it.second;
        for (auto& node_name_it : p_system->NodeByName) {
            IBNode *p_node = node_name_it.second;
            // make sure this node should be connected to member nodes in this system
            // (e.g. this node isn't and aggregation node)
            if (p_node->type != IB_SW_NODE)
                continue;

            if (p_node->getSuitablePlane() == -1)
                continue;

            anyErr += CheckBlackMambaNodeFNMPorts(mlid, p_node,
                            min_max_planes.first, min_max_planes.second);
        }
    }

    for (auto& node_name_it : swGroup.visits) {
        IBNode *p_curr_node = node_name_it.first;

        if (!p_curr_node)
            continue;
        // skip all checked nodes
        if (sys_min_max_planes.count(p_curr_node->p_system))
            continue;
        if (!p_curr_node->hasFNMPort())
            continue;

        list_phys_ports curr_mft_ports = p_curr_node->getMFTPortsForMLid(mlid);
        for (phys_port_t port_num : curr_mft_ports) {
            IBPort *p_port = p_curr_node->getPort(port_num);
            if (!p_port)
                continue;
            if (p_port->isFNMPort()) {
                cout << "-E- FNM Port " << p_port->getName()
                     << " shouldn't be in the multicast tree"
                     << " (node shouldn't have active multicast FNM ports)." << endl;
                anyErr++;
            }
        }

    }

    // # Summary
    // foreach switch in the fabric
    //  if has MFT for mlid
    //    if !swGroup[sw] or !SENDERS[sw]
    //      report as garbadge MFT
    //    if swGroup[sw] collect GROUP_MEMS[swGroup[sw]]

    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end();
        nI++ ) {

        IBNode *p_node = (*nI).second;

        if ( swGroup.isVisited(p_node) ) {
            groupMembers[swGroup.getRootNode(p_node)].insert(p_node);
        } else {
            //if node is a switch and not part of the senderOnly group
            if ( p_node->type == IB_SW_NODE && !soGroup.isVisited(p_node)) {

                // see if we have an MFT entry by the given lid:
                list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);
                if ( !portNums.empty() ) {
                    list_phys_ports::iterator fI =
                    find(portNums.begin(), portNums.end(), 0);
                    if ( portNums.end() != fI ) {
                        cout << "-E- Switch: " << p_node->name
                        << " is not connected to MLID:"
                        << buffer << endl;
                        anyErr++;
                    } else {
                        cout << "-W- Switch: " << p_node->name
                        << " has unconnected MFT entries for MLID:"
                        << buffer << endl;
                    }
                }
            }
        }
    }

    if ( groupMembers.size() > 1 ) {
        cout << "-E- Found FullMembers" << groupMembers.size()
        << " connection groups for MLID:"
        << buffer << endl;
        int numGrps = 0;
        for ( map<IBNode*, set<IBNode*> >::iterator gI = groupMembers.begin();
            gI != groupMembers.end(); gI++ ) {
            numGrps++;
            list <string> cas;
            list <string> sws;
            int numCAs = 0;
            int numSWs = 0;
            for ( set<IBNode*>::iterator mI = (*gI).second.begin();
                mI != (*gI).second.end(); mI++ ) {
                IBNode *pNode = *mI;
                if ( pNode->type == IB_SW_NODE ) {
                    numSWs++;
                    sws.push_back(pNode->name);
                } else {
                    numCAs++;
                    cas.push_back(pNode->name);
                }
            }
            cout << "    Group:" << numGrps << " has "<< numCAs << " CAs: "
            << compressNames(cas) << endl;
            cout << "    Group:" << numGrps << " has "<< numSWs << " SWs: "
            << compressNames(sws) << endl;
        }
        anyErr++;
    }

    return(0);
}

int
SubnMgtCheckMCGrp(IBFabric *p_fabric,
                  lid_t mlid,
                  class McastGroupInfo &mcast_info)
{
    int anyErr = 0;
    char buffer[128] = {0};
    list< IBNode *> groupSwitches;
    list< IBNode *> groupHCAs;
    list< IBPort *> groupSendRecvMemPorts;
    list< IBPort *> sendOnlyPorts;
    bool is_sharp_mctree = false;

    for (map_mcast_members::iterator mI = mcast_info.m_members.begin();
         mI != mcast_info.m_members.end();
         mI++) {
        class McastGroupMemberInfo &mcast_grp_mem = (*mI).second;
        IBPort * p_port = (*mI).first;

        if (mcast_grp_mem.is_sender_only)
            sendOnlyPorts.push_back(p_port);
        else {
            if (p_port->p_node->type == IB_SW_NODE) {
                //check that have port 0 in MFT (full memeber must have port 0
                list_phys_ports portNums = p_port->p_node->getMFTPortsForMLid(mlid);
                if ( !portNums.empty() ) {
                    list_phys_ports::iterator fI =
                        find(portNums.begin(), portNums.end(), 0);
                    if ( portNums.end() == fI ) {
                        cout << "-E- Node: " << p_port->p_node->name
                            << " is a full member and missing port 0 for MLID:"
                            << buffer << endl;
                        anyErr++;
                    }
                }
            }
            groupSendRecvMemPorts.push_back(p_port);
        }

        if (p_port->p_node->type == IB_SW_NODE)
            groupSwitches.push_back(p_port->p_node);
        else if (p_port->p_node->type == IB_CA_NODE)
            groupHCAs.push_back(p_port->p_node);
    }

    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);
    cout << "-I- Multicast Group:" << buffer << " has:" << groupSwitches.size()
    << " Switches and:" << groupHCAs.size() << " HCAs which includes: "<< groupSendRecvMemPorts.size()
    << " FullMember ports and:" << sendOnlyPorts.size() << " SenderOnly ports" << endl;

    if ( groupSwitches.empty() && groupHCAs.empty()) return(0);
    if (groupSendRecvMemPorts.empty()) return(0);
    if (anyErr) return anyErr;

    is_sharp_mctree = isAggregationNodeInList(groupSwitches);
    if (!is_sharp_mctree) {
        // if APort's aren't MFT symmetric, exit before other checks
        anyErr = checkFabricAPortMFTSymmetry(p_fabric, mlid, groupSwitches);
        anyErr += checkFabricAPortMFTSymmetry(p_fabric, mlid, groupHCAs);
    }

    anyErr += SubnMgtCheckMCGrpByMemPortLists(p_fabric, mlid,
                                             groupSendRecvMemPorts, sendOnlyPorts);
    return anyErr;
}

// Check a multicast group dealing only with full members
int
SubnMgtCheckMCGrp(IBFabric *p_fabric,
                  uint16_t mlid)
{
    list< IBNode *> groupSwitches;
    list< IBPort *> groupSendRecvMemPorts;
    list< IBPort *> sendOnlyPorts;
    int anyErr = 0;
    bool is_sharp_mctree = false;

    // find all switches that are part of this mcgrp.
    // fill in the groupSwitches, groupSwitchesWithHCAs and groupHCAs
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end();
        nI++ ) {

        IBNode *p_node = (*nI).second;
        if ( p_node->type != IB_SW_NODE ) continue;

        // see if we have an MFT entry by the given lid:
        list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);

        if ( portNums.empty() ) continue;

        groupSwitches.push_back(p_node);

        // find all CAs connected to the group by following the links that
        // are marked in the MFT that connect to the group.
        for ( list_phys_ports::iterator lI = portNums.begin();
            lI != portNums.end();
            lI++ ) {
            IBPort *p_port = p_node->getPort(*lI);

            // handle member switch port 0
            if ( p_port && (*lI)==0 )
                groupSendRecvMemPorts.push_back(p_port);

            // we do not count disconnected ports
            if ( p_port && p_port->p_remotePort &&
                 (p_port->p_remotePort->p_node->type != IB_SW_NODE) ) {
                groupSendRecvMemPorts.push_back(p_port->p_remotePort);
            }
        }
    }

    char buffer[128];
    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);
    cout << "-I- Multicast Group:" << buffer << " has:" << groupSwitches.size()
    << " switches and:" << groupSendRecvMemPorts.size()
    << " FullMember ports" << endl;

    if ( ! groupSwitches.size() ) return(0);
    if ( ! groupSendRecvMemPorts.size() ) return(0);
    // if APort's aren't MFT symmetric, exit before other checks
    is_sharp_mctree = isAggregationNodeInList(groupSwitches);
    if (!is_sharp_mctree) {
        anyErr = checkFabricAPortMFTSymmetry(p_fabric, mlid, groupSwitches);
    }

    anyErr += SubnMgtCheckMCGrpByMemPortLists(p_fabric, mlid,
                                             groupSendRecvMemPorts, sendOnlyPorts);
    return anyErr;
}

///////////////////////////////////////////////////////////////////////////////
// Check all multicast groups :
// 1. all switches holding it are connected
// 2. No loops (i.e. a single BFS with no returns).
int
SubnMgtCheckFabricMCGrps(IBFabric *p_fabric)
{
    int anyErrs = 0;
    cout << "-I- Scanning all multicast groups for loops and connectivity..."
    << endl;
    if (p_fabric->McastGroups.empty()) {
        for ( set_uint16::iterator sI = p_fabric->mcGroups.begin();
                sI != p_fabric->mcGroups.end();
                sI++ )
            anyErrs += SubnMgtCheckMCGrp(p_fabric, *sI);
    } else {
        for (map_mcast_groups::iterator sI = p_fabric->McastGroups.begin();
                sI != p_fabric->McastGroups.end();
                sI++)
            anyErrs+= SubnMgtCheckMCGrp(p_fabric, (*sI).first, (*sI).second);
    }

    if ( anyErrs )
        cout << "-E- " << anyErrs << " multicast group checks failed" << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;
    return anyErrs;
}


///////////////////////////////////////////////////////////////////////////////
struct upDnBfsEntry {
   IBNode *pNode;
   IBNode *pTurnNode;
   uint8_t inPort;
   int     down;
};

// Given a starting point switch for multicast traversal and an mlid
// make sure all traversals are following up/down rules.
// return 1 oif there is a violation
int
SubnReportNonUpDownMulticastGroupFromCaSwitch(IBFabric *p_fabric,
        IBNode *pSwitchNode,
        uint16_t mlid)
{
    map< IBNode *, uint8_t, less < IBNode *> > visitedNodeFromPort;
    list< upDnBfsEntry > nodesQueue;
    upDnBfsEntry thisStep, nextStep;
    rank_t thisNodeRank, remNodeRank;
    int anyErr = 0;
    char buffer[128];
    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);

    thisStep.down = 0;
    thisStep.pNode = pSwitchNode;
    thisStep.inPort = 0; // start as we got in from port 0 we can go out
                        //  any port
    thisStep.pTurnNode = NULL;
    nodesQueue.push_back(thisStep);

    // we do the BFS through the queue
    while (nodesQueue.size()) {
        thisStep = nodesQueue.front();
        nodesQueue.pop_front();

        // we keep track of all visited nodes
        visitedNodeFromPort[thisStep.pNode] = thisStep.inPort;

        // get the list of MC group ports for this mlid of this node
        list_phys_ports portNums = thisStep.pNode->getMFTPortsForMLid(mlid);

        if (!IsRankAssigned(thisStep.pNode))
            exit(1);

        thisNodeRank = thisStep.pNode->rank;
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Visiting node:" << thisStep.pNode->name
                  << " dir:" << (thisStep.down ? "DOWN" : "UP") << endl;

        // go over all MC output ports of the current node ignoring the input
        for (list_phys_ports::iterator pnI = portNums.begin();
                pnI != portNums.end(); pnI++) {
            phys_port_t pn = (*pnI);

            // ignore the port we got here through.
            if (pn == thisStep.inPort)
                continue;

            IBPort *pPort = thisStep.pNode->getPort(pn);
            if (! pPort || ! pPort->p_remotePort)
                continue;

            // get the remote node
            IBNode *pRemNode = pPort->p_remotePort->p_node;

            // we ignore remote HCAs
            if (pRemNode->type != IB_SW_NODE)
                continue;

            // if we already visited this node - it is a loop!
            map< IBNode *, uint8_t, less < IBNode *> >::iterator vI =
                    visitedNodeFromPort.find(pRemNode);

            if (vI != visitedNodeFromPort.end()) {
                int prevPort =  (*vI).second ;
                cout << "-E- Found a loop on MLID:" << buffer
                        << " got to node:" << pRemNode->name
                        << " through port:" << (unsigned int)pPort->p_remotePort->num
                        << " previoulsy visited through port:" << prevPort << endl;
                anyErr++;
                continue;
            }

            if (!IsRankAssigned(pRemNode))
                exit(1);

            remNodeRank = pRemNode->rank;
            // if we are goin up make sure we did not go down before.
            if (remNodeRank < thisNodeRank) {
                // we go up - was it a down?
                if (thisStep.down) {
                    cout << "-E- Non Up/Down on MLID:" << buffer
                            << " turning UP from:" << thisStep.pNode->name
                            << "/P" << (unsigned int)pn << "("<< thisNodeRank << ") to node:"
                            << pRemNode->name << "/P" << (unsigned int)pPort->p_remotePort->num
                            << "(" << remNodeRank << ") previoulsy turned down:"
                            << thisStep.pTurnNode->name << endl;
                    anyErr++;
                    continue;
                } else {
                    // we turned down here so track it:
                    nextStep.down = 0;
                    nextStep.pTurnNode = NULL;
                }
            } else {
                nextStep.pTurnNode = thisStep.pNode;
                nextStep.down = 1;
            }

            // push the node into next steps:
            nextStep.pNode = pRemNode;
            nextStep.inPort = pPort->p_remotePort->num;
            nodesQueue.push_back(nextStep);
        }
    }
    return anyErr;
}


///////////////////////////////////////////////////////////////////////////////
// Find any routes that exist in the FDB's from CA to CA and do not adhare to
// the up/down rules. Report any crossing of the path.
int
SubnReportNonUpDownMulticastGroupCa2CaPaths(IBFabric *p_fabric, uint16_t mlid)
{
    int numBadPaths = 0, paths = 0;
    char buffer[128];
    snprintf(buffer, sizeof(buffer), "0x%04X", mlid);

    cout << "-I- Tracing Multicast Group:" << buffer
            << " CA to CA paths for Credit Loops potential ..."
            << endl;

    // find all HCAs in the MGRP
    list< IBNode *> groupSwitchesConnToHCAs;

    // find all switches that are part of this mcgrp and connects to HCAs
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE)
            continue;

        // see if we have an MFT entry by the given lid:
        list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);

        if (portNums.empty())
            continue;

        // find all HCAs connected to the group by following the links that
        // are marked in the MFT that connect to the group.
        for (list_phys_ports::iterator lI = portNums.begin();
                lI != portNums.end();
                lI++) {
            IBPort *p_port = p_node->getPort(*lI);

            if (p_port && p_port->p_remotePort &&
                    p_port->p_remotePort->p_node->type != IB_SW_NODE) {
                groupSwitchesConnToHCAs.push_back(p_node);
                break;
            }
        }
    }

    cout << "-I- Multicast group:" << buffer << " has:"
            << groupSwitchesConnToHCAs.size()
            << " Switches connected to HCAs" << endl;

    // from each HCA traverse BFS through the switches.
    for (list_pnode::iterator lI = groupSwitchesConnToHCAs.begin();
            lI != groupSwitchesConnToHCAs.end();
            lI++) {
        // avoid too many errors reported:
        if (numBadPaths > 100)
            break;

        numBadPaths +=
                SubnReportNonUpDownMulticastGroupFromCaSwitch(
                        p_fabric, (*lI), mlid);
        paths++;
    }

    if (numBadPaths) {
        if (numBadPaths > 100)
            cout << "-W- Stopped checking multicast groups after 100 errors" << endl;
        cout << "-E- Found:" << numBadPaths
                << " Multicast:" << buffer
                << " CA to CA paths that can cause credit loops." << endl;
    } else
        cout << "-I- No credit loops found traversing:" << paths
               << " leaf switches for Multicast LID:" << buffer << endl;
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Check all multicast groups :
// 1. all switches holding it are connected
// 2. No loops (i.e. a single BFS with no returns).
int
SubnMgtCheckFabricMCGrpsForCreditLoopPotential(IBFabric *p_fabric)
{
    int anyErrs = 0;
    cout << "-I- Scanning all multicast groups for Credit Loops Potential ..." << endl;

    for (set_uint16::iterator sI = p_fabric->mcGroups.begin();
            sI != p_fabric->mcGroups.end();
            sI++)
        anyErrs +=
                SubnReportNonUpDownMulticastGroupCa2CaPaths(p_fabric, *sI);

    if (anyErrs)
        cout << "-E- " << anyErrs << " multicast groups failed" << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;
    return anyErrs;
}


///////////////////////////////////////////////////////////////////////////////
// Fill in the FDB tables assuming the fabric is a fat tree.
// Also assumes the fabric is already ranked from roots downwards.
// We also assume MinHop tables are defined.
//
// LIMITATION: Currently only LMC=0 is supported
//
// The main idea behind this algorithm is that if we have enough
// root ports we can assign each port one destination LID.
// Then we propagate that selection by allowing routing to that LID
// only through that top root and downwards - unless closes at lower levels
//
// ALGORITHM:
// * Initialize a WORKMAP of all HCA LIDs - also count them
// * Calc total switch ports versus number of HCA lids.
// * If we do not have enough switch ports abort.
// * Loop on all rank=0 switches.
//   * For each port select one of the LIDs that it MinHop to and are
//     still in WORKMAP
//   * Traverse forward to that LID assigning LFT
//
// Traverse Forward to LID:
// * Given switch and target LID
// * Find the out-port to be used for going to that LID
//   (use # paths to select from many?)
// * Set FDB to that LID (actually done by the Backward traversal)
// * Recurse to the direction of the LID (need Min hop for that)
// * Perform Backward traversal through all ports connected to lower
//   level switches in-port = out-port
//
// Traverse Backward to LID:
// * Given current switch and LID, in-port
// * Set FDB to target LID to the "in-port"
// * Recurse through all ports connected to lower level switches
//   not including the in-port

static inline void
markPortUtilization(IBPort *p_port)
{
    p_port->counter1++;
}

// given source and destination nodes find the port with lowest
// utilization (subscriptions) and return its number
static phys_port_t
getLowestUtilzedPortFromTo(IBNode *p_fromNode, IBNode *p_toNode)
{
    unsigned int minUtil=0;
    phys_port_t minUtilPortNum=0;
    IBPort *p_port;

    for (unsigned int i = 1; i <= p_fromNode->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        p_port = p_fromNode->getPort(pn);

        if (! p_port)
            continue;
        if (! p_port->p_remotePort)
            continue;
        if (p_port->p_remotePort->p_node != p_toNode)
            continue;

        // the hops should match the min
        if ((minUtilPortNum == 0) || (p_port->counter1 < minUtil)) {
            minUtilPortNum = pn;
            minUtil = p_port->counter1;
        }
    }

    return( minUtilPortNum );
}

// given a node and a target LID find the port that has min hops
// to that LID and lowest utilization
static phys_port_t
getLowestUtilzedPortToLid(IBNode *p_node, lid_t dLid)
{
    unsigned int minUtil = 0;
    phys_port_t minUtilPortNum = 0;
    IBPort *p_port;

    // get the minimal hop count from this node:
    uint8_t minHop = p_node->getHops(NULL,dLid);

    for (unsigned int i = 1; i <= p_node->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        p_port = p_node->getPort(pn);

        if (! p_port)
            continue;
        if (! p_port->p_remotePort)
            continue;

        // the hops should match the min
        if (p_node->getHops(p_port, dLid) == minHop) {
            if ((minUtilPortNum == 0) || (p_port->counter1 < minUtil)) {
                minUtilPortNum = pn;
                minUtil = p_port->counter1;
            }
        }
    }
    return( minUtilPortNum );
}

int
SubnMgtFatTreeBwd(IBNode *p_node, uint16_t dLid, phys_port_t outPortNum)
{
    IBPort* p_port = NULL;

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- SubnMgtFatTreeBwd from:" << p_node->name << " dlid:" << dLid
               << " out-port:" << outPortNum << endl;

    // Set FDB to target LID to the "in-port"
    p_node->setLFTPortForLid(dLid, outPortNum);

    p_port = p_node->getPort(outPortNum);
    if (!p_port) {
        cout << "-E- the node:" << p_node->name
             << " has no port with number:" << outPortNum << endl;
        exit(1);
    }

    // mark this port was utilized
    markPortUtilization(p_port);

    // get the remote node to avoid decending down through it
    IBNode *p_origRemNode = p_port->p_remotePort->p_node;

    // Recurse through all ports connected to lower level switches not
    // including the in-port
    for (unsigned int i = 1; i <= p_node->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        if (pn == outPortNum)
            continue;

        p_port = p_node->getPort(pn);
        if (!p_port || !p_port->p_remotePort)
            continue;

        IBNode *p_remNode = p_port->p_remotePort->p_node;
        // we might have several ports
        if (p_remNode == p_origRemNode)
            continue;
        if (p_remNode->type != IB_SW_NODE)
            continue;
        // avoid going up or sideways in the tree
        if (p_node->rank >= p_remNode->rank)
            continue;

        // avoid loops by inspecting the FDB value of the remote port
        if (p_remNode->getLFTPortForLid(dLid) != IB_LFT_UNASSIGNED)
            continue;

        // select the best port from the remote node to this one:
        phys_port_t remPortNum = getLowestUtilzedPortFromTo(p_remNode, p_node);
        SubnMgtFatTreeBwd(p_remNode, dLid, remPortNum);
    }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
int
SubnMgtFatTreeFwd(IBNode *p_node, uint16_t dLid)
{
    phys_port_t outPortNum = 0;
    IBPort *p_port;

    // Find the out-port to be used for going to that LID
    // (use # paths to select from many?)
    outPortNum = getLowestUtilzedPortToLid(p_node, dLid);

    if (!outPortNum) {
        cout << "-E- fail to find output port for switch:" << p_node->name
                << " to LID:" << dLid << endl;
        exit(1);
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- SubnMgtFatTreeFwd from:" << p_node->name << " dlid:" << dLid
        << " through port:" << outPortNum << endl;

    p_port = p_node->getPort(outPortNum);
    if (!p_port) {
        cout << "-E- the switch:" << p_node->name
                << " has no port with number:" << outPortNum << endl;
        exit(1);
    }

    // Set FDB to that LID (actually done by the Backward traversal)

    // Recurse to the direction of the LID
    if (p_port->p_remotePort->p_node->type == IB_SW_NODE)
        SubnMgtFatTreeFwd(p_port->p_remotePort->p_node, dLid);

    // Perform Backward traversal through all ports connected to lower
    // level switches in-port = out-port
    SubnMgtFatTreeBwd(p_node, dLid, outPortNum);
   return(0);
}


///////////////////////////////////////////////////////////////////////////////
int
SubnMgtFatTreeRoute(IBFabric *p_fabric)
{
    IBNode *p_node;
    IBPort *p_port = NULL;

    cout << "-I- Using Fat Tree Routing" << endl;

    // HACK we currently do not support LMC > 0
    if (p_fabric->caLmc > 0) {
        cout << "-E- Fat Tree Router does not support LMC > 0 yet" << endl;
        return(1);
    }

    list<IBNode*> rootNodes;
    set<lid_t> unRoutedLids;

    int numHcaPorts = 0;
    int numRootPorts = 0;

    // get all root nodes
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;

        // if not a switch just count
        if (p_node->type != IB_SW_NODE) {
            // count all ports that are connected
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                p_port = p_node->getPort((phys_port_t)pn);
                if (p_port && p_port->p_remotePort) {
                    numHcaPorts++;
                    unRoutedLids.insert(p_port->base_lid);
                }
            }
        } else {
            // we only crae now about root switch ports
            if (p_node->rank == 0) {
                rootNodes.push_back(p_node);
                // count all ports that are connected
                for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                    p_port = p_node->getPort((phys_port_t)pn);
                    if (p_port && p_port->p_remotePort)
                        numRootPorts++;
                }
            }
        }
    }

    // do we have enough root ports?
    if (numRootPorts < numHcaPorts) {
        cout << "-E- Can Route Fat-Tree - not enough root ports:"
                << numRootPorts << " < HCA ports:" << numHcaPorts << endl;
        return(1);
    }

    // Loop on all rank=0 switches.
    for (list<IBNode *>::iterator lI = rootNodes.begin();
            lI != rootNodes.end();
            lI++) {
        set<lid_t> switchAllocatedLids;

        p_node = *lI;

        // Foreach port select one of the LIDs that it MinHop to and are
        // still in WORKMAP
        for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
            p_port = p_node->getPort((phys_port_t)pn);
            if (! p_port || !p_port->p_remotePort)
                continue;

            // since most fat trees are fully connected we can simply try from the
            // head of the unRoutedLids...
            int found = 0;
            for ( set<lid_t>::iterator mI = unRoutedLids.begin();
                    !found && (mI != unRoutedLids.end());
                    mI++) {
                // is this dLid in our port min hop?
                uint16_t dLid = *mI;
                // is it in this port min hop?
                if (p_node->getHops(NULL, dLid) == p_node->getHops(p_port, dLid)) {
                    found = 1;
                    // remove from set
                    unRoutedLids.erase(mI);

                    // as it is important we handle lids in order we just collect
                    // them here and later use in order
                    switchAllocatedLids.insert(dLid);

                    // escape the for loop
                    break;
                }
            }
        } // all ports of switch

        if(!p_port) {
            cout << "-E- Cannot Route Fat-Tree - Internal error port is NULL. Node name is "
                 << p_node->getName() << endl;
            continue;
        }

        // now handle all allocated lids of this switch:
        for ( set<lid_t>::iterator alI = switchAllocatedLids.begin();
                alI != switchAllocatedLids.end();
                alI++) {
            lid_t dLid = *alI;
            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                cout << "-V- Routing to LID:" << dLid << " through root port:"
                << p_port->getName() << endl;
            // Traverse forward to that LID assigning LFT
            SubnMgtFatTreeFwd(p_node, dLid);
        }
    } // all rank 0 switches

    // double check no more HCA LIDs to route...
    if (unRoutedLids.size()) {
        cout << "-E- " << unRoutedLids.size()
                   << " lids still not routed:" << endl;
        for( set<lid_t>::iterator sI = unRoutedLids.begin();
                sI != unRoutedLids.end();
                sI++)
            cout << "   " << *sI << endl;
        return(1);
    }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Recursivly DFS backwards through all switch ports
// (except for the port to the lids) and fill in HCAS reached
static int
dfsBackToCAByLftToDLIDs(IBNode *node,
        list<lid_t> &dstLids,
        phys_port_t dstPortNum,
        set<IBNode *> &visitedNodes,
        map<IBPort *, list<lid_t> > &HCAPortsLids)
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-V- Visiting " << node->name << " searching for lids:";
        for (list<lid_t>::const_iterator lI = dstLids.begin();
                lI != dstLids.end(); lI++) cout << *lI << ",";
        cout << endl;
    }

    if (node->type != IB_SW_NODE) {
        IBPort* port = node->getPort(dstPortNum);
        HCAPortsLids[port] = dstLids;
        return(0);
    }

    // first check which dst lid is still valid
    list<lid_t> subDstLids;
    for (list<lid_t>::const_iterator lI = dstLids.begin();
            lI != dstLids.end(); lI++) {
        lid_t lid = *lI;
        if (node->getLFTPortForLid(lid) == dstPortNum) {
            subDstLids.push_front(lid);
        }
    }

    // no paths to the dlids through the port
    if (subDstLids.size() == 0) {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Dead end" << endl;
        return(0);
    }

    // mark the node visited now
    visitedNodes.insert(node);

    // DFS through the other ports
    for (unsigned int i = 1; i <= node->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        if (pn != dstPortNum) {
            IBPort *port = node->getPort(pn);
            if (!port || ! port->p_remotePort) continue;
            // set remPort [IBPort_p_remotePort_get $port]
            // if {$remPort == ""} {continue}
            IBNode *p_remNode = port->p_remotePort->p_node;
            if (visitedNodes.find(p_remNode) == visitedNodes.end()) {
                phys_port_t remPortNum = port->p_remotePort->num;
                dfsBackToCAByLftToDLIDs(p_remNode, subDstLids, remPortNum,
                        visitedNodes, HCAPortsLids);
            }
        }
    }
    return(0);
}

// return 1 if LID is reachable by LFT from given port
/* UNUSED
static int
isThereLFTPathToLID(IBPort *p_port, lid_t lid)
{
    if (p_port->base_lid == lid)
        return(1);

    set<IBNode *> visitedNodes;
    visitedNodes.insert(p_port->p_node);
    while (p_port) {
        // get remote port
        IBPort *p_remPort = p_port->p_remotePort;
        if (!p_remPort)
            return(0);

        if (p_remPort->base_lid == lid)
            return(1);

        // if we already visited we are in a loop
        if (visitedNodes.find(p_remPort->p_node) != visitedNodes.end())
            return(0);

        visitedNodes.insert(p_remPort->p_node);

        // if it is a switch
        if (p_remPort->p_node->type == IB_SW_NODE) {
            phys_port_t pn = p_remPort->p_node->getLFTPortForLid(lid);
            p_port = p_remPort->p_node->getPort(pn);
        } else {
            p_port = NULL;
        }
    }
    return(0);
}
*/
// Obtain all the CA to CA port pairs going through the
// given port
/* UNUSED
int
SubnReportCA2CAPathsThroughSWPort(IBPort *p_port)
{
    // if not a switch port error
    if (p_port->p_node->type != IB_SW_NODE) {
        cout << "-E- Provided port:" << p_port->getName()
	           << " is not a switch port" << endl;
        return(1);
    }

    IBNode *node = p_port->p_node;

    // obtain the DLIDs on the given switch port
    list<lid_t> lidsThroughPort;
    for (lid_t i = 0; i < node->LFT[0].size(); i++) { //!!!PLFT
        if (node->getLFTEntry(i) == p_port->num) {
            // TODO: validate there is really a path to that node
            if (isThereLFTPathToLID(p_port, i)) { //!!!PLFT
                lidsThroughPort.push_front(i);
            } else {
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- LID:" << i
                    << " pointed by LFT but is not reachable from:"
                    << p_port->getName() << endl;
            }
        }
    }
    if (lidsThroughPort.size() == 0) {
        cout << "-W- No paths through port:" << p_port->getName() << endl;
        return(1);
    }

    set<IBNode *> visitedNodes;
    map<IBPort *, list<lid_t> > HCAPortsLids;

    dfsBackToCAByLftToDLIDs(node, lidsThroughPort, p_port->num,
            visitedNodes, HCAPortsLids);
    IBFabric *fabric = node->p_fabric;
    if (HCAPortsLids.size()) {
        map<IBPort *, list<lid_t> >::const_iterator pI;
        for (pI = HCAPortsLids.begin(); pI != HCAPortsLids.end(); pI++) {
            IBPort *port = (*pI).first;
            list<lid_t>::const_iterator lI;
            cout << "From:" << port->getName() << " SLID:" << port->base_lid << endl;
            for (lI = (*pI).second.begin(); lI != (*pI).second.end(); lI++) {
                IBPort *dPort = fabric->getPortByLid(*lI);
                cout << "   To:" << dPort->getName() << " DLID:" << dPort->base_lid
                        << endl;
            }
        }
    }
    return(0);
}
*/

set_lid SubnMgmtGetUsedUnicastLids(IBFabric *p_fabric)
{
    set_lid used_lids;

    //Using of PortByLid and VPortByLid should be safe for XDR.
    //They are both vectors that associate relevnat port to its LID.
    //XDR network could have 4 ports with the same LID but containers should
    //still keep an object.

    //collect lids from all ports
    for (const auto& p_port: p_fabric->PortByLid) {
        if (!p_port)
            continue;

        used_lids.insert(p_port->base_lid);
    }

    //collect lids from all virtual ports
    for (const auto& p_vport: p_fabric->VPortByLid) {
        if (!p_vport)
            continue;

        used_lids.insert(p_vport->get_vlid());
    }

    //collect FLIDs
    for (const auto& entry : p_fabric->RoutersByFLID)
        used_lids.insert(entry.first);

    return used_lids;
}


////////////////////////////////////////////////////////
int
SubnMgtVerifyAREmptyGroups(IBFabric *p_fabric)
{
    cout << "-I- Verifying that empty AR groups are not used in routing tables ... " << endl;

    int anyError = 0;
    const set_lid used_lids = SubnMgmtGetUsedUnicastLids(p_fabric);

    for (u_int8_t plft = 0; plft < MAX_PLFT_NUM; ++plft) {
        for (lid_t dlid : used_lids) {
            //Assume situation that a LID appears in an AR routing table but it is not assigned to any Node.
            //In this case the LID's group could be empty. Take out such LIDs from the validation.
            //I mean the following:
            //0x000c    18             Free           2
            //the lid 0x000c is NOT used and its group 2 could be an empty one.

            for (const auto& p_switch : p_fabric->Switches) {
                if (!p_switch->isAREnable() && !p_switch->isHBFEnable())
                    continue;

                u_int16_t arGroup = p_switch->getARLFTPortGroupForLid(dlid, plft);
                if (!arGroup || arGroup == IB_AR_LFT_UNASSIGNED)
                    continue;

                if (p_switch->isARPortGroupEmpty(arGroup)) {
                    anyError++;
                    cout << "-E- Empty AR group:" << arGroup
                         << " PLFT:" << +plft
                         << " for " << (p_fabric->isFLID(dlid) ? "FLID:" : "LID:") << dlid  
                         << " is found on the switch:" << p_switch->getName()
                         << endl;
                }
            }
        }
    }

    return anyError;
}

////////////////////////////////////////////////////////
int
SubnMgtVerifyAllARCaToCaRoutes(IBFabric *p_fabric)
{
    int anyError = 0, paths = 0;

    //FabricUtilsVerboseLevel = 0xff; //!!!
    g_clearedCache = 0;
    cout << "-I- Verifying all CA to CA AR paths ... " << endl;
    // initialize the histograms:

    // NOTE: since it is OK to have just one of the DLIDs assigned to a single
    // port be available as path from any given source we do not complain
    // about dead paths unless all of them from sPort to dPort are not avaiable

    // go over all ports in the fabric
    ARTraceRouteInfo globalRouteInfo;
    //ARTraceRouteInfo lid2lidRouteInfo;
    ARTraceRouteInfo *pArRouteInfo = NULL;

    ARTraceRouteNodeInfo::prepare(p_fabric);

    set_lid flids;

    //go over all CA dest lid ordered by connected SW
    for ( set_pnode::iterator nI = p_fabric->Switches.begin();
          nI != p_fabric->Switches.end(); nI++ ) {

        IBNode *p_node = *nI;

        //clear cache for new base SW
        ARTraceRouteNodeInfo::clearDB(p_fabric);

        for ( phys_port_t pn = 1; pn <= p_node->numPorts; pn++ ) {
            IBPort *p_port = p_node->getPort((phys_port_t)pn);
            if ( !p_port || !p_port->p_remotePort ||
                 !p_port->p_remotePort->p_node ||
                 p_port->p_remotePort->p_node->type == IB_SW_NODE)
                continue;

            IBPort *p_dstPort = p_port->p_remotePort;
            IBNode *p_dstNode = p_dstPort->p_node;

            // skip ANs
            if (p_port->isSpecialPort() &&
                    p_port->getSpecialPortType() != IB_SPECIAL_PORT_ROUTER)
                continue;

            lid_t dLid;
            uint8_t lmc;
            p_dstNode->getLidAndLMC(p_dstPort->num, dLid, lmc);

            // skip zero dlid
            if (dLid == 0)
                continue;

            // go over all LMC values of the destination:
            lid_t pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
            set_lid all_lids;

            // go over all LMC combinations:
            for ( lid_t l = 0; l < pathsPerDlid; l++ )
                all_lids.insert(lid_t(dLid + l));

            // go over all vports:
            map_vportnum_vport::iterator vports_iter = p_dstPort->VPorts.begin();
            for ( ; vports_iter != p_dstPort->VPorts.end(); ++vports_iter) {
                IBVPort *p_vport =  vports_iter->second;
                if (!p_vport ||
                    !p_fabric->getVPortByLid(p_vport->get_vlid())||
                    p_vport->getVPortNum() == 0)
                    continue;
                all_lids.insert(p_vport->get_vlid());
            }

            // go over all FLIDs for routers
            if (p_dstNode->type == IB_RTR_NODE) {
                for (set_lid::iterator it = p_dstNode->remote_enabled_flids.begin();
                        it != p_dstNode->remote_enabled_flids.end(); ++it) {
                    // check if this flid already checked
                    if (flids.find(*it) == flids.end()) {
                        flids.insert(*it);
                        all_lids.insert(*it);
                    }
                }
            }

            // go over all dlids of this node
            for (set_lid::iterator lids_iter = all_lids.begin();
                    lids_iter != all_lids.end(); ++lids_iter) {

                //check cache for new dLID, might clear DB
                ARTraceRouteNodeInfo::checkDB(p_fabric, *lids_iter);

                // go over all the rest of the ports - but only once - so skip non base lids
                for ( set_pnode::iterator itrSw = p_fabric->Switches.begin();
                        itrSw != p_fabric->Switches.end(); itrSw++ ) {

                    IBNode *p_node = *itrSw;
                    list_lid slids;

                    if (p_node->routing_cache.CA_lids.empty())
                        continue;

                    if (p_node->routing_cache.representative)
                        slids.push_back(p_node->routing_cache.CA_slid);
                    else
                        slids = p_node->routing_cache.CA_lids;

                    for (list_lid::iterator it = slids.begin(); it != slids.end(); it++ ) {
                        lid_t sLid = *it;

                        // Avoid tracing to itself
                        if ( sLid == dLid )
                            continue;

                        const IBPort *p_source_port = p_fabric->getPortByLid(sLid);

                        if (p_source_port &&
                                p_source_port->p_node &&
                                p_source_port->p_node->skipRoutingChecks)
                            continue;

                        //avoid checking router to flid or router to router.
                        //should not be any routing path from router to flid.
                        //routing from router to router has never been checked before
                        //and it may not exist either
                        if (p_source_port && p_source_port->p_node &&
                                p_source_port->p_node->type == IB_RTR_NODE &&
                                (p_fabric->isFLID(*lids_iter) || p_dstNode->type == IB_RTR_NODE)) {
                            if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                                cout << "-V- Skip validation from lid:" << sLid
                                     << " (router: " <<  p_source_port->p_node->getName()
                                     << ") to lid:" << *lids_iter
                                     <<  endl;

                            continue;
                        }

                        // if dlid is flid - init all routers.
                        bool isFLID = p_fabric->isFLID(*lids_iter);
                        if (isFLID)
                            for (IBNode* p_router : p_fabric->Routers)
                                 p_router->appData3.val = 0;

                        paths++;

                        // now go and verify the path:
                        ARTraceRouteByLFT(p_fabric, sLid, *lids_iter, pArRouteInfo);
                        if (pArRouteInfo) {
                            //ARTraceRouteByLFT returns RouteInfo of connected SW
                            //so we can use it as a child in updateRouteStatistics
                            globalRouteInfo.updateRouteStatistics(pArRouteInfo);
                            if ( pArRouteInfo->getGoodPathCount() == 0 )
                                anyError++;
                        } else
                            anyError++;

                        // if destination lid is flid -
                        // check that if we reach a router the flid is enabled on it.
                        if (isFLID)
                            for (IBNode* p_router : p_fabric->Routers)
                                if (p_router->appData3.val &&
                                        p_router->remote_enabled_flids.find(*lids_iter) == p_router->remote_enabled_flids.end()) {
                                    anyError++;

                                    cout << "-E- Dead end to flid:" << *lids_iter
                                         << " at router:" << p_router->name
                                         << ". The flid is not enabled on the router" << endl;
                                }
                    } // all slids
                } // for on sources
            } // all_lids (lmc values + vlids)

            /*
            if ( lid2lidRouteInfo.getGoodPathCount() == 0 ) {
                cout << "-E- Fail to find a path from:"
                << p_srcPort->p_node->name << "/"
                << (unsigned int)p_srcPort->num
                << " to:" << p_dstPort->p_node->name << "/"
                << (unsigned int)p_dstPort->num
                << endl;
                anyError++;
            }
            lid2lidRouteInfo.clearRouteStatistics();
            */

        } // all ports connected to same SW
    } // all SW

    globalRouteInfo.dumpRouteStatistics();

    if ( anyError )
        cout << "-E- Found " << anyError << " missing paths"
        << " out of:" << paths << " paths" << endl;
    else
        cout << "-I- Scanned:" << paths << " CA to CA paths " << endl;


    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-I- Cashe was cleard:" << g_clearedCache
            << " times" << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;

    ARTraceRouteNodeInfo::cleanup(p_fabric);

    return anyError;
}

int
SubnMgtVerifyFNMPorts(IBFabric *p_fabric, bool useAR)
{
    int anyError = 0;

    // for each node with FNM ports, save the lids that are routed from it's FNM ports
    map<IBNode *, set<lid_t>> fnm_routed_lids;

    // Check that FNM ports are only used to route LIDs inside the same system
    for (IBNode *p_node : p_fabric->Switches) {
        if (!p_node || !p_node->hasFNMPort())
            continue;
        fnm_routed_lids[p_node];
        for (u_int8_t pLFT = 0; pLFT <= p_node->getMaxPLFT(); ++pLFT) {
            u_int16_t lfdbTop = (u_int16_t)p_node->LFT[pLFT].size();
            if (p_node->isPLFTEnabled())
                lfdbTop = p_node->getLFDBTop(pLFT);

            for (lid_t dst_lid = 1; dst_lid <= lfdbTop; ++dst_lid) {
                list<phys_port_t> lft_ports;
                IBPort *p_dest_port = p_fabric->getPortByLid(dst_lid);
                if (!p_dest_port)
                    continue;

                p_node->getLFTPortListForLid(dst_lid, pLFT, useAR, lft_ports);
                for (phys_port_t lft_port_num : lft_ports) {
                    IBPort *lft_port = p_node->getPort(lft_port_num);
                    if (!lft_port || !lft_port->isFNMPort())
                        continue;

                    fnm_routed_lids[p_node].insert(dst_lid);

                    if (p_dest_port->p_node->system_guid_get() != p_node->system_guid_get()) {
                        cout << "-E- FNM Port " << lft_port->getName() << " used to route LID "
                             << dst_lid << " in different system" << endl;
                        anyError++;
                    }
                } // all out ports
            } // all LIDs
        } // all pLFTs
    } // all Swtiches

    // For each BlackMamba node, check that all other nodes in the
    // system can be reached via FNM ports.
    for (IBNode *p_node : p_fabric->Switches) {
        // target LIDs for FNM
        if (!p_node || !p_node->hasFNMPort())
            continue;

        // check that node is BlackMamba
        if (p_node->getSuitablePlane() == -1 || p_node->isPrismaSwitch())
            continue;

        set<lid_t>& lids = fnm_routed_lids[p_node];
        // go over all other nodes in the current BlackMamba system, and make
        // sure that they are routed to by FNM ports of the current node.
        for (auto& node_name_pair : p_node->p_system->NodeByName) {
            IBNode *p_other_node = node_name_pair.second;
            if (!p_other_node || (p_other_node == p_node) || !p_other_node->hasFNMPort())
                continue;
            if (p_other_node->getSuitablePlane() == -1 || p_other_node->isPrismaSwitch())
                continue;
            // skip planes without LIDs
            if (!p_other_node->getFirstLid())
                continue;

            if (!lids.count(p_other_node->getFirstLid())) {
                cout << "-E- FNM Ports on " << p_node->getName() 
                     << " do not route to " << p_other_node->getName() 
                     << " (LID " << p_other_node->getFirstLid() << ")"
                     << " in the same system" << endl;
                anyError++;
            }
        }
    }

    return anyError;
}

int SubnMgtCountSkipRoutingChecksNodes(IBFabric *p_fabric)
{
    int rc = 0;
    uint64_t num = 0;

    //go over all Nodes
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
          nI != p_fabric->NodeByName.end(); nI++ ) {

        IBNode *p_node = nI->second;

        if (p_node->skipRoutingChecks)
          num++;
    }

    if (num)
        cout << "-W- " << num << " nodes are being skipped from routing checks"
             << " due to data retrieving issues." << endl;

    return rc;
}

int
SubnMgtCheckSL2VLTables(IBFabric *p_fabric)
{
    int anyError = 0;

    cout << "-I- Check all SL2VL Tables ... " << endl;

    //go over all Nodes
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
          nI != p_fabric->NodeByName.end(); nI++ ) {

        IBNode *p_node = (*nI).second;

        if (p_node->skipRoutingChecks)
          continue;

        anyError += p_node->checkSL2VLTable();
    }

    if(anyError == 0)
        cout << "-I- All SL2VL tables are OK." << endl;

    return anyError;
}

string ports_str(list_phys_ports& ports)
{
    stringstream sstr;

    if (ports.size()) {
        list_phys_ports::iterator it = ports.begin();
        sstr << +*it;
        it++;

        for (; it != ports.end(); it++)
            sstr << ", " << +*it;
    }

    return sstr.str();
}

int SubnMgtValidateARRouting(IBFabric *p_fabric)
{
    int anyError = 0;
    set_pnode::iterator swI;
    IBNode *cur_switch;
    list_phys_ports min_hop_ports;
    list_phys_ports ar_lft_ports;
    list_phys_ports::iterator mh_iter;
    list_phys_ports::iterator ar_iter;
    phys_port_t staticPort = 0;
    uint16_t group = 0;

    // check that routing algo is 'ar_updn' or 'ar_ftree'.
    if (p_fabric->GetRoutingEngine() != IB_ROUTING_AR_UPDN &&
            p_fabric->GetRoutingEngine() != IB_ROUTING_AR_FTREE) {
        cout << "-E- AR Validation is for ar_updn or ar_ftree only." << endl;
        anyError = 1;
        return anyError;
    }

    // Build Up-Down MinHop tables on switches
    anyError = SubnMgtCalcUpDnMinHopTbls(p_fabric);
    if (anyError) {
        cout << "-E- Failed to build Up-Down MinHop Tables." << endl;
        return anyError;
    }

    cout << "-I- Comparing AR LFTs against calculated Up-Down MinHop Tables." << endl;
    bool success = true;

    // go over all switch nodes:
    for (swI = p_fabric->Switches.begin(); swI != p_fabric->Switches.end(); swI++) {
        cur_switch = *swI;

        // go over all lids
        for (lid_t lid = p_fabric->minLid; lid <= p_fabric->maxLid; lid++) {

            IBPort* p_port = p_fabric->PortByLid[lid];

            if (!p_port || !p_port->p_node ||             // skip lids that not in the fabric.
                    p_port->p_node->type == IB_SW_NODE || // skip switch lids.
                    p_port->p_node->isSpecialNode())      // skip special nodes (ANs) lids.
                continue;

            bool failed = false;
            min_hop_ports.clear();
            ar_lft_ports.clear();

            // go over all ports
            if (cur_switch->MinHopsTable[lid][0] != IB_HOP_UNASSIGNED) {
                for (phys_port_t pn = 1; pn <= cur_switch->numPorts; pn++) {
                    // collect all min-hop ports
                    // MinHopsTable[lid][0] holds the min-hop from all ports
                    if (cur_switch->MinHopsTable[lid][pn] == cur_switch->MinHopsTable[lid][0])
                        min_hop_ports.push_back(pn);
                } // all ports
            }

            // take the AR ports group
            // for trees topologies there is only PLFT 0
            staticPort = cur_switch->getLFTPortForLid(lid, 0);
            group = cur_switch->getARLFTPortGroupForLid(lid, 0);
            cur_switch->getLFTPortListForLid(staticPort, group, ar_lft_ports);


            /*** compare min_hop_ports and ar_lft_ports ***/

            // in case there is no port in MinHop Table but tere is one port in AR LFT
            // we shouldn't report an error because it can happen for example if a CA
            // is connected to a root switch and there is no updn MinHop route from other
            // root switches to this CA.
            if (min_hop_ports.empty() && ar_lft_ports.size() == 1) {
                cout << "-I- Found 1 port in AR LFT and no ports in calculated "
                     << "Up-Down MinHop Tables." << endl
                     << "    On switch: " << cur_switch->name << ", for lid: " << lid << endl
                     << "    ar_lft_port = "  << ports_str(ar_lft_ports)  << endl;
                continue;
            }

            // more ports in AR LFT than in updn MinHop tables - Error.
            if (ar_lft_ports.size() > min_hop_ports.size()) {
                cout << "-E- Found more ports in AR LFT than in calculated "
                     << "Up-Down MinHop Tables." << endl
                     << "    On switch: " << cur_switch->name << ", for lid: " << lid << endl;
                failed = true;
            }

            // less ports in AR LFT than in updn MinHop tables - Warning.
            if (ar_lft_ports.size() < min_hop_ports.size()) {
                cout << "-W- Found less ports in AR LFT than in calculated "
                     << "Up-Down MinHop Tables." << endl
                     << "    On switch: " << cur_switch->name << ", for lid: " << lid << endl;
                failed = true;
            }

            // same size but different port - Warning.
            if (ar_lft_ports.size() == min_hop_ports.size()) {
                mh_iter = min_hop_ports.begin();
                ar_iter = ar_lft_ports.begin();
                for (; mh_iter != min_hop_ports.end(); mh_iter++, ar_iter++) {
                    if (*mh_iter != *ar_iter) {
                        cout << "-W- Found different ports in AR LFT than in "
                                "calculated Up-Down MinHop Tables." << endl
                             << "    On switch: " << cur_switch->name << ", for lid: " << lid << endl;
                        failed = true;
                        break;
                    }
                }
            }

            // if pFRN enabled check that static port is part of AR group.
            if (cur_switch->ispFRNEnabled()) {

                if ((find(ar_lft_ports.begin(), ar_lft_ports.end(), staticPort)) == ar_lft_ports.end()) {

                    cout << "-W- pFRN enabled but static port is not part of AR group." << endl
                         << "    On switch: " << cur_switch->name << ", for lid: " << lid << endl;
                    success = false;
                }
            }

            if (failed) {
                cout << "    min_hop_ports = " << ports_str(min_hop_ports) << endl;
                cout << "    ar_lft_ports = "  << ports_str(ar_lft_ports)  << endl;
                success = false;
            }
        } // all lids
    } // all switches

    if (success)
        cout << "-I- AR Validation finished successfully!" << endl;

    return anyError;
}

