/*
 * Copyright (c) 2004-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <dirent.h>
#include <unistd.h>
#include <getopt.h>
#include <fstream>

#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/CredLoops.h>
#include <infiniband/ibdm/IBDiagnetFileNames.h>

#define IBDMCHK_FLAG_VERSION 0X10000

using namespace std;

void
show_usage() {
  cout << "Usage: there are two modes: Design/Verify" << endl;
  cout << "  Design: ibdmchk [-v][-h][--version][-u][-r <roots file>] -t <topology file> -n <SM Node> -p <SM Port> [-e] [-l <lmc>]" << endl;
  cout << "  Verify: ibdmchk [-v][-h][--version][-u][-r <roots file>] [-s <subnet file>] [-f <fdb file>] [-m <mcfdb file>] [-M] [-a] -l <lmc>[:<sw-lmc>] \n";
  cout << "                             [-c <path-sl file>] [-d <sl2vl file>] [-P <PLFT pile>] [-V <vl2vl file>] [-A <AR file>] [-F <FAR file>] \n"
          "                             [-L <FLID file>] [-D <sa-dump file>] [-I <hierarchy info file>] [-o <output directory>]\n\n" << endl;
}

void
show_help() {
  cout
    << " IBDM Topology - Checker\n"
    << "--------------------------\n"
    << "\n"
    << "Usage: \n"
    << " ibdmchk has two operation modes: cluster design and verification.\n"
    << "\n"
    << "  CLUSTER DESIGN:\n"
    << "  Usage:\n"
    << "   ibdmchk [-v][-h][--version][-u][-r <roots file>] -t <topology file> -n <SM Node> -p <SM Port> [-e] [-l <lmc>]\n\n"
    << "  Description:\n"
    << "   The Design mode is intended to be used before the cluster is built. It provides \n"
    << "   basic checks of the specified network as described by a topology file.\n"
    << "   After simulating the SM LID assignment and routing algorithms it provides reports\n"
    << "   of the CA to CA paths depth histogram and credit deadlock potential \n"
    << "   in the resulting routing scheme.\n\n"
    << "  Arguments (required):\n"
    << "  -t|--topology <file> = the topology file specifying the fabric.\n"
    << "  -n|--node <SM Node> = the name of the Subnet Manager node (syntax: <Topo-File-System>/U1)\n"
    << "  -p|--port <SM Port> = the port number by which the SM nodes is attached to the fabric.\n"
    << "\n"
    << "  Options:\n"
    << "  -v|--verbose = verbose mode\n"
    << "  -h|--help = provide this help message\n"
    << "  --version = Prints the version of the tool\n"
    << "  -l|--lmc <lmc> = LMC value > 0 means assigning 2^lmc lids to each port.\n"
    << "  -e|--enh = use enhanced routing algorithm when LMC > 0 and report the resulting paths\n"
    << "       correlation (using same system/node) histogram\n"
    << "  -u|--updn = use up/down routing algorithm instead of OpenSM min-hop.\n"
	 << "       Also selects Up/Down credit loop check algorithm.\n"
    << "  -r|--roots <file> = a file with all the roots node names (one on each line).\n"
    << "\n"
    << "  CLUSTER VERIFICATION:\n"
    << "  Usage:\n"
    << "   ibdmchk [-v][-h][--version][-u][-r <roots file>] [-s <subnet file>] [-f <fdb file>] [-m <mcfdb file>] [-M] [-a] -l <lmc>[:<sw-lmc>] \n"
    << "                  [-c <path-sl file>] [-d <sl2vl file>] [-P <PLFT pile>] [-V <vl2vl file>] [-A <AR file>] [-F <FAR file>] \n"
    << "                  [-D <sa-dump file>] [-I <hierarchy info file>] [-E <entry plane filter file>] [-o <output directory>] \n\n"
    << "  Description:\n"
    << "   After the cluster is built and OpenSM is run (using flag -D 0x43) it reports the\n"
    << "   subnet and FDB tables into the files osm-subnet.lst, osm.fdbs and osm.fdbs in\n"
    << "   /var/log/ (or subnet.lst, osm.fdbs and osm.mcfdbs into /tmp in older versions).\n"
    << "   If more than one SL is known to be used additional file holding CAxCA->SL mapping \n"
    << "   is generated (format: 0xsrc_guid dlid sl) . In this case the SL2VL mapping is \n"
    << "   optionally supplied in an additional file (format: 0xsw_guid inport outport 0x(sl0)(sl1),\n"
    << "   0x(sl2)(sl3),...). Without SL2VL mapping file an identity mapping is used.\n"
    << "   Based on these files the utility checks all CA to CA connectivity. Further analysis\n"
    << "   for credit deadlock potential is performed and reported. \n"
    << "   In case of an LMC > 0 it reports histograms for how many systems and nodes\n"
    << "   are common between the different paths for the same port pairs.\n"
    << "\n"
    << "  Arguments (required):\n"
    << "  -l|--lmc <lmc>[:<sw-lmc>] = LMC value > 0 means assigning 2^lmc lids to each port.\n"
    << "       if the given parameter is in the format <digit>:<digit> then the first digit\n"
    << "       is the LMC used for CAs and the second is the LMC for Switches.\n"
    << "\n"
    << "  Options:\n"
    << "  -v|--verbose = verbose mode\n"
    << "  -h|--help = provide this help message\n"
    << "  --version = Prints the version of the tool\n"
    << "  -R|--dir = Path to ibdiagnet's output directory to be used for input file."
    << "     overrun by normal file flags (like --slvl, --mcfdb, and others).\n"
    << "  -s|--subnet <file> = OpenSM subnet.lst file (/var/log/osm-subnet.lst or /tmp/subnet.lst)\n"
    << "  -f|--fdb <file> = OpenSM dump of Ucast LFDB. Use -D 0x41 to generate it.\n"
    << "     (default is /var/log/osm.fdbs or /tmp/osm.fdbs).\n"
    << "  -m|--mcfdb <file> = OpenSM dump of Multicast LFDB. Use -D 0x41 to generate it.\n"
    << "     (default is /var/log/osm.mcfdbs or /tmp/osm.mcfdbs).\n"
    << "  -c|--psl <file> = CAxCA->SL mapping. Each line holds: srcguid dlid sl \n"
    << "  -d|--slvl <file> = SL2VL mapping. Each line holds: swguid iport oport 0x(sl0)(sl1) 0x(sl2)(sl3)...\n"
    << "  -r|--roots <file> = a file holding all root nodes guids (one per line).\n"
    << "  -P|--PLFT <file> = a file holding private LFT mapping on switches.\n"
    << "  -V|--vl2vl <file> = a file holding vl2vl configuration on switches.\n"
    << "  -A|--AR <file> = a file holding adaptive routing configuration on switches.\n"
    << "  -F|--FAR <file> = a file holding full adaptive routing configuration on switches.\n"
    << "  -L|--FLID <file> = a file holding FLID configuration on routers.\n"
    << "  -D|--sa_dump <file> = a file holding multicast group definitions, used for mcast credit loop check.\n"
    << "  -I|--phi <file> = a file holding port hierarchy info.\n"
    << "  -E|--epf <file> = a file holding entry plane filter.\n"
    << "  -o|--outdir <directory> = the directory for analysis report files.\n"
    << "  -u|--updn = selects Up/Down credit loop check algorithm rather than the generic one.\n"
    << "  -M|--MFT = include multicast routing in credit loops analysis.\n"
    << "  -a|--all = verify all paths: not only CA-CA but also SW-CA and SW-SW\n"
    << "\n"
    << "Author: Eitan Zahavi, Mellanox Technologies LTD.\n"
    << endl;
}

list <IBNode *>
ParseRootNodeNamesFile( IBFabric *p_fabric, string fileName)
{
  ifstream namesFile;
  namesFile.open(fileName.c_str(), ifstream::in);
  if (!namesFile.is_open())
  {
    cout << "-F- Given roots file could not be opened!" << endl;
    exit(1);
  }

  char name[256];
  list <IBNode *> roots;
  cout << "-I- Parsing Root Node Names from file:" << fileName << endl;
  namesFile.getline(name,256);
  while (namesFile.good())
  {
    IBNode *pNode = p_fabric->getNode(string(name));
    if (!pNode)
    {
      cout << "-E- Fail to find node:" << name << endl;
    }
    else
    {
      roots.push_back(pNode);
    }
    namesFile.getline(name,256);
  }
  cout << "-I- Defined " << roots.size() << " root nodes" << endl;
  namesFile.close();
  return roots;
}

list <IBNode *>
ParseRootNodeGuidsFile( IBFabric *p_fabric, string fileName)
{
  ifstream guidsFile;
  guidsFile.open(fileName.c_str(), ifstream::in);
  if (!guidsFile.is_open())
  {
    cout << "-F- Given roots file could not be opened!" << endl;
    exit(1);
  }

  uint64_t guid;
  list <IBNode *> roots;

  string line;

  guidsFile >> line;
  while (guidsFile.good())
  {
    guid = strtoull(line.c_str(), NULL, 0);

    IBNode *pNode = p_fabric->getNodeByGuid(guid);
    if (!pNode)
    {
      char buffer[20];
      snprintf(buffer, sizeof(buffer), "0x%016" PRIx64, guid);
      cout << "-E- Fail to find node guid:" << buffer << endl;
    }
    else
    {
      roots.push_back(pNode);
    }
    guidsFile >> line;
  }

  guidsFile.close();
  return roots;
}

class FileNames {
public:
  string subnetFile = string("");
  string fdbFile = string("");
  string mcFdbFile = string("");
  string TopoFile = string("");
  string ibdiagnetOutputDir = string("");
  string SmNodeName = string("");
  string PSLFile = string("");
  string SLVLFile = string("");
  string VL2VLFile = string("");
  string ARFile = string("");
  string FARFile = string("");
  string FLIDFile = string("");
  string SADumpFile = string("");
  string PHIFile  = string("");
  string EPFFile  = string("");
  string PLFTFile = string("");
  string RootsFileName = string("");
  string outFilesDir = string("");

  /**
   * @brief use the folder in ibdiagnetOutputDir to
   *        fill all variables in the class that have existing files.
   */
  int setFromIbdiagnetOutputDir();
};

int FileNames::setFromIbdiagnetOutputDir() {
  int rc = 0;
  string dir_path = ibdiagnetOutputDir;
  if (!dir_path.size())
    return rc;

  if (dir_path.back() != '/')
      dir_path += '/';

  DIR *dir = nullptr;
  dir = opendir(dir_path.c_str());
  int err = errno;
  string input_type;
  if (!dir) {
    if (err == ENOENT) {
      cout << "-E- ibdiagnet output folder " << dir_path << "doesn't exist." << endl;
    } else if (err == ENOTDIR) {
      cout << "-E- path given as ibdiagnet output folder " << dir_path
            << "isn't a folder." << endl;
    } else {
      cout << "-E- Failed to open ibdiagnet output folder "
            << dir_path << "." << endl;
    }

    return -1;
  }

  struct dirent *file_entry;

  while ((file_entry = readdir(dir))) {
    string curr_file_name = file_entry->d_name;

    if (!subnetFile.size() && (curr_file_name == LST_FILE)) {
      subnetFile = dir_path + curr_file_name;
      cout << "-I- Using LST file: " << subnetFile << endl;
    }else if (!fdbFile.size() && (curr_file_name == UC_FDBS_FILE)) {
      fdbFile = dir_path + curr_file_name;
      cout << "-I- Using FDBS file: " << fdbFile << endl;
    }else if (!mcFdbFile.size() && (curr_file_name == MC_FDBS_FILE)) {
      mcFdbFile = dir_path + curr_file_name;
      cout << "-I- Using MCFDBS file: " << mcFdbFile << endl;
    }else if (!SLVLFile.size() && (curr_file_name == SLVL_FILE)) {
      SLVLFile = dir_path + curr_file_name;
      cout << "-I- Using SLVL file: " << SLVLFile << endl;
    }else if (!VL2VLFile.size() && (curr_file_name == VL2VL_FILE)) {
      VL2VLFile = dir_path + curr_file_name;
      cout << "-I- Using VL2VL file: " << VL2VLFile << endl;
    }else if (!ARFile.size() && (curr_file_name == AR_FILE)) {
      ARFile = dir_path + curr_file_name;
      cout << "-I- Using AR file: " << ARFile << endl;
    }else if (!FARFile.size() && (curr_file_name == FAR_FILE)) {
      FARFile = dir_path + curr_file_name;
      cout << "-I- Using FAR file: " << FARFile << endl;
    }else if (!FLIDFile.size() && (curr_file_name == FLID_DUMP_FILE)) {
      FLIDFile = dir_path + curr_file_name;
      cout << "-I- Using FLID file: " << FLIDFile << endl;
    }else if (!PHIFile.size() && (curr_file_name == PORT_HI_INFO_FILE)) {
      PHIFile = dir_path + curr_file_name;
      cout << "-I- Using Hierarchy Info file: " << PHIFile << endl;
    }else if (!EPFFile.size() && (curr_file_name == END_PORT_PF_FILE)) {
      EPFFile = dir_path + curr_file_name;
      cout << "-I- Using EPF file: " << EPFFile << endl;
    }else if (!PLFTFile.size() && (curr_file_name == PLFT_FILE)) {
      PLFTFile = dir_path + curr_file_name;
      cout << "-I- Using PLFT file: " << PLFTFile << endl;
    }
  }

  closedir(dir);

  return rc;
}

int main (int argc, char **argv) {

  FabricUtilsVerboseLevel = FABU_LOG_ERROR;
  FileNames names;
  int EnhancedRouting = 0;
  uint8_t lmc = 0;
  uint8_t swLmc = 0;
  int SmPortNum  = -1;
  int UseUpDown = 0;
  int AllPaths = 0;
  int CheckMFTCredLoops = 0;
  int CheckSWUnicastPathsCredLoops = 0;
  char *pColon;

  /*
   * Parsing of Command Line
   */

  int next_option;
  const char * const short_option = "vho:R:l:s:f:m:et:p:n:uar:c:d:V:A:F:L:D:I:E:P:M";
  /*
	 In the array below, the 2nd parameter specified the number
	 of arguments as follows:
	 0: no arguments
	 1: argument
	 2: optional
  */
  const option long_option[] =
	 {
		{  "verbose",  1,   NULL,   'v'},
		{  "help",     0,   NULL,   'h'},
		{  "version",  0,   NULL,   IBDMCHK_FLAG_VERSION},
		{  "dir",      1,   NULL,   'R'},
		{  "lmc",      1,   NULL,   'l'},
		{  "subnet",   1,   NULL,   's'},
		{  "fdb",      1,   NULL,   'f'},
		{  "mcfdb",    1,   NULL,   'm'},
		{  "psl",      1,   NULL,   'c'},
		{  "slvl",     1,   NULL,   'd'},
		{  "vl2vl",    1,   NULL,   'V'},
		{  "AR",       1,   NULL,   'A'},
		{  "FAR",      1,   NULL,   'F'},
		{  "FLID",     1,   NULL,   'L'},
		{  "sa_dump",  1,   NULL,   'D'},
		{  "phi",      1,   NULL,   'I'},
		{  "epf",      1,   NULL,   'E'},
		{  "PLFT",     1,   NULL,   'P'},
		{  "outdir",   1,   NULL,   'o'},
		{  "topology", 1,   NULL,   't'},
		{  "node",     1,   NULL,   'n'},
		{  "port",     1,   NULL,   'p'},
		{  "enh",      0,   NULL,   'e'},
		{  "updn",     0,   NULL,   'u'},
		{  "roots",    1,   NULL,   'r'},
		{  "all",      0,   NULL,   'a'},
		{  "MFT",      0,   NULL,   'M'},
		{  NULL,       0,   NULL,	 0 }	/* Required at the end of the array */
	 };

  printf("-------------------------------------------------\n");
  do
  {
	 next_option = getopt_long(argc, argv, short_option,long_option, NULL);
	 switch(next_option)
	 {
	 case 'h':
		show_help();
		return 0;
		break;

         case IBDMCHK_FLAG_VERSION:
		printf("ibdmchk version: %s\n", get_ibdm_version());
		return 0;

	 case 'v':
		/*
		  Verbose Mode
		*/
		FabricUtilsVerboseLevel |= FABU_LOG_VERBOSE;
		printf(" Verbose Mode\n");
		break;

         case 'l':
                /*
                  Specifies lmc
                */
                if ((pColon = strchr(optarg,':'))) {
                  lmc = (uint8_t)atoi(optarg);
                  swLmc = (uint8_t)atoi(pColon+1);
                } else {
                  lmc = (uint8_t)atoi(optarg);
                  swLmc = (uint8_t)lmc;
                }
                break;

	 case 'o':
         	/*
          	*     Output
          	*/
       		names.outFilesDir = string(optarg);
       		break;
   case 'R':
		/*
		  Specifies input directory (ibdiagnet output)
		*/
		names.ibdiagnetOutputDir = string(optarg);
		break;

	 case 'f':
		/*
		  Specifies FDB
		*/
		names.fdbFile = string(optarg);
		break;

	 case 'm':
		/*
		  Specifies Multicast FDB
		*/
		names.mcFdbFile = string(optarg);
		break;

	 case 's':
		/*
		  Specifies Subnet
		*/
		names.subnetFile = string(optarg);
		break;

	 case 'c':
	        /*
		  Specifies CAxCA->SL
		*/
	        names.PSLFile = string(optarg);
	        break;

	 case 'd':
	        /*
		  Specifies SL->VL
		*/
	        names.SLVLFile = string(optarg);
	        break;

	 case 'V':
	        /*
		  Specifies VL2VL
		*/
	        names.VL2VLFile = string(optarg);
	        break;

	 case 'A':
	        /*
		  Specifies AR
		*/
		if (!names.FARFile.empty()) {
	            cout << "-E- AR flag and FAR can't be running together." << endl;
	            exit(1);
	        }
	        names.ARFile = string(optarg);
	        break;

	 case 'F':
	        /*
		  Specifies FAR
		*/
		if (!names.ARFile.empty()) {
	            cout << "-E- FAR flag and AR can't be running together." << endl;
	            exit(1);
	        }
	        names.FARFile = string(optarg);
	        break;

	 case 'L':
	        /*
		  Specifies FLID file
		*/
	        names.FLIDFile = string(optarg);
	        break;

	 case 'D':
	        /*
		  Specifies AR
		*/
	        names.SADumpFile = string(optarg);
	        break;
	 case 'I':
	        /*
	          Port Hierarchy Info
	        */
	        names.PHIFile = string(optarg);
	        break;
	 case 'E':
	        /*
	          Entry Plane Filter
	        */
	        names.EPFFile = string(optarg);
	        break;
	 case 'P':
	        /*
		  Specifies PLFT
		*/
	        names.PLFTFile = string(optarg);
	        break;

	 case 't':
		/*
		  Specifies Topology
		*/
		names.TopoFile = string(optarg);
		break;

	 case 'n':
		/*
		  Specifies SM Node
		*/
		names.SmNodeName = string(optarg);
		break;

	 case 'p':
		/*
		  Specifies SM Port Num
		*/
		SmPortNum = atoi(optarg);
		break;

	 case 'e':
		/*
		  Enhanced Routing
		*/
		EnhancedRouting = 1;
		break;

	 case 'r':
		/*
		  Use Roots File
		*/
		names.RootsFileName = string(optarg);
		break;

	 case 'u':
		/*
		  Use Up Down Routing
		*/
		UseUpDown = 1;
		break;

	 case 'a':
		/*
		  Verify ALL paths
		*/
		AllPaths = 1;
		CheckSWUnicastPathsCredLoops = 1;
		break;
	 case 'M':
		/*
		  Include MFT in credit loops check
		*/
		CheckMFTCredLoops = 1;
		break;

	 case -1:
		break; /* done with option */
	 default: /* something wrong */
		show_usage();
		exit(1);
	 }
  }
  while(next_option != -1);

  if (names.ibdiagnetOutputDir.size()) {
    names.setFromIbdiagnetOutputDir();
  }

  IBFabric fabric;

  /* based on the topology file we decide what mode we are: */
  if (names.TopoFile.size()) {
    printf("IBDMCHK Cluster Design Mode:\n");
    printf(" Topology File .. %s\n", names.TopoFile.c_str());
    printf(" SM Node ........ %s\n", names.SmNodeName.c_str());
    printf(" SM Port ........ %u\n", SmPortNum);
    printf(" LMC ............ %u\n", lmc);

    if (EnhancedRouting && lmc > 0)
      printf(" Using Enhanced Routing\n");
    if (names.RootsFileName.size())
      printf(" Roots File ..... %s\n", names.RootsFileName.c_str());

    if (fabric.parseTopology(names.TopoFile)) {
      cout << "-E- Fail to parse topology file:" << names.TopoFile << endl;
      exit(1);
    }

    // get the SM Port
    IBNode *p_smNode = fabric.getNode(names.SmNodeName);
    if (! p_smNode ) {
      cout << "-E- Fail to find SM node:" << names.SmNodeName << endl;
      exit(1);
    }

    IBPort *p_smPort = p_smNode->getPort((phys_port_t)SmPortNum);
    if (! p_smPort) {
      cout <<  "-E- Fail to find SM Port: " << names.SmNodeName
	   << "/" << SmPortNum << endl;
      exit(1);
    }

    if (!p_smPort->p_remotePort)
        cout << "-W- The port: " << p_smPort->getName()
             << " specifed as SM port is not connected to any remote port" << endl;

    // assign lids
    if (SubnMgtAssignLids(p_smPort,lmc)) {
      cout << "-E- Fail to assign LIDs." << endl;
      exit(1);
    }

    // we need to run min hop calculation anyway
    if (SubnMgtCalcMinHopTables(&fabric)) {
      cout << "-E- Fail to update Min Hops Tables." << endl;
      exit(1);
    }

    if (UseUpDown) {
      list <IBNode *> rootNodes;
      if (names.RootsFileName.size()) {
        rootNodes = ParseRootNodeNamesFile(&fabric, names.RootsFileName);
      }
      else {
        rootNodes = SubnMgtFindRootNodesByMinHop(&fabric);
      }

      if (!rootNodes.empty()) {
        cout << "-I- Recognized " << rootNodes.size() << " root nodes:" << endl;
        for (list <IBNode *>::iterator nI = rootNodes.begin();
             nI != rootNodes.end(); nI++) {
          cout << " " << (*nI)->name << endl;
        }
        cout << "---------------------------------------------------------------------------\n" << endl;

        SubnRankFabricNodesByRootNodes(&fabric, rootNodes);

        if (SubnMgtCalcUpDnMinHopTbls(&fabric)) {
          cout << "-E- Fail to update Min Hops Tables." << endl;
          exit(1);
        }
      } else {
        cout << "-E- Fail to recognize any root nodes. Up/Down is not active!" << endl;
      }
    }

    if (!EnhancedRouting) {

      if (SubnMgtOsmRoute(&fabric)) {
        cout << "-E- Fail to update LFT Tables." << endl;
        exit(1);
      }
    } else {
      if (SubnMgtOsmEnhancedRoute(&fabric)) {
        cout << "-E- Fail to update LFT Tables." << endl;
        exit(1);
      }
    }


  } else {
	  int anyMissingFile = 0;
	  // resolve the default files
	  if (names.fdbFile.size() == 0) {
		  if (access("/var/log/osm.fdbs",R_OK) == 0)
			  names.fdbFile = string("/var/log/osm.fdbs");
		  else if (access("/tmp/osm.fdbs",R_OK) == 0)
			  names.fdbFile = string("/tmp/osm.fdbs");
		  else {
			  cout << "-E- Could not find a readble osm.fdbs in /var/log or  /tmp" << endl;
			  anyMissingFile = 1;
		  }
	  }
	  if (names.mcFdbFile.size() == 0) {
		  if (access("/var/log/osm.mcfdbs",R_OK) == 0)
			  names.mcFdbFile = string("/var/log/osm.mcfdbs");
		  else if (access("/tmp/osm.mcfdbs",R_OK) == 0)
			  names.mcFdbFile = string("/tmp/osm.mcfdbs");
		  else {
			  cout << "-E- Could not find a readble osm.mcfdbs in /var/log or  /tmp" << endl;
			  anyMissingFile = 1;
		  }
	  }

	  if (names.subnetFile.size() == 0) {
		  if (access("/var/log/osm-subnet.lst",R_OK) == 0)
			  names.subnetFile = string("/var/log/osm-subnet.lst");
		  else if (access("/tmp/subnet.lst",R_OK) == 0)
			  names.subnetFile = string("/tmp/subnet.lst");
		  else {
			  cout << "-E- Could not find a readble /var/log/osm-subnet.lst or /tmp/subnet.lst" << endl;
			  anyMissingFile = 1;
		  }
	  }
	  if (anyMissingFile)
		  exit(1);

    printf(" IBDMCHK OpenSM Routing Verification Mode:\n");
    printf(" FDB File = %s\n", names.fdbFile.c_str());
    printf(" MCFDB File = %s\n", names.mcFdbFile.c_str());
    printf(" Subnet File = %s\n", names.subnetFile.c_str());
    if (names.PSLFile.size() >0)
      printf(" SL File = %s\n", names.PSLFile.c_str());
    else
      printf(" SL File = NONE, using single SL\n");
    if (names.SLVLFile.size() >0)
      printf(" SLVL File = %s\n", names.SLVLFile.c_str());
    else
      printf(" SLVL File = NONE, using identity mapping\n");
    if (names.VL2VLFile.size() >0)
      printf(" VL2VL File = %s\n", names.VL2VLFile.c_str());

    printf(" LMC CA/SW = %u/%u\n", lmc, swLmc);
    printf("-------------------------------------------------\n");

    fabric.caLmc = lmc;
    fabric.swLmc = swLmc;

    if (fabric.parseSubnetLinks(names.subnetFile)) {
      cout << "-E- Fail to parse subnet file:" << names.subnetFile << endl;
      exit(1);
    }

    fabric.markNodesAsSpecialByDescriptions();

    if (names.PLFTFile.size() && fabric.parsePLFTFile(names.PLFTFile)) {
      cout << "-E- Fail to parse PLFT  file:" << names.PLFTFile << endl;
      exit(1);
    }

    if (fabric.parseFdbFile(names.fdbFile)) {
      cout << "-E- Fail to parse fdb file:" << names.fdbFile << endl;
      exit(1);
    }

    if (names.PSLFile.size() && fabric.parsePSLFile(names.PSLFile)) {
      cout << "-E- Fail to parse SL file:" << names.PSLFile << endl;
      exit(1);
    }

    fabric.setNumVLs(fabric.getNumSLs());

    if (names.SLVLFile.size() && fabric.parseSLVLFile(names.SLVLFile)) {
      cout << "-E- Fail to parse SL file:" << names.SLVLFile << endl;
      exit(1);
    }

    if (names.VL2VLFile.size() && fabric.parseVL2VLFile(names.VL2VLFile)) {
      cout << "-E- Fail to parse VL2VL file:" << names.VL2VLFile << endl;
      exit(1);
    }

    if (names.ARFile.size() && fabric.parseARFile(names.ARFile)) {
      cout << "-E- Fail to parse AR file:" << names.ARFile << endl;
      exit(1);
    }

    if (names.FARFile.size() && fabric.parseFARFile(names.FARFile)) {
      cout << "-E- Fail to parse FAR file:" << names.FARFile << endl;
      exit(1);
    }

    if (CheckMFTCredLoops && names.SADumpFile.size() &&
        fabric.parseSADumpFile(names.SADumpFile)) {
      cout << "-E- Fail to parse SA Dump file:" << names.SADumpFile << endl;
      exit(1);
    }

    if (fabric.parseMCFdbFile(names.mcFdbFile)) {
      cout << "-E- Fail to parse multicast fdb file:" << names.mcFdbFile << endl;
      exit(1);
    }

    if (names.FLIDFile.size() && fabric.parseFLIDFile(names.FLIDFile)) {
      cout << "-E- Fail to parse FLID file:" << names.FLIDFile << endl;
      exit(1);
    }

    if (names.PHIFile.size() && fabric.parsePortHierarchyInfoFile(names.PHIFile)) {
      cout << "-E- Fail to parse port gierarchy info file:" << names.PHIFile << endl;
      exit(1);
    }

    if (names.EPFFile.size() && fabric.parseEPFFile(names.EPFFile)) {
      cout << "-E- Fail to parse entry plane filter file:" << names.EPFFile << endl;
      exit(1);
    }


  }

  cout << endl << endl;
  int anyErr = 0;

  if (names.ARFile.size() || names.FARFile.size()) {
      if (SubnMgtVerifyAllARCaToCaRoutes(&fabric)) {
          cout << "-E- Some Point to Point AR Traversals Failed." << endl;
          anyErr++;
      }

      if (SubnMgtVerifyAREmptyGroups(&fabric)) {
          cout << "-E- Empty AR Groups Validation Failed."
              << endl << endl;
          anyErr++;
      }
  } else {
    if (SubnMgtVerifyAllCaToCaRoutes(&fabric, names.outFilesDir.size() > 0 ?
          names.outFilesDir.c_str() : NULL)) {
      cout << "-E- Some Point to Point Traversals Failed." << endl;
      anyErr++;
    }
  }

  if (SubnMgtVerifyFNMPorts(&fabric, (names.ARFile.size() || names.FARFile.size()))) {
    cout << "-E- FNM Ports Routing Check Failed." << endl;
    anyErr++;
  }

  if (AllPaths && SubnMgtVerifyAllRoutes(&fabric)) {
    cout << "-E- Some Point to Point Traversals Failed." << endl;
    anyErr++;
  }

  if (SubnMgtCheckFabricMCGrps(&fabric)) {
    cout << "-E- Some Multicast Groups Routing Check Failed." << endl;
    anyErr++;
  }

  list <IBNode *> rootNodes;
  if (UseUpDown) {
	 if (names.RootsFileName.size()) {
		if (names.TopoFile.size()) {
		  rootNodes = ParseRootNodeNamesFile(&fabric, names.RootsFileName);
		} else {
		  rootNodes = ParseRootNodeGuidsFile(&fabric, names.RootsFileName);
		}
	 } else {
         // propagate lid hop count matrix:
         if (SubnMgtCalcMinHopTables(&fabric)) {
           cout << "-E- Fail to update Min Hops Tables." << endl;
           exit(1);
         }

		rootNodes = SubnMgtFindRootNodesByMinHop(&fabric);
	 }
  }

  if (!rootNodes.empty()) {
    cout << "-I- Recognized " << rootNodes.size() << " root nodes:" << endl;
    for (list <IBNode *>::iterator nI = rootNodes.begin();
         nI != rootNodes.end(); nI++) {
      cout << " " << (*nI)->name << endl;
    }
    cout << "---------------------------------------------------------------------------\n" << endl;

    // rank the fabric by these roots
    SubnRankFabricNodesByRootNodes(&fabric, rootNodes);
    ofstream rankFile;
    string err_message;
    int rc = IBFabric::OpenFile("/tmp/ibdmchk.node_ranking", rankFile, false, err_message);
    if (rc) {
        cout << "-E- " << err_message << endl;
    } else {
        rankFile << "-I- Node Ranking:" << endl;
        for(set_pnode::const_iterator nI = fabric.Switches.begin();
             nI != fabric.Switches.end();  ++nI) {
            const IBNode *p_curr_node = *nI;
            if (!p_curr_node) {
               cout << "DB error - found null node in Switches set";
               exit(1);
            }

            if (!IsRankAssigned(p_curr_node))
               exit(1);

            rankFile << p_curr_node->name << " rank:" << p_curr_node->rank << endl;
        }
        rankFile.close();
    }

    // report non up down paths:
    anyErr |= SubnReportNonUpDownCa2CaPaths(&fabric);

    // report non up down Multicast Groups
    anyErr |= SubnMgtCheckFabricMCGrpsForCreditLoopPotential(&fabric);

    // Analyze the fabric links to generate a list of src,dst pairs list
    // to cover all fabric links
    LinkCoverageAnalysis(&fabric, rootNodes);

  } else {
    cout << "-I- Using full credit loop check." << endl;
    CredLoopMode(CheckSWUnicastPathsCredLoops, CheckMFTCredLoops);
    anyErr |= CrdLoopAnalyze(&fabric, ((!names.ARFile.empty()) || (!names.FARFile.empty())));
  }

  exit(anyErr);
}
