/*
 * Copyright (c) 2023-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"

/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */

int Ibis::ClassEMadGetSet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        const data_func_set_t & attribute_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    struct MAD_Class_E class_e_mad;
    CLEAR_STRUCT(class_e_mad);

    // build class C lid route header
    this->CommonMadHeaderBuild(&class_e_mad.MAD_Header_Common,
                               IBIS_IB_CLASS_E,
                               method,
                               attribute_id,
                               attribute_modifier);

    class_e_mad.VS_Key = this->key_manager.GetKey(lid, IBIS_VS_KEY);

    // send the mad
    IBIS_RETURN(this->MadGetSet(lid,
                                1,
                                sl,
                                IBIS_IB_DEFAULT_QP1_QKEY,
                                IBIS_IB_CLASS_E,
                                method,
                                attribute_id,
                                attribute_modifier,
                                IBIS_NVL_RDM_OFFSET_CLASS_E,
                                data_func_set_t(&class_e_mad, IBIS_FUNC_LST(MAD_Class_E)),
                                attribute_data,
                                p_clbck_data));
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLClassPortInfoGet(u_int16_t lid,
        uint8_t sl,
        struct IB_ClassPortInfo *p_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_class_port_info);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo (Class 0xE) Get MAD lid = %u\n", lid);

    int rc = this->ClassEMadGetSet(lid,
            0 /*sl*/,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,  /* attribute modifier */
            data_func_set_t(p_class_port_info, IBIS_FUNC_LST(IB_ClassPortInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLClassPortInfoSet(u_int16_t lid,
        uint8_t sl,
        struct IB_ClassPortInfo *p_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo (Class 0xE) Set MAD lid = %u\n", lid);

    int rc = this->ClassEMadGetSet(lid,
            0 /*sl*/,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,  /* attribute modifier */
            data_func_set_t(p_class_port_info, IBIS_FUNC_LST(IB_ClassPortInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionInfoSet(u_int16_t lid,
        uint8_t sl,
        struct NVLReductionInfo *p_nvl_reduction_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionInfo (Class 0xE) Set MAD lid = %u\n", lid);

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_NVL_REDUCTION_INFO,
            0,  /* attribute modifier */
            data_func_set_t(p_nvl_reduction_info, IBIS_FUNC_LST(NVLReductionInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionInfoGet(u_int16_t lid,
        uint8_t sl,
        struct NVLReductionInfo *p_nvl_reduction_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_nvl_reduction_info);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionInfo (Class 0xE) Get MAD lid = %u\n", lid);

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_NVL_REDUCTION_INFO,
            0,  /* attribute modifier */
            data_func_set_t(p_nvl_reduction_info, IBIS_FUNC_LST(NVLReductionInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionPortInfoSet(u_int16_t lid,
        uint8_t sl,
        uint16_t port_number,
        uint8_t profile_select,
        struct NVLReductionPortInfo *p_nvl_reduction_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionPortInfo (Class 0xE) Set MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_number;
    attr_modifier |= profile_select << 31;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_NVL_REDUCTION_PORT_INFO,
            attr_modifier,
            data_func_set_t(p_nvl_reduction_port_info, IBIS_FUNC_LST(NVLReductionPortInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionPortInfoGet(u_int16_t lid,
        uint8_t sl,
        uint16_t port_number,
        uint8_t profile_select,
        struct NVLReductionPortInfo *p_nvl_reduction_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    CLEAR_STRUCT(*p_nvl_reduction_port_info);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionPortInfo (Class 0xE) Get MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_number;
    attr_modifier |= profile_select << 31;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_NVL_REDUCTION_PORT_INFO,
            attr_modifier,
            data_func_set_t(p_nvl_reduction_port_info, IBIS_FUNC_LST(NVLReductionPortInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionProfilesConfigSet(u_int16_t lid,
        uint8_t sl,
        uint8_t block_id,
        uint8_t feature_id,
        struct NVLReductionProfilesConfig *p_nvl_reduction_profiles_config,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionPortInfo (Class 0xE) Set MAD lid = %u\n", lid);

    uint32_t attr_modifier = block_id;
    attr_modifier |= feature_id << 8;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_NVL_REDUCTION_PROFILES_CONFIG,
            attr_modifier,
            data_func_set_t(p_nvl_reduction_profiles_config, IBIS_FUNC_LST(NVLReductionProfilesConfig)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionProfilesConfigGet(u_int16_t lid,
        uint8_t sl,
        uint8_t block_id,
        uint8_t feature_id,
        struct NVLReductionProfilesConfig *p_nvl_reduction_profiles_config,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    CLEAR_STRUCT(*p_nvl_reduction_profiles_config);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionPortInfo (Class 0xE) Get MAD lid = %u\n", lid);

    uint32_t attr_modifier = block_id;
    attr_modifier |= feature_id << 8;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_NVL_REDUCTION_PROFILES_CONFIG,
            attr_modifier,
            data_func_set_t(p_nvl_reduction_profiles_config, IBIS_FUNC_LST(NVLReductionProfilesConfig)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionForwardingTableSet(u_int16_t lid,
        uint8_t sl,
        uint16_t port_number,
        uint16_t block_index,
        struct NVLReductionForwardingTable *p_reduction_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionForwardingTable (Class 0xE) Set MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_number;
    attr_modifier |= block_index << 16;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_REDUCTION_FORWARDING_TABLE,
            attr_modifier,
            data_func_set_t(p_reduction_forwarding_table, IBIS_FUNC_LST(NVLReductionForwardingTable)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionForwardingTableGet(u_int16_t lid,
        uint8_t sl,
        uint16_t port_number,
        uint16_t block_index,
        struct NVLReductionForwardingTable *p_reduction_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    CLEAR_STRUCT(*p_reduction_forwarding_table);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionForwardingTable (Class 0xE) Get MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_number;
    attr_modifier |= block_index << 16;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_REDUCTION_FORWARDING_TABLE,
            attr_modifier,
            data_func_set_t(p_reduction_forwarding_table, IBIS_FUNC_LST(NVLReductionForwardingTable)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLPenaltyBoxConfigSet(u_int16_t lid,
        uint8_t sl,
        uint8_t block_index,
        struct NVLPenaltyBoxConfig *p_penalty_box_config,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLPenaltyBoxConfig (Class 0xE) Set MAD lid = %u\n", lid);

    uint32_t attr_modifier = block_index;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_PENALTY_BOX_CONFIG,
            attr_modifier,
            data_func_set_t(p_penalty_box_config, IBIS_FUNC_LST(NVLPenaltyBoxConfig)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLPenaltyBoxConfigGet(u_int16_t lid,
        uint8_t sl,
        uint8_t block_index,
        struct NVLPenaltyBoxConfig *p_penalty_box_config,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    CLEAR_STRUCT(*p_penalty_box_config);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLPenaltyBoxConfig (Class 0xE) Get MAD lid = %u\n", lid);

    uint32_t attr_modifier = block_index;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_PENALTY_BOX_CONFIG,
            attr_modifier,
            data_func_set_t(p_penalty_box_config, IBIS_FUNC_LST(NVLPenaltyBoxConfig)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::NVLReductionConfigureMLIDMonitorsSet(u_int16_t lid,
        uint8_t sl,
        uint16_t port_select,
        uint8_t profile_number_enable,
        struct NVLReductionConfigureMLIDMonitors *p_reduction_configure_mlid_monitors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionConfigureMLIDMonitors (Class 0xE) Set MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_select;
    attr_modifier |= profile_number_enable << 31;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_NVL_REDUCTION_CONFIGURE_MLID_MONITORS,
            attr_modifier,
            data_func_set_t(p_reduction_configure_mlid_monitors, IBIS_FUNC_LST(NVLReductionConfigureMLIDMonitors)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */

int Ibis::NVLReductionConfigureMLIDMonitorsGet(u_int16_t lid,
        uint8_t sl,
        uint16_t port_select,
        uint8_t profile_number_enable,
        struct NVLReductionConfigureMLIDMonitors *p_reduction_configure_mlid_monitors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_reduction_configure_mlid_monitors);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionConfigureMLIDMonitors (Class 0xE) Get MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_select;
    attr_modifier |= profile_number_enable << 31;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_NVL_REDUCTION_CONFIGURE_MLID_MONITORS,
            attr_modifier,
            data_func_set_t(p_reduction_configure_mlid_monitors, IBIS_FUNC_LST(NVLReductionConfigureMLIDMonitors)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::NVLReductionRoundingModeSet(u_int16_t lid,
            uint8_t sl,
            uint16_t port_select,
            uint8_t profile_number_enable,
            struct NVLReductionRoundingMode *p_reduction_rounding_mode,
            const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionRoundingModeSet (Class 0xE) Set MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_select;
    attr_modifier |= profile_number_enable << 31;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_NVL_REDUCTION_ROUTING_MODE,
            attr_modifier,
            data_func_set_t(p_reduction_rounding_mode, IBIS_FUNC_LST(NVLReductionRoundingMode)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::NVLReductionRoundingModeGet(u_int16_t lid,
            uint8_t sl,
            uint16_t port_select,
            uint8_t profile_number_enable,
            struct NVLReductionRoundingMode *p_reduction_rounding_mode,
            const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_reduction_rounding_mode);

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionRoundingModeGet (Class 0xE) Get MAD lid = %u\n", lid);

    uint32_t attr_modifier = port_select;
    attr_modifier |= profile_number_enable << 31;

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_NVL_REDUCTION_ROUTING_MODE,
            attr_modifier,
            data_func_set_t(p_reduction_rounding_mode, IBIS_FUNC_LST(NVLReductionRoundingMode)),
            p_clbck_data);

     IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */

int Ibis::NVLReductionCountersSet(u_int16_t lid,
        uint8_t sl,
        struct NVLReductionCounters *p_reduction_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionCounters (Class 0xE) Set MAD lid = %u\n", lid);

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_RDM_ATTR_NVL_REDUCTION_COUNTERS,
            0, /* attribute modifier */
            data_func_set_t(p_reduction_counters, IBIS_FUNC_LST(NVLReductionCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */

int Ibis::NVLReductionCountersGet(u_int16_t lid,
        uint8_t sl,
        struct NVLReductionCounters *p_reduction_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER; 

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending NVLReductionCounters (Class 0xE) Get MAD lid = %u\n", lid);

    int rc = this->ClassEMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_RDM_ATTR_NVL_REDUCTION_COUNTERS,
            0, /* attribute modifier */
            data_func_set_t(p_reduction_counters, IBIS_FUNC_LST(NVLReductionCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::RegisterClassETrapsHandler(mad_handler_callback_func_t callback_func,
                                     void *context) {

    return RegisterMadHandler(
            IBIS_IB_CLASS_E,
            IBIS_IB_ATTR_NOTICE,
            IBIS_IB_MAD_METHOD_TRAP,
            IBIS_NVL_RDM_OFFSET_CLASS_E,
            (const unpack_data_func_t)MAD_Class_E_unpack,
            (const dump_data_func_t)MAD_Class_E_dump,
            (const unpack_data_func_t)RDMNotice_unpack,
            (const dump_data_func_t)RDMNotice_dump,
            callback_func,
            context);
}

int Ibis::RepressClassETrap(ib_address_t *p_ib_address,
                            MAD_Class_E *p_mad_class_e,
                            RDMNotice *p_notice) {

    p_mad_class_e->MAD_Header_Common.Method = IBIS_IB_MAD_METHOD_TRAP_REPRESS;
    return MadGetSet(p_ib_address,
            IBIS_IB_CLASS_E,
            IBIS_IB_MAD_METHOD_TRAP_REPRESS,
            IBIS_NVL_RDM_OFFSET_CLASS_E,
            data_func_set_t(p_mad_class_e, IBIS_FUNC_LST(MAD_Class_E)),
            data_func_set_t(p_notice, IBIS_FUNC_LST(RDMNotice)),
            NULL);
}
